@@ -41,12 +41,15 @@
 #include <string.h>
 #include <unistd.h>
 #include <signal.h>
-#include <pthread.h>
+#include <termios.h>
 #include <sys/resource.h>
 
 #ifdef HAVE_EXECINFO_H
 #include <execinfo.h>
 #endif
+#if BACKWARD_HAS_DW == 1
+#include "backward-cpp/backward.hpp"
+#endif
 
 #include <sys/types.h>
 #include <sys/time.h>
@@ -54,7 +57,10 @@
 #include <sys/wait.h>
 #include <sys/param.h>
 
+#include <mutex>
 #include <thread>
+#include <vector>
+#include <algorithm>
 
 #ifdef HAVE_PCRE_H
 #include <pcre.h>
@@ -83,9 +89,10 @@
 #  error "SysV or X/Open-compatible Curses header file required"
 #endif
 
+#include "opt_util.hh"
 #include "lnav_log.hh"
-#include "pthreadpp.hh"
 #include "enum_util.hh"
+#include "auto_mem.hh"
 
 static const size_t BUFFER_SIZE = 256 * 1024;
 static const size_t MAX_LOG_LINE_SIZE = 2048;
@@ -104,10 +111,30 @@ nonstd::optional<FILE *> lnav_log_file;
 lnav_log_level_t lnav_log_level = lnav_log_level_t::DEBUG;
 const char *lnav_log_crash_dir;
 nonstd::optional<const struct termios *> lnav_log_orig_termios;
-static pthread_mutex_t lnav_log_mutex = PTHREAD_MUTEX_INITIALIZER;
+// NOTE: This mutex is leaked so that it is not destroyed during exit.
+// Otherwise, any attempts to log will fail.
+static std::mutex *lnav_log_mutex = new std::mutex();
 
-std::vector<log_state_dumper*> log_state_dumper::DUMPER_LIST;
-std::vector<log_crash_recoverer*> log_crash_recoverer::CRASH_LIST;
+static std::vector<log_state_dumper*>& DUMPER_LIST()
+{
+    static std::vector<log_state_dumper*> retval;
+
+    return retval;
+}
+static std::vector<log_crash_recoverer*> CRASH_LIST;
+
+struct thid {
+    static uint32_t COUNTER;
+
+    thid() noexcept : t_id(COUNTER++) {}
+
+    uint32_t t_id;
+};
+
+uint32_t thid::COUNTER = 0;
+
+thread_local thid current_thid;
+thread_local std::string thread_log_prefix;
 
 static struct {
     size_t lr_length;
@@ -118,6 +145,7 @@ static struct {
     0,
     BUFFER_SIZE,
     0,
+    {}
 };
 
 static const char *LEVEL_NAMES[] = {
@@ -162,7 +190,7 @@ void log_argv(int argc, char *argv[])
     const char *log_path = getenv("LNAV_LOG_PATH");
 
     if (log_path != nullptr) {
-        lnav_log_file = fopen(log_path, "a");
+        lnav_log_file = make_optional_from_nullable(fopen(log_path, "a"));
     }
 
     log_info("argv[%d] =", argc);
@@ -171,6 +199,11 @@ void log_argv(int argc, char *argv[])
     }
 }
 
+void log_set_thread_prefix(std::string prefix)
+{
+    // thread_log_prefix = std::move(prefix);
+}
+
 void log_host_info()
 {
     char cwd[MAXPATHLEN];
@@ -256,7 +289,7 @@ void log_msg(lnav_log_level_t level, const char *src_file, int line_number,
         return;
     }
 
-    mutex_guard mg(lnav_log_mutex);
+    std::lock_guard<std::mutex> log_lock(*lnav_log_mutex);
 
     {
         // get the base name of the file.  NB: can't use basename() since it
@@ -278,7 +311,7 @@ void log_msg(lnav_log_level_t level, const char *src_file, int line_number,
     auto line = log_alloc();
     prefix_size = snprintf(
         line, MAX_LOG_LINE_SIZE,
-        "%4d-%02d-%02dT%02d:%02d:%02d.%03d %s %s:%d ",
+        "%4d-%02d-%02dT%02d:%02d:%02d.%03d %s t%u %s:%d ",
         localtm.tm_year + 1900,
         localtm.tm_mon + 1,
         localtm.tm_mday,
@@ -286,9 +319,18 @@ void log_msg(lnav_log_level_t level, const char *src_file, int line_number,
         localtm.tm_min,
         localtm.tm_sec,
         (int)(curr_time.tv_usec / 1000),
-        LEVEL_NAMES[to_underlying(level)],
+        LEVEL_NAMES[lnav::enums::to_underlying(level)],
+        current_thid.t_id,
         src_file,
         line_number);
+#if 0
+    if (!thread_log_prefix.empty()) {
+        prefix_size += snprintf(
+            &line[prefix_size], MAX_LOG_LINE_SIZE - prefix_size,
+            "%s ",
+            thread_log_prefix.c_str());
+    }
+#endif
     rc = vsnprintf(&line[prefix_size], MAX_LOG_LINE_SIZE - prefix_size,
         fmt, args);
     if (rc >= (ssize_t)(MAX_LOG_LINE_SIZE - prefix_size)) {
@@ -305,7 +347,7 @@ void log_msg(lnav_log_level_t level, const char *src_file, int line_number,
 
 void log_msg_extra(const char *fmt, ...)
 {
-    mutex_guard mg(lnav_log_mutex);
+    std::lock_guard<std::mutex> mg(*lnav_log_mutex);
     va_list args;
 
     va_start(args, fmt);
@@ -321,7 +363,7 @@ void log_msg_extra(const char *fmt, ...)
 
 void log_msg_extra_complete()
 {
-    mutex_guard mg(lnav_log_mutex);
+    std::lock_guard<std::mutex> mg(*lnav_log_mutex);
     auto line = log_alloc();
     line[0] = '\n';
     log_ring.lr_length += 1;
@@ -333,7 +375,7 @@ void log_msg_extra_complete()
 
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wunused-result"
-static void sigabrt(int sig)
+static void sigabrt(int sig, siginfo_t *info, void *ctx)
 {
     char crash_path[1024], latest_crash_path[1024];
     int fd;
@@ -376,6 +418,43 @@ static void sigabrt(int sig)
         (void)write(fd, log_ring.lr_data, log_ring.lr_length);
 #ifdef HAVE_EXECINFO_H
         backtrace_symbols_fd(frames, frame_count, fd);
+#endif
+#if BACKWARD_HAS_DW == 1
+        {
+            ucontext_t *uctx = static_cast<ucontext_t *>(ctx);
+            void *error_addr = nullptr;
+
+#ifdef REG_RIP // x86_64
+            error_addr = reinterpret_cast<void *>(uctx->uc_mcontext.gregs[REG_RIP]);
+#elif defined(REG_EIP) // x86_32
+            error_addr = reinterpret_cast<void *>(uctx->uc_mcontext.gregs[REG_EIP]);
+#endif
+
+            backward::StackTrace st;
+
+            if (error_addr) {
+                st.load_from(error_addr, 32, reinterpret_cast<void *>(uctx),
+                             info->si_addr);
+            } else {
+                st.load_here(32, reinterpret_cast<void *>(uctx), info->si_addr);
+            }
+            backward::TraceResolver tr;
+
+            tr.load_stacktrace(st);
+            for (size_t lpc = 0; lpc < st.size(); lpc++) {
+                auto trace = tr.resolve(st[lpc]);
+                char buf[1024];
+
+                snprintf(buf, sizeof(buf),
+                         "Frame %lu:%s:%s (%s:%d)\n",
+                         lpc,
+                         trace.object_filename.c_str(),
+                         trace.object_function.c_str(),
+                         trace.source.filename.c_str(),
+                         trace.source.line);
+                write(fd, buf, strlen(buf));
+            }
+        }
 #endif
         log_ring.lr_length = 0;
         log_ring.lr_frag_start = BUFFER_SIZE;
@@ -383,7 +462,7 @@ static void sigabrt(int sig)
 
         log_host_info();
 
-        for (auto lsd : log_state_dumper::DUMPER_LIST) {
+        for (auto lsd : DUMPER_LIST()) {
             lsd->log_state();
         }
 
@@ -408,7 +487,7 @@ static void sigabrt(int sig)
     }
 
     lnav_log_orig_termios | [](auto termios) {
-        for (auto lcr : log_crash_recoverer::CRASH_LIST) {
+        for (auto lcr : CRASH_LIST) {
             lcr->log_crash_recover();
         }
 
@@ -466,16 +545,39 @@ static void sigabrt(int sig)
 
 void log_install_handlers()
 {
-    signal(SIGABRT, sigabrt);
-    signal(SIGSEGV, sigabrt);
-    signal(SIGBUS, sigabrt);
-    signal(SIGILL, sigabrt);
-    signal(SIGFPE, sigabrt);
+    const size_t stack_size = 8 * 1024 * 1024;
+    const int sigs[] = {
+        SIGABRT,
+        SIGSEGV,
+        SIGBUS,
+        SIGILL,
+        SIGFPE,
+    };
+    static auto_mem<void> stack_content;
+
+    stack_t ss;
+
+    stack_content = malloc(stack_size);
+    ss.ss_sp = stack_content;
+    ss.ss_size = stack_size;
+    ss.ss_flags = 0;
+    sigaltstack(&ss, nullptr);
+    for (const auto sig : sigs) {
+        struct sigaction sa;
+
+        memset(&sa, 0, sizeof(sa));
+        sa.sa_flags = SA_SIGINFO | SA_ONSTACK | SA_NODEFER | SA_RESETHAND;
+        sigfillset(&sa.sa_mask);
+        sigdelset(&sa.sa_mask, sig);
+        sa.sa_sigaction = sigabrt;
+
+        sigaction(sig, &sa, nullptr);
+    }
 }
 
 void log_abort()
 {
-    sigabrt(SIGABRT);
+    raise(SIGABRT);
     _exit(1);
 }
 
@@ -512,18 +614,14 @@ void log_pipe_err(int fd)
 
 log_state_dumper::log_state_dumper()
 {
-    DUMPER_LIST.push_back(this);
+    DUMPER_LIST().push_back(this);
 }
 
 log_state_dumper::~log_state_dumper()
 {
-    auto iter = std::find(DUMPER_LIST.begin(),
-                          DUMPER_LIST.end(),
-                          this);
-    if (iter != DUMPER_LIST.end()) {
-        DUMPER_LIST.erase(std::find(DUMPER_LIST.begin(),
-                                    DUMPER_LIST.end(),
-                                    this));
+    auto iter = std::find(DUMPER_LIST().begin(), DUMPER_LIST().end(), this);
+    if (iter != DUMPER_LIST().end()) {
+        DUMPER_LIST().erase(iter);
     }
 }
 
