@@ -10,7 +10,8 @@
 
 #pragma once
 
-#include <iostream>
+#include <stdio.h>
+#include <exception>
 #include <functional>
 #include <type_traits>
 
@@ -33,7 +34,10 @@ namespace types {
 
         E val;
     };
-}
+
+    template<>
+    struct Err<void> { };
+};
 
 template<typename T, typename CleanT = typename std::decay<T>::type>
 types::Ok<CleanT> Ok(T&& val) {
@@ -49,6 +53,10 @@ types::Err<CleanE> Err(E&& val) {
     return types::Err<CleanE>(std::forward<E>(val));
 }
 
+inline types::Err<void> Err() {
+    return {};
+}
+
 template<typename T, typename E> struct Result;
 
 namespace details {
@@ -554,7 +562,7 @@ struct Storage {
 
     void construct(types::Ok<T> ok)
     {
-        new (&storage_) T(ok.val);
+        new (&storage_) T(std::move(ok.val));
         initialized_ = true;
     }
     void construct(types::Err<E> err)
@@ -629,16 +637,22 @@ struct Storage<void, E> {
         }
     }
 
-    template<typename U>
+    template<typename U,
+        typename = std::enable_if_t<!std::is_same<U, void>::value>>
     const U& get() const {
         return *reinterpret_cast<const U *>(&storage_);
     }
 
-    template<typename U>
-    U& get() {
+    template<typename U,
+        typename = std::enable_if_t<!std::is_same<U, void>::value>>
+    typename std::add_lvalue_reference<U>::type get() {
         return *reinterpret_cast<U *>(&storage_);
     }
 
+    template<typename U,
+        typename = std::enable_if_t<std::is_same<U, void>::value>>
+    void get() {}
+
     type storage_;
     bool initialized_;
 };
@@ -758,7 +772,7 @@ struct Result {
 
     T expect(const char* str) const {
         if (!isOk()) {
-            std::fprintf(stderr, "%s\n", str);
+            ::fprintf(stderr, "%s\n", str);
             std::terminate(); 
         }
         return expect_impl(std::is_same<T, void>());
@@ -766,10 +780,10 @@ struct Result {
 
     template<typename Func>
     auto map(Func func) const {
-        auto value = this->storage().template get<T>();
-        using return_type = decltype(func(value));
+        using return_type = decltype(func(T{}));
 
         if (this->isOk()) {
+            auto value = this->storage().template get<T>();
             auto res = func(value);
             return Result<return_type, E>(types::Ok<return_type>(std::move(res)));
         }
@@ -790,17 +804,32 @@ struct Result {
     }
 
     template<typename Func>
-    Result<T, E> then(Func func) const {
+    Result<void, E> then(Func func) {
         if (this->isOk()) {
-            func(this->storage().template get<T>());
+            func(std::move(this->storage().template get<T>()));
+
+            return Ok();
         }
 
-        return *this;
+        return Err(std::move(this->storage().template get<E>()));
     }
 
     template<typename Func>
-    Result<T, E> otherwise(Func func) const {
-        return details::otherwise(*this, func);
+    Result<typename std::result_of<Func>::type, E> then(Func func) {
+        if (this->isOk()) {
+            return Ok(func(std::move(this->storage().template get<T>())));
+        }
+
+        return Err(std::move(this->storage().template get<E>()));
+    }
+
+    template<typename Func>
+    void otherwise(Func func) {
+        if (this->isOk()) {
+            return;
+        }
+
+        func(std::move(this->storage().template get<E>()));
     }
 
     template<typename Func,
@@ -835,6 +864,14 @@ struct Result {
         return defaultValue;
     }
 
+    template<typename Func>
+    auto unwrapOrElse(Func func) const {
+        if (isOk()) {
+            return storage().template get<T>();
+        }
+        return func(this->storage().template get<E>());
+    }
+
     template<typename U = T>
     typename std::enable_if<
         !std::is_same<U, void>::value,
@@ -842,10 +879,38 @@ struct Result {
     >::type
     unwrap() const {
         if (isOk()) {
-            return storage().template get<U>();
+            return std::move(storage().template get<U>());
         }
 
-        std::fprintf(stderr, "Attempting to unwrap an error Result\n");
+        ::fprintf(stderr, "Attempting to unwrap an error Result\n");
+        std::terminate();
+    }
+
+    template<typename U = T>
+    typename std::enable_if<
+        !std::is_same<U, void>::value,
+        U
+    >::type
+    unwrap() {
+        if (isOk()) {
+            return std::move(storage().template get<U>());
+        }
+
+        ::fprintf(stderr, "Attempting to unwrap an error Result\n");
+        std::terminate();
+    }
+
+    template<typename U = T>
+    typename std::enable_if<
+        std::is_same<U, void>::value,
+        U
+    >::type
+    unwrap() const {
+        if (isOk()) {
+            return;
+        }
+
+        ::fprintf(stderr, "Attempting to unwrap an error Result\n");
         std::terminate();
     }
 
@@ -854,7 +919,7 @@ struct Result {
             return storage().template get<E>();
         }
 
-        std::fprintf(stderr, "Attempting to unwrapErr an ok Result\n");
+        ::fprintf(stderr, "Attempting to unwrapErr an ok Result\n");
         std::terminate();
     }
 
@@ -905,9 +970,8 @@ bool operator==(const Result<T, E>& lhs, types::Err<E> err) {
     ({                                                             \
         auto res = __VA_ARGS__;                                    \
         if (!res.isOk()) {                                         \
-            typedef details::ResultErrType<decltype(res)>::type E; \
-            return types::Err<E>(res.storage().get<E>());          \
+            typedef typename ::details::ResultErrType<decltype(res)>::type E; \
+            return ::types::Err<E>(res.storage().template get<E>()); \
         }                                                          \
-        typedef details::ResultOkType<decltype(res)>::type T;      \
-        res.storage().get<T>();                                    \
+        res.unwrap();                         \
     })
