@@ -29,11 +29,16 @@
 
 #include "config.h"
 
-#include "lnav.hh"
+#include "base/humanize.time.hh"
+#include "lnav_util.hh"
+#include "ansi_scrubber.hh"
 #include "vtab_module.hh"
 #include "relative_time.hh"
 #include "field_overlay_source.hh"
 #include "readline_highlighters.hh"
+#include "vtab_module_json.hh"
+#include "log_format_ext.hh"
+#include "log_vtab_impl.hh"
 
 using namespace std;
 
@@ -41,7 +46,8 @@ json_string extract(const char *str);
 
 void field_overlay_source::build_summary_lines(const listview_curses &lv)
 {
-    textfile_sub_source &tss = lnav_data.ld_text_source;
+    auto& tc = dynamic_cast<const textview_curses &>(lv);
+    textfile_sub_source &tss = this->fos_tss;
     logfile_sub_source &lss = this->fos_lss;
 
     this->fos_summary_lines.clear();
@@ -55,44 +61,42 @@ void field_overlay_source::build_summary_lines(const listview_curses &lv)
 
         lv.get_dimensions(height, width);
         free_rows = height - filled_rows - vis_line_t(this->fos_lines.size());
-        if (free_rows < 2 || lnav_data.ld_flags & LNF_HEADLESS) {
+        if (free_rows < 2 || !this->fos_show_status) {
             this->fos_summary_lines.clear();
         }
         else {
-            string last_time, time_span;
+            string time_span;
             double error_rate = 0.0;
 
             if (lv.get_inner_height() == 0) {
-                last_time = "No log messages";
                 time_span = "None";
             }
             else {
                 logline *first_line, *last_line;
-                time_t now = time(NULL);
+                time_t now = time(nullptr);
 
                 first_line = lss.find_line(lss.at(vis_line_t(0)));
                 last_line = lss.find_line(lss.at(lv.get_bottom()));
-                last_time = "Last message: " ANSI_BOLD_START + precise_time_ago(
-                    last_line->get_timeval(), true) + ANSI_NORM;
-                duration2str(last_line->get_time_in_millis() -
-                             first_line->get_time_in_millis(),
-                             time_span);
+                time_span = humanize::time::duration::from_tv(
+                    last_line->get_timeval() - first_line->get_timeval())
+                    .to_string();
 
                 time_t local_now = convert_log_time_to_local(now);
                 time_t five_minutes_ago = local_now - (5 * 60 * 60);
                 time_t ten_secs_ago = local_now - 10;
 
-                vis_line_t from_five_min_ago = lss.find_from_time(five_minutes_ago);
-                vis_line_t from_ten_secs_ago = lss.find_from_time(ten_secs_ago);
-                vis_bookmarks &bm = lnav_data.ld_views[LNV_LOG].get_bookmarks();
-                bookmark_vector<vis_line_t> &error_bookmarks =
-                    bm[&logfile_sub_source::BM_ERRORS];
+                auto from_five_min_ago_opt = lss.find_from_time(five_minutes_ago);
+                auto from_ten_secs_ago_opt = lss.find_from_time(ten_secs_ago);
+                auto &bm = tc.get_bookmarks();
+                auto error_bm_iter = bm.find(&logfile_sub_source::BM_ERRORS);
 
-                if (now > last_line->get_time() && from_five_min_ago != -1) {
+                if (now > last_line->get_time() && from_five_min_ago_opt &&
+                    error_bm_iter != bm.end()) {
+                    auto& error_bookmarks = error_bm_iter->second;
                     auto five_min_lower =
                         lower_bound(error_bookmarks.begin(),
                                     error_bookmarks.end(),
-                                    from_five_min_ago);
+                                    from_five_min_ago_opt.value());
                     if (five_min_lower != error_bookmarks.end()) {
                         double error_count = distance(
                             five_min_lower, error_bookmarks.end());
@@ -104,11 +108,11 @@ void field_overlay_source::build_summary_lines(const listview_curses &lv)
                         }
                         error_rate = error_count / time_diff;
 
-                        if (from_ten_secs_ago != -1) {
+                        if (from_ten_secs_ago_opt) {
                             auto ten_sec_lower =
                                 lower_bound(error_bookmarks.begin(),
                                             error_bookmarks.end(),
-                                            from_ten_secs_ago);
+                                            from_ten_secs_ago_opt.value());
                             if (ten_sec_lower != error_bookmarks.end()) {
                                 double recent_error_count = distance(
                                     ten_sec_lower, error_bookmarks.end());
@@ -133,25 +137,23 @@ void field_overlay_source::build_summary_lines(const listview_curses &lv)
             attr_line_t &sum_line = this->fos_summary_lines.back();
             if (tss.empty()) {
                 sum_line.with_ansi_string(
-                    "       %s; "
+                    "       "
                         "Files: " ANSI_BOLD("%'2d") "; "
                         ANSI_ROLE("Error rate") ": " ANSI_BOLD(
                         "%'.2lf") "/min; "
                         "Time span: " ANSI_BOLD("%s"),
-                    last_time.c_str(),
                     lss.file_count(),
                     view_colors::VCR_ERROR,
                     error_rate,
                     time_span.c_str());
             } else {
                 sum_line.with_ansi_string(
-                    "       %s; "
+                    "       "
                         "Log Files: " ANSI_BOLD("%'2d") "; "
                         "Text Files: " ANSI_BOLD("%'2d") "; "
                         ANSI_ROLE("Error rate") ": " ANSI_BOLD(
                         "%'.2lf") "/min; "
                         "Time span: " ANSI_BOLD("%s"),
-                    last_time.c_str(),
                     lss.file_count(),
                     tss.size(),
                     view_colors::VCR_ERROR,
@@ -208,14 +210,15 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
     content_line_t cl = lss.at(lv.get_top());
     std::shared_ptr<logfile> file = lss.find(cl);
     auto ll = file->begin() + cl;
-    log_format *format = file->get_format();
+    auto format = file->get_format();
     bool display = false;
 
-    if (ll->is_time_skewed()) {
+    if (ll->is_time_skewed() ||
+        ll->get_msg_level() == log_level_t::LEVEL_INVALID) {
         display = true;
     }
-    if (this->fos_active) {
-        display = true;
+    if (!this->fos_contexts.empty()) {
+        display = display || this->fos_contexts.top().c_show;
     }
 
     this->build_meta_line(lv, this->fos_lines, lv.get_top());
@@ -228,8 +231,27 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
         return;
     }
 
+    if (ll->get_msg_level() == LEVEL_INVALID) {
+        for (const auto& sattr : this->fos_log_helper.ldh_line_attrs) {
+            if (sattr.sa_type != &SA_INVALID) {
+                continue;
+            }
+
+            auto emsg = fmt::format("   Invalid log message: {}",
+                                    (const char *) sattr.sa_value.sav_ptr);
+            auto al = attr_line_t(emsg)
+                .with_attr(string_attr(line_range{1, 2},
+                                       &view_curses::VC_GRAPHIC,
+                                       ACS_LLCORNER))
+                .with_attr(string_attr(line_range{0, 22},
+                                       &view_curses::VC_ROLE,
+                                       view_colors::VCR_INVALID_MSG));
+            this->fos_lines.emplace_back(al);
+        }
+    }
+
     char old_timestamp[64], curr_timestamp[64], orig_timestamp[64];
-    struct timeval curr_tv, offset_tv, orig_tv, diff_tv = { 0 };
+    struct timeval curr_tv, offset_tv, orig_tv, diff_tv = { 0, 0 };
     attr_line_t time_line;
     string &time_str = time_line.get_string();
     struct line_range time_lr;
@@ -247,8 +269,8 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
         time_str.append("   Out-Of-Time-Order Message");
         time_lr.lr_start = 3;
         time_lr.lr_end = time_str.length();
-        time_line.with_attr(string_attr(time_lr, &view_curses::VC_STYLE,
-                                        vc.attrs_for_role(view_colors::VCR_SKEWED_TIME)));
+        time_line.with_attr(string_attr(time_lr, &view_curses::VC_ROLE,
+                                        view_colors::VCR_SKEWED_TIME));
         time_str.append(" --");
     }
 
@@ -259,7 +281,9 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
     time_line.with_attr(string_attr(time_lr, &view_curses::VC_STYLE, A_BOLD));
     time_str.append(" -- ");
     time_lr.lr_start = time_str.length();
-    time_str.append(precise_time_ago(ll->get_timeval(), true));
+    time_str.append(humanize::time::point::from_tv(ll->get_timeval())
+                        .with_convert_to_local(true)
+                        .as_precise_time_ago());
     time_lr.lr_end = time_str.length();
     time_line.with_attr(string_attr(time_lr, &view_curses::VC_STYLE, A_BOLD));
 
@@ -290,13 +314,14 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
             time_lr.lr_end = time_str.length();
             time_line.with_attr(string_attr(
                 time_lr,
-                &view_curses::VC_STYLE,
-                vc.attrs_for_role(view_colors::VCR_SKEWED_TIME)));
+                &view_curses::VC_ROLE,
+                view_colors::VCR_SKEWED_TIME));
 
             timersub(&curr_tv, &actual_tv, &diff_tv);
             time_str.append(";  Diff: ");
             time_lr.lr_start = time_str.length();
-            str2reltime(diff_tv, time_str);
+            time_str.append(humanize::time::duration::from_tv(diff_tv)
+                                .to_string());
             time_lr.lr_end = time_str.length();
             time_line.with_attr(string_attr(
                 time_lr,
@@ -321,25 +346,32 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
         time_str.append(offset_str);
     }
 
-    if (this->fos_active || diff_tv.tv_sec > 0) {
+    if ((!this->fos_contexts.empty() && this->fos_contexts.top().c_show) ||
+        diff_tv.tv_sec > 0) {
         this->fos_lines.emplace_back(time_line);
     }
 
-    if (!this->fos_active) {
+    if (this->fos_contexts.empty() || !this->fos_contexts.top().c_show) {
         return;
     }
 
-    this->fos_known_key_size = 0;
+    this->fos_known_key_size = LOG_BODY.length();
     this->fos_unknown_key_size = 0;
 
     for (auto & ldh_line_value : this->fos_log_helper.ldh_line_values) {
-        int this_key_size = ldh_line_value.lv_name.size();
+        auto& meta = ldh_line_value.lv_meta;
+        int this_key_size = meta.lvm_name.size();
 
-        if (ldh_line_value.lv_kind == logline_value::VALUE_STRUCT) {
+        if (!this->fos_contexts.empty()) {
+            this_key_size += this->fos_contexts.top().c_prefix.length();
+        }
+        if (meta.lvm_kind == value_kind_t::VALUE_STRUCT) {
             this_key_size += 9;
         }
-        this->fos_known_key_size = max(
-            this->fos_known_key_size, this_key_size);
+        if (!meta.lvm_struct_name.empty()) {
+            this_key_size += meta.lvm_struct_name.size() + 11;
+        }
+        this->fos_known_key_size = max(this->fos_known_key_size, this_key_size);
     }
 
     for (auto iter = this->fos_log_helper.ldh_parser->dp_pairs.begin();
@@ -353,7 +385,7 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
                 this->fos_unknown_key_size, (int)colname.length());
     }
 
-    log_format *lf = this->fos_log_helper.ldh_file->get_format();
+    auto lf = this->fos_log_helper.ldh_file->get_format();
     if (!lf->get_pattern_regex(cl).empty()) {
         attr_line_t pattern_al;
         std::string &pattern_str = pattern_al.get_string();
@@ -369,14 +401,20 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
         this->fos_lines.emplace_back(" No known message fields");
     }
 
-    const log_format *last_format = NULL;
+    const log_format *last_format = nullptr;
 
     for (auto & lv : this->fos_log_helper.ldh_line_values) {
-        string format_name = lv.lv_format->get_name().to_string();
+        if (!lv.lv_meta.lvm_format) {
+            continue;
+        }
+
+        auto curr_format = lv.lv_meta.lvm_format.value();
+        auto curr_elf = dynamic_cast<external_log_format *>(curr_format);
+        string format_name = curr_format->get_name().to_string();
         attr_line_t al;
         string str, value_str = lv.to_string();
 
-        if (lv.lv_format != last_format) {
+        if (curr_format != last_format) {
             this->fos_lines.emplace_back(" Known message fields for table " +
                                          format_name +
                                          ":");
@@ -384,32 +422,61 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
                 line_range(32, 32 + format_name.length()),
                 &view_curses::VC_STYLE,
                 vc.attrs_for_ident(format_name) | A_BOLD));
-            last_format = lv.lv_format;
+            last_format = curr_format;
         }
 
-        str = "   " + lv.lv_name.to_string();
-        str.append(this->fos_known_key_size - lv.lv_name.size() + 3, ' ');
+        std::string field_name, orig_field_name;
+        if (lv.lv_meta.lvm_struct_name.empty()) {
+            if (curr_elf && curr_elf->elf_body_field == lv.lv_meta.lvm_name) {
+                field_name = LOG_BODY;
+            } else if (curr_elf && curr_elf->lf_timestamp_field == lv.lv_meta.lvm_name) {
+                field_name = LOG_TIME;
+            } else {
+                field_name = lv.lv_meta.lvm_name.to_string();
+            }
+            orig_field_name = field_name;
+            if (!this->fos_contexts.empty()) {
+                field_name = this->fos_contexts.top().c_prefix + field_name;
+            }
+            str = "   " + field_name;
+        } else {
+            auto_mem<char, sqlite3_free> jgetter;
+
+            jgetter = sqlite3_mprintf("   jget(%s, '/%q')",
+                                      lv.lv_meta.lvm_struct_name.get(),
+                                      lv.lv_meta.lvm_name.get());
+            str = jgetter;
+        }
+        str.append(this->fos_known_key_size - (str.length() - 3), ' ');
         str += " = " + value_str;
 
-        al.with_string(str)
-                .with_attr(string_attr(
-                        line_range(3, 3 + lv.lv_name.size()),
-                        &view_curses::VC_STYLE,
-                        vc.attrs_for_ident(lv.lv_name.to_string())));
+        al.with_string(str);
+        if (lv.lv_meta.lvm_struct_name.empty()) {
+            auto prefix_len = field_name.length() - orig_field_name.length();
+            al.with_attr(string_attr(
+                line_range(3 + prefix_len, 3 + prefix_len + field_name.size()),
+                &view_curses::VC_STYLE,
+                vc.attrs_for_ident(orig_field_name)));
+        } else {
+            al.with_attr(string_attr(
+                line_range(8, 8 + lv.lv_meta.lvm_struct_name.size()),
+                &view_curses::VC_STYLE,
+                vc.attrs_for_ident(lv.lv_meta.lvm_struct_name)));
+        }
 
         this->fos_lines.emplace_back(al);
         this->add_key_line_attrs(this->fos_known_key_size);
 
-        if (lv.lv_kind == logline_value::VALUE_STRUCT) {
+        if (lv.lv_meta.lvm_kind == value_kind_t::VALUE_STRUCT) {
             json_string js = extract(value_str.c_str());
 
             al.clear()
               .append("   extract(")
-              .append(lv.lv_name.get(),
+              .append(lv.lv_meta.lvm_name.get(),
                       &view_curses::VC_STYLE,
-                      vc.attrs_for_ident(lv.lv_name.get(), lv.lv_name.size()))
+                      vc.attrs_for_ident(lv.lv_meta.lvm_name))
               .append(")")
-              .append(this->fos_known_key_size - lv.lv_name.size() - 9 + 3, ' ')
+              .append(this->fos_known_key_size - lv.lv_meta.lvm_name.size() - 9 + 3, ' ')
               .append(" = ")
               .append((const char *) js.js_content.in(), js.js_len);
             this->fos_lines.emplace_back(al);
@@ -438,6 +505,28 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
         }
     }
 
+    if (!this->fos_log_helper.ldh_xml_pairs.empty()) {
+        this->fos_lines.emplace_back(" XML fields:");
+    }
+
+    for (const auto& xml_pair : this->fos_log_helper.ldh_xml_pairs) {
+        auto_mem<char, sqlite3_free> qname;
+        auto_mem<char, sqlite3_free> xp_call;
+
+        qname = sql_quote_ident(xml_pair.first.first.get());
+        xp_call = sqlite3_mprintf("xpath(%Q, %s)",
+                                  xml_pair.first.second.c_str(),
+                                  qname.in());
+        this->fos_lines.emplace_back(fmt::format(
+            "   {} = {}", xp_call, xml_pair.second));
+        this->add_key_line_attrs(0);
+    }
+
+    if (!this->fos_contexts.empty() &&
+        !this->fos_contexts.top().c_show_discovered) {
+        return;
+    }
+
     if (this->fos_log_helper.ldh_parser->dp_pairs.empty()) {
         this->fos_lines.emplace_back(" No discovered message fields");
     }
@@ -448,8 +537,8 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
             &view_curses::VC_STYLE,
             vc.attrs_for_ident("logline")
         ));
-        attr_line_t &al = this->fos_lines.back();
-        string &disc_str = al.get_string();
+        auto &al = this->fos_lines.back();
+        auto &disc_str = al.get_string();
 
         al.with_attr(string_attr(
             line_range(disc_str.length(), -1),
@@ -458,15 +547,13 @@ void field_overlay_source::build_field_lines(const listview_curses &lv)
         disc_str.append(this->fos_log_helper.ldh_msg_format);
     }
 
-    data_parser::element_list_t::iterator iter;
-
-    iter = this->fos_log_helper.ldh_parser->dp_pairs.begin();
+    auto iter = this->fos_log_helper.ldh_parser->dp_pairs.begin();
     for (size_t lpc = 0;
          lpc < this->fos_log_helper.ldh_parser->dp_pairs.size(); lpc++, ++iter) {
-        string &name = this->fos_log_helper.ldh_namer->cn_names[lpc];
-        string val = this->fos_log_helper.ldh_parser->get_element_string(
+        auto &name = this->fos_log_helper.ldh_namer->cn_names[lpc];
+        auto val = this->fos_log_helper.ldh_parser->get_element_string(
                 iter->e_sub_elements->back());
-        attr_line_t al("   " + name + " = " + val);
+        attr_line_t al(fmt::format("   {} = {}", name, val));
 
         al.with_attr(string_attr(
                 line_range(3, 3 + name.length()),
@@ -521,7 +608,7 @@ void field_overlay_source::build_meta_line(const listview_curses &lv,
 
             const auto *tc = dynamic_cast<const textview_curses *>(&lv);
             if (tc) {
-                const textview_curses::highlight_map_t &hm = tc->get_highlights();
+                const auto &hm = tc->get_highlights();
                 auto hl_iter = hm.find({highlight_source_t::PREVIEW, "search"});
 
                 if (hl_iter != hm.end()) {
