@@ -29,56 +29,56 @@
 
 #include "config.h"
 
-#include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
 
-#include "lnav.hh"
-#include "auto_mem.hh"
+#include "base/injector.bind.hh"
+#include "base/lnav.gzip.hh"
 #include "base/lnav_log.hh"
-#include "sql_util.hh"
-#include "file_vtab.hh"
+#include "file_collection.hh"
+#include "logfile.hh"
 #include "session_data.hh"
 #include "vtab_module.hh"
+#include "log_format.hh"
+#include "file_vtab.cfg.hh"
 
 using namespace std;
 
 struct lnav_file : public tvt_iterator_cursor<lnav_file> {
     using iterator = vector<shared_ptr<logfile>>::iterator;
 
+    static constexpr const char *NAME = "lnav_file";
     static constexpr const char *CREATE_STMT = R"(
 -- Access lnav's open file list through this table.
 CREATE TABLE lnav_file (
     device integer,       -- The device the file is stored on.
     inode integer,        -- The inode for the file on the device.
     filepath text,        -- The path to the file.
+    mimetype text,        -- The MIME type for the file.
     format text,          -- The log file format for the file.
     lines integer,        -- The number of lines in the file.
-    time_offset integer   -- The millisecond offset for timestamps.
+    time_offset integer,  -- The millisecond offset for timestamps.
+
+    content BLOB HIDDEN   -- The contents of the file.
 );
 )";
 
-    struct vtab {
-        sqlite3_vtab base;
-
-        explicit operator sqlite3_vtab *() {
-            return &this->base;
-        };
-    };
+    explicit lnav_file(file_collection& fc) : lf_collection(fc) {
+    }
 
     iterator begin() {
-        return lnav_data.ld_files.begin();
+        return this->lf_collection.fc_files.begin();
     }
 
     iterator end() {
-        return lnav_data.ld_files.end();
+        return this->lf_collection.fc_files.end();
     }
 
     int get_column(const cursor &vc, sqlite3_context *ctx, int col) {
         auto lf = *vc.iter;
         const struct stat &st = lf->get_stat();
         const string &name = lf->get_filename();
-        log_format *format = lf->get_format();
+        auto format = lf->get_format();
         const char *format_name =
             format != nullptr ? format->get_name().get() : nullptr;
 
@@ -93,18 +93,66 @@ CREATE TABLE lnav_file (
                 to_sqlite(ctx, name);
                 break;
             case 3:
-                to_sqlite(ctx, format_name);
+                to_sqlite(ctx, fmt::format("{}", lf->get_text_format()));
                 break;
             case 4:
+                to_sqlite(ctx, format_name);
+                break;
+            case 5:
                 to_sqlite(ctx, (int64_t) lf->size());
                 break;
-            case 5: {
+            case 6: {
                 auto tv = lf->get_time_offset();
                 int64_t ms = (tv.tv_sec * 1000LL) + tv.tv_usec / 1000LL;
 
                 to_sqlite(ctx, ms);
                 break;
             }
+            case 7: {
+                auto& cfg = injector::get<const file_vtab::config&>();
+                auto lf_stat = lf->get_stat();
+
+                if (lf_stat.st_size > cfg.fvc_max_content_size) {
+                    sqlite3_result_error(ctx, "file is too large", -1);
+                } else {
+                    auto fd = lf->get_fd();
+                    auto_mem<char> buf;
+                    buf = (char *) malloc(lf_stat.st_size);
+                    auto rc = pread(fd, buf, lf_stat.st_size, 0);
+
+                    if (rc == -1) {
+                        auto errmsg = fmt::format("unable to read file: {}",
+                                                  strerror(errno));
+
+                        sqlite3_result_error(ctx, errmsg.c_str(),
+                                             errmsg.length());
+                    } else if (rc != lf_stat.st_size) {
+                        auto errmsg = fmt::format("short read of file: {} < {}",
+                                                  rc, lf_stat.st_size);
+
+                        sqlite3_result_error(ctx, errmsg.c_str(),
+                                             errmsg.length());
+                    } else if (lnav::gzip::is_gzipped(buf, rc)) {
+                        lnav::gzip::uncompress(lf->get_unique_path(), buf, rc)
+                            .then([ctx](auto uncomp) {
+                                auto pair = uncomp.release();
+
+                                sqlite3_result_blob64(ctx,
+                                                      pair.first,
+                                                      pair.second,
+                                                      free);
+                            })
+                            .otherwise([ctx](auto msg) {
+                                sqlite3_result_error(ctx,
+                                                     msg.c_str(),
+                                                     msg.size());
+                            });
+                    } else {
+                        sqlite3_result_blob64(ctx, buf.release(), rc, free);
+                    }
+                }
+                break;
+            }
             default:
                 ensure(0);
                 break;
@@ -130,10 +178,12 @@ CREATE TABLE lnav_file (
                    int64_t device,
                    int64_t inode,
                    std::string path,
+                   const char *text_format,
                    const char *format,
                    int64_t lines,
-                   int64_t time_offset) {
-        auto lf = lnav_data.ld_files[rowid];
+                   int64_t time_offset,
+                   const char *content) {
+        auto lf = this->lf_collection.fc_files[rowid];
         struct timeval tv = {
             (int) (time_offset / 1000LL),
             (int) (time_offset / (1000LL * 1000LL)),
@@ -147,15 +197,15 @@ CREATE TABLE lnav_file (
                     "real file paths cannot be updated, only symbolic ones");
             }
 
-            auto iter = lnav_data.ld_file_names.find(lf->get_filename());
+            auto iter = this->lf_collection.fc_file_names.find(lf->get_filename());
 
-            if (iter != lnav_data.ld_file_names.end()) {
-                auto loo = iter->second;
+            if (iter != this->lf_collection.fc_file_names.end()) {
+                auto loo = std::move(iter->second);
 
-                lnav_data.ld_file_names.erase(iter);
+                this->lf_collection.fc_file_names.erase(iter);
 
                 loo.loo_include_in_session = true;
-                lnav_data.ld_file_names[path] = loo;
+                this->lf_collection.fc_file_names[path] = std::move(loo);
                 lf->set_filename(path);
 
                 init_session();
@@ -163,21 +213,16 @@ CREATE TABLE lnav_file (
             }
         }
 
-
         return SQLITE_OK;
     };
+
+    file_collection &lf_collection;
 };
 
+struct injectable_lnav_file : vtab_module<lnav_file> {
+    using vtab_module<lnav_file>::vtab_module;
+    using injectable = injectable_lnav_file(file_collection&);
+};
 
-int register_file_vtab(sqlite3 *db)
-{
-    static vtab_module<lnav_file> LNAV_FILE_MODULE;
-
-    int rc;
-
-    rc = LNAV_FILE_MODULE.create(db, "lnav_file");
-
-    ensure(rc == SQLITE_OK);
-
-    return rc;
-}
+static auto file_binder = injector::bind_multiple<vtab_module_base>()
+     .add<injectable_lnav_file>();
