@@ -29,6 +29,9 @@
 
 #include "config.h"
 
+#include "base/enum_util.hh"
+#include "base/func_util.hh"
+
 #include "lnav.hh"
 
 #include "filter_sub_source.hh"
@@ -38,20 +41,25 @@
 using namespace std;
 
 filter_sub_source::filter_sub_source()
-    : fss_change_wire(*this, &filter_sub_source::rl_change),
-      fss_perform_wire(*this, &filter_sub_source::rl_perform),
-      fss_abort_wire(*this, &filter_sub_source::rl_abort),
-      fss_display_match_wire(*this, &filter_sub_source::rl_display_matches),
-      fss_display_next_wire(*this, &filter_sub_source::rl_display_next)
 {
-    this->filter_context.set_highlighter(readline_regex_highlighter)
+    this->fss_regex_context.set_highlighter(readline_regex_highlighter)
         .set_append_character(0);
-    this->fss_editor.add_context(LNM_FILTER, this->filter_context);
-    this->fss_editor.set_change_action(&this->fss_change_wire);
-    this->fss_editor.set_perform_action(&this->fss_perform_wire);
-    this->fss_editor.set_abort_action(&this->fss_abort_wire);
-    this->fss_editor.set_display_match_action(&this->fss_display_match_wire);
-    this->fss_editor.set_display_next_action(&this->fss_display_next_wire);
+    this->fss_editor.add_context(lnav::enums::to_underlying(filter_lang_t::REGEX),
+                                 this->fss_regex_context);
+    this->fss_sql_context.set_highlighter(readline_sqlite_highlighter)
+        .set_append_character(0);
+    this->fss_editor.add_context(lnav::enums::to_underlying(filter_lang_t::SQL),
+                                 this->fss_sql_context);
+    this->fss_editor.set_change_action(bind_mem(
+        &filter_sub_source::rl_change, this));
+    this->fss_editor.set_perform_action(bind_mem(
+        &filter_sub_source::rl_perform, this));
+    this->fss_editor.set_abort_action(bind_mem(
+        &filter_sub_source::rl_abort, this));
+    this->fss_editor.set_display_match_action(
+        bind_mem(&filter_sub_source::rl_display_matches, this));
+    this->fss_editor.set_display_next_action(
+        bind_mem(&filter_sub_source::rl_display_next, this));
     this->fss_match_view.set_sub_source(&this->fss_match_source);
     this->fss_match_view.set_height(0_vl);
     this->fss_match_view.set_show_scrollbar(true);
@@ -72,12 +80,13 @@ bool filter_sub_source::list_input_handle_key(listview_curses &lv, int ch)
     }
 
     switch (ch) {
-        case '\t':
-        case KEY_BTAB:
-        case 'q':
-            lnav_data.ld_mode = LNM_PAGING;
-            lnav_data.ld_filter_view.reload_data();
-            return true;
+        case 'f': {
+            auto top_view = *lnav_data.ld_view_stack.top();
+            auto tss = top_view->get_sub_source();
+
+            tss->toggle_apply_filters();
+            break;
+        }
         case ' ': {
             textview_curses *top_view = *lnav_data.ld_view_stack.top();
             text_sub_source *tss = top_view->get_sub_source();
@@ -127,35 +136,41 @@ bool filter_sub_source::list_input_handle_key(listview_curses &lv, int ch)
             shared_ptr<text_filter> tf = *(fs.begin() + lv.get_selection());
 
             fs.delete_filter(tf->get_id());
-            tss->text_filters_changed();
             lv.reload_data();
+            tss->text_filters_changed();
             return true;
         }
         case 'i': {
             textview_curses *top_view = *lnav_data.ld_view_stack.top();
             text_sub_source *tss = top_view->get_sub_source();
             filter_stack &fs = tss->get_filters();
+            auto filter_index = fs.next_index();
 
-            if (fs.full()) {
+            if (!filter_index) {
+                lnav_data.ld_filter_help_status_source.fss_error
+                    .set_value("error: too many filters");
                 return true;
             }
 
-            auto ef = make_shared<empty_filter>(text_filter::type_t::INCLUDE, fs.next_index());
+            auto ef = make_shared<empty_filter>(text_filter::type_t::INCLUDE,
+                                                *filter_index);
             fs.add_filter(ef);
             lv.set_selection(vis_line_t(fs.size() - 1));
             lv.reload_data();
 
             this->fss_editing = true;
 
-            add_view_text_possibilities(&this->fss_editor, LNM_FILTER, "*", top_view);
-            lnav_data.ld_filter_status_source.tss_prompt.set_value(
-                "Enter a regular expression to match lines to filter in:");
+            add_view_text_possibilities(&this->fss_editor,
+                                        lnav::enums::to_underlying(filter_lang_t::REGEX),
+                                        "*",
+                                        top_view);
             this->fss_editor.set_window(lv.get_window());
             this->fss_editor.set_visible(true);
-            this->fss_editor.set_y(lv.get_y() + (int) (lv.get_selection() - lv.get_top()));
-            this->fss_editor.set_left(22);
+            this->fss_editor.set_y(
+                lv.get_y() + (int) (lv.get_selection() - lv.get_top()));
+            this->fss_editor.set_left(25);
             this->fss_editor.set_width(this->tss_view->get_width() - 8 - 1);
-            this->fss_editor.focus(LNM_FILTER, "", "");
+            this->fss_editor.focus(lnav::enums::to_underlying(filter_lang_t::REGEX), "", "");
             this->fss_filter_state = true;
             ef->disable();
             return true;
@@ -164,27 +179,33 @@ bool filter_sub_source::list_input_handle_key(listview_curses &lv, int ch)
             textview_curses *top_view = *lnav_data.ld_view_stack.top();
             text_sub_source *tss = top_view->get_sub_source();
             filter_stack &fs = tss->get_filters();
+            auto filter_index = fs.next_index();
 
-            if (fs.full()) {
+            if (!filter_index) {
+                lnav_data.ld_filter_help_status_source.fss_error
+                    .set_value("error: too many filters");
                 return true;
             }
 
-            auto ef = make_shared<empty_filter>(text_filter::type_t::EXCLUDE, fs.next_index());
+            auto ef = make_shared<empty_filter>(text_filter::type_t::EXCLUDE,
+                                                *filter_index);
             fs.add_filter(ef);
             lv.set_selection(vis_line_t(fs.size() - 1));
             lv.reload_data();
 
             this->fss_editing = true;
 
-            add_view_text_possibilities(&this->fss_editor, LNM_FILTER, "*", top_view);
-            lnav_data.ld_filter_status_source.tss_prompt.set_value(
-                "Enter a regular expression to match lines to filter out:");
+            add_view_text_possibilities(&this->fss_editor,
+                                        lnav::enums::to_underlying(filter_lang_t::REGEX),
+                                        "*",
+                                        top_view);
             this->fss_editor.set_window(lv.get_window());
             this->fss_editor.set_visible(true);
-            this->fss_editor.set_y(lv.get_y() + (int) (lv.get_selection() - lv.get_top()));
-            this->fss_editor.set_left(22);
+            this->fss_editor.set_y(
+                lv.get_y() + (int) (lv.get_selection() - lv.get_top()));
+            this->fss_editor.set_left(25);
             this->fss_editor.set_width(this->tss_view->get_width() - 8 - 1);
-            this->fss_editor.focus(LNM_FILTER, "", "");
+            this->fss_editor.focus(lnav::enums::to_underlying(filter_lang_t::REGEX), "", "");
             this->fss_filter_state = true;
             ef->disable();
             return true;
@@ -203,18 +224,39 @@ bool filter_sub_source::list_input_handle_key(listview_curses &lv, int ch)
 
             this->fss_editing = true;
 
-            add_view_text_possibilities(&this->fss_editor, LNM_FILTER, "*", top_view);
+            add_view_text_possibilities(&this->fss_editor,
+                                        lnav::enums::to_underlying(filter_lang_t::REGEX),
+                                        "*",
+                                        top_view);
+            add_filter_expr_possibilities(&this->fss_editor,
+                                          lnav::enums::to_underlying(filter_lang_t::SQL),
+                                          "*");
             this->fss_editor.set_window(lv.get_window());
             this->fss_editor.set_visible(true);
-            this->fss_editor.set_y(lv.get_y() + (int) (lv.get_selection() - lv.get_top()));
-            this->fss_editor.set_left(22);
+            this->fss_editor.set_y(
+                lv.get_y() + (int) (lv.get_selection() - lv.get_top()));
+            this->fss_editor.set_left(25);
             this->fss_editor.set_width(this->tss_view->get_width() - 8 - 1);
-            this->fss_editor.focus(LNM_FILTER, "", tf->get_id().c_str());
+            this->fss_editor.focus(lnav::enums::to_underlying(tf->get_lang()), "");
+            this->fss_editor.rewrite_line(0, tf->get_id().c_str());
             this->fss_filter_state = tf->is_enabled();
             tf->disable();
             tss->text_filters_changed();
             return true;
         }
+        case 'n': {
+            execute_command(lnav_data.ld_exec_context, "next-mark search");
+            return true;
+        }
+        case 'N': {
+            execute_command(lnav_data.ld_exec_context, "prev-mark search");
+            return true;
+        }
+        case '/': {
+            execute_command(lnav_data.ld_exec_context,
+                            "prompt search-filters");
+            return true;
+        }
         default:
             log_debug("unhandled %x", ch);
             break;
@@ -225,7 +267,7 @@ bool filter_sub_source::list_input_handle_key(listview_curses &lv, int ch)
 
 size_t filter_sub_source::text_line_count()
 {
-    return (lnav_data.ld_view_stack.top() | [] (auto tc) {
+    return (lnav_data.ld_view_stack.top() | [](auto tc) {
         text_sub_source *tss = tc->get_sub_source();
         filter_stack &fs = tss->get_filters();
 
@@ -263,7 +305,11 @@ void filter_sub_source::text_value_for_line(textview_curses &tc, int line,
             value_out.append(" IN ");
             break;
         case text_filter::EXCLUDE:
-            value_out.append("OUT ");
+            if (tf->get_lang() == filter_lang_t::REGEX) {
+                value_out.append("OUT ");
+            } else {
+                value_out.append("    ");
+            }
             break;
         default:
             ensure(0);
@@ -271,9 +317,10 @@ void filter_sub_source::text_value_for_line(textview_curses &tc, int line,
     }
 
     if (this->fss_editing && line == tc.get_selection()) {
-        snprintf(hits, sizeof(hits), "%6s hits | ", "-");
+        snprintf(hits, sizeof(hits), "%9s hits | ", "-");
     } else {
-        snprintf(hits, sizeof(hits), "%6d hits | ", tss->get_filtered_count_for(tf->get_index()));
+        snprintf(hits, sizeof(hits), "%'9d hits | ",
+                 tss->get_filtered_count_for(tf->get_index()));
     }
     value_out.append(hits);
 
@@ -288,11 +335,12 @@ void filter_sub_source::text_attrs_for_line(textview_curses &tc, int line,
     text_sub_source *tss = top_view->get_sub_source();
     filter_stack &fs = tss->get_filters();
     shared_ptr<text_filter> tf = *(fs.begin() + line);
-    bool selected = lnav_data.ld_mode == LNM_FILTER && line == tc.get_selection();
-    int bg = selected ? COLOR_WHITE : COLOR_BLACK;
+    bool selected =
+        lnav_data.ld_mode == LNM_FILTER && line == tc.get_selection();
 
     if (selected) {
-        value_out.emplace_back(line_range{0, 1}, &view_curses::VC_GRAPHIC, ACS_RARROW);
+        value_out.emplace_back(line_range{0, 1}, &view_curses::VC_GRAPHIC,
+                               ACS_RARROW);
     }
 
     chtype enabled = tf->is_enabled() ? ACS_DIAMOND : ' ';
@@ -304,19 +352,22 @@ void filter_sub_source::text_attrs_for_line(textview_curses &tc, int line,
                                vcolors.ansi_to_theme_color(COLOR_GREEN));
     }
 
-    int fg = tf->get_type() == text_filter::INCLUDE ? COLOR_GREEN : COLOR_RED;
-    value_out.emplace_back(line_range{4, 7}, &view_curses::VC_FOREGROUND,
-                           vcolors.ansi_to_theme_color(fg));
+    int fg_role = tf->get_type() == text_filter::INCLUDE ?
+        view_colors::VCR_OK : view_colors::VCR_ERROR;
+    value_out.emplace_back(line_range{4, 7},
+                           &view_curses::VC_ROLE_FG,
+                           fg_role);
     value_out.emplace_back(line_range{4, 7}, &view_curses::VC_STYLE, A_BOLD);
 
-    value_out.emplace_back(line_range{8, 14}, &view_curses::VC_STYLE, A_BOLD);
-    value_out.emplace_back(line_range{20, 21}, &view_curses::VC_GRAPHIC, ACS_VLINE);
+    value_out.emplace_back(line_range{8, 17}, &view_curses::VC_STYLE, A_BOLD);
+    value_out.emplace_back(line_range{23, 24}, &view_curses::VC_GRAPHIC,
+                           ACS_VLINE);
 
-    fg = selected ? COLOR_BLACK : COLOR_WHITE;
-    value_out.emplace_back(line_range{0, -1}, &view_curses::VC_FOREGROUND,
-                           vcolors.ansi_to_theme_color(fg));
-    value_out.emplace_back(line_range{0, -1}, &view_curses::VC_BACKGROUND,
-                           vcolors.ansi_to_theme_color(bg));
+    if (selected) {
+        value_out.emplace_back(line_range{0, -1},
+                               &view_curses::VC_ROLE,
+                               view_colors::VCR_FOCUSED);
+    }
 }
 
 size_t filter_sub_source::text_size_for_line(textview_curses &tc, int line,
@@ -342,33 +393,75 @@ void filter_sub_source::rl_change(readline_curses *rc)
     auto iter = fs.begin() + this->tss_view->get_selection();
     shared_ptr<text_filter> tf = *iter;
     string new_value = rc->get_line_buffer();
-    auto_mem<pcre> code;
-    const char *errptr;
-    int eoff;
 
-    if ((code = pcre_compile(new_value.c_str(),
-                             PCRE_CASELESS,
-                             &errptr,
-                             &eoff,
-                             nullptr)) == nullptr) {
-        lnav_data.ld_filter_status_source.tss_error
-                 .set_value("error: %s", errptr);
-    } else {
-        textview_curses::highlight_map_t &hm = top_view->get_highlights();
-        highlighter hl(code.release());
-        int color;
+    switch (tf->get_lang()) {
+        case filter_lang_t::NONE:
+            break;
+        case filter_lang_t::REGEX: {
+            auto_mem<pcre> code;
+            const char *errptr;
+            int eoff;
 
-        if (tf->get_type() == text_filter::EXCLUDE) {
-            color = COLOR_RED;
-        } else {
-            color = COLOR_GREEN;
+            if ((code = pcre_compile(new_value.c_str(),
+                                     PCRE_CASELESS,
+                                     &errptr,
+                                     &eoff,
+                                     nullptr)) == nullptr) {
+                lnav_data.ld_filter_help_status_source.fss_error
+                    .set_value("error: %s", errptr);
+            } else {
+                auto &hm = top_view->get_highlights();
+                highlighter hl(code.release());
+                int color;
+
+                if (tf->get_type() == text_filter::EXCLUDE) {
+                    color = COLOR_RED;
+                } else {
+                    color = COLOR_GREEN;
+                }
+                hl.with_attrs(
+                    view_colors::ansi_color_pair(COLOR_BLACK, color) | A_BLINK);
+
+                hm[{highlight_source_t::PREVIEW, "preview"}] = hl;
+                top_view->set_needs_update();
+                lnav_data.ld_filter_help_status_source.fss_error.clear();
+            }
+            break;
         }
-        hl.with_attrs(
-            view_colors::ansi_color_pair(COLOR_BLACK, color) | A_BLINK);
+        case filter_lang_t::SQL: {
+            auto full_sql = fmt::format("SELECT 1 WHERE {}", new_value);
+            auto_mem<sqlite3_stmt> stmt(sqlite3_finalize);
+#ifdef SQLITE_PREPARE_PERSISTENT
+            auto retcode = sqlite3_prepare_v3(lnav_data.ld_db.in(),
+                                              full_sql.c_str(),
+                                              full_sql.size(),
+                                              SQLITE_PREPARE_PERSISTENT,
+                                              stmt.out(),
+                                              nullptr);
+#else
+            auto retcode = sqlite3_prepare_v2(lnav_data.ld_db.in(),
+                                                  full_sql.c_str(),
+                                                  full_sql.size(),
+                                                  stmt.out(),
+                                                  nullptr);
+#endif
+            if (retcode != SQLITE_OK) {
+                lnav_data.ld_filter_help_status_source.fss_error
+                    .set_value("error: %s", sqlite3_errmsg(lnav_data.ld_db));
+            } else {
+                auto set_res = lnav_data.ld_log_source
+                    .set_preview_sql_filter(stmt.release());
 
-        hm[{highlight_source_t::PREVIEW, "preview"}] = hl;
-        top_view->set_needs_update();
-        lnav_data.ld_filter_status_source.tss_error.clear();
+                if (set_res.isErr()) {
+                    lnav_data.ld_filter_help_status_source.fss_error
+                        .set_value("error: %s", set_res.unwrapErr().c_str());
+                } else {
+                    top_view->set_needs_update();
+                    lnav_data.ld_filter_help_status_source.fss_error.clear();
+                }
+            }
+            break;
+        }
     }
 }
 
@@ -380,30 +473,68 @@ void filter_sub_source::rl_perform(readline_curses *rc)
     auto iter = fs.begin() + this->tss_view->get_selection();
     shared_ptr<text_filter> tf = *iter;
     string new_value = rc->get_value();
-    auto_mem<pcre> code;
-    const char *errptr;
-    int eoff;
 
     if (new_value.empty()) {
         this->rl_abort(rc);
-    } else if ((code = pcre_compile(new_value.c_str(),
-                                    PCRE_CASELESS,
-                                    &errptr,
-                                    &eoff,
-                                    nullptr)) == nullptr) {
-        this->rl_abort(rc);
     } else {
-        tf->lf_deleted = true;
-        tss->text_filters_changed();
+        top_view->get_highlights().erase({highlight_source_t::PREVIEW, "preview"});
+        switch (tf->get_lang()) {
+            case filter_lang_t::NONE:
+            case filter_lang_t::REGEX: {
+                auto_mem<pcre> code;
+                const char *errptr;
+                int eoff;
 
-        auto pf = make_shared<pcre_filter>(tf->get_type(),
-                                           new_value, tf->get_index(), code.release());
+                if ((code = pcre_compile(new_value.c_str(),
+                                         PCRE_CASELESS,
+                                         &errptr,
+                                         &eoff,
+                                         nullptr)) == nullptr) {
+                    this->rl_abort(rc);
+                } else {
+                    tf->lf_deleted = true;
+                    tss->text_filters_changed();
 
-        *iter = pf;
-        tss->text_filters_changed();
+                    auto pf = make_shared<pcre_filter>(tf->get_type(),
+                                                       new_value, tf->get_index(),
+                                                       code.release());
+
+                    *iter = pf;
+                    tss->text_filters_changed();
+                }
+                break;
+            }
+            case filter_lang_t::SQL: {
+                auto full_sql = fmt::format("SELECT 1 WHERE {}", new_value);
+                auto_mem<sqlite3_stmt> stmt(sqlite3_finalize);
+#ifdef SQLITE_PREPARE_PERSISTENT
+                auto retcode = sqlite3_prepare_v3(lnav_data.ld_db.in(),
+                                                  full_sql.c_str(),
+                                                  full_sql.size(),
+                                                  SQLITE_PREPARE_PERSISTENT,
+                                                  stmt.out(),
+                                                  nullptr);
+#else
+                auto retcode = sqlite3_prepare_v2(lnav_data.ld_db.in(),
+                                                  full_sql.c_str(),
+                                                  full_sql.size(),
+                                                  stmt.out(),
+                                                  nullptr);
+#endif
+                if (retcode != SQLITE_OK) {
+                    this->rl_abort(rc);
+                } else {
+                    lnav_data.ld_log_source.set_sql_filter(
+                        new_value, stmt.release());
+                    tss->text_filters_changed();
+                }
+                break;
+            }
+        }
     }
 
-    lnav_data.ld_filter_status_source.tss_prompt.clear();
+    lnav_data.ld_log_source.set_preview_sql_filter(nullptr);
+    lnav_data.ld_filter_help_status_source.fss_prompt.clear();
     this->fss_editing = false;
     this->fss_editor.set_visible(false);
     this->tss_view->reload_data();
@@ -417,11 +548,13 @@ void filter_sub_source::rl_abort(readline_curses *rc)
     auto iter = fs.begin() + this->tss_view->get_selection();
     shared_ptr<text_filter> tf = *iter;
 
-    lnav_data.ld_filter_status_source.tss_prompt.clear();
-    lnav_data.ld_filter_status_source.tss_error.clear();
+    lnav_data.ld_log_source.set_preview_sql_filter(nullptr);
+    lnav_data.ld_filter_help_status_source.fss_prompt.clear();
+    lnav_data.ld_filter_help_status_source.fss_error.clear();
     top_view->get_highlights().erase({highlight_source_t::PREVIEW, "preview"});
     top_view->reload_data();
     fs.delete_filter("");
+    this->tss_view->reload_data();
     this->fss_editor.set_visible(false);
     this->fss_editing = false;
     this->tss_view->set_needs_update();
@@ -458,7 +591,8 @@ void filter_sub_source::rl_display_matches(readline_curses *rc)
 
         this->fss_match_view.set_selection(selected_line);
         this->fss_match_source.replace_with(al);
-        this->fss_match_view.set_height(std::min(vis_line_t(matches.size()), 3_vl));
+        this->fss_match_view.set_height(
+            std::min(vis_line_t(matches.size()), 3_vl));
     }
 
     this->fss_match_view.set_window(this->tss_view->get_window());
@@ -476,8 +610,7 @@ void filter_sub_source::rl_display_next(readline_curses *rc)
 
     if (tc.get_top() >= (tc.get_top_for_last_row() - 1)) {
         tc.set_top(0_vl);
-    }
-    else {
+    } else {
         tc.shift_top(tc.get_height());
     }
 }
