@@ -30,14 +30,12 @@
 #include "config.h"
 
 #include <string.h>
-#include <stdlib.h>
-#include <unistd.h>
 #include <sys/stat.h>
 #include <glob.h>
 #include <grp.h>
 #include <pwd.h>
 
-#include "auto_mem.hh"
+#include "base/injector.hh"
 #include "base/lnav_log.hh"
 #include "sql_util.hh"
 #include "vtab_module.hh"
@@ -68,6 +66,7 @@ enum {
 };
 
 struct fstat_table {
+    static constexpr const char *NAME = "fstat";
     static constexpr const char *CREATE_STMT = R"(
 CREATE TABLE fstat (
     st_parent TEXT,
@@ -92,14 +91,6 @@ CREATE TABLE fstat (
 );
 )";
 
-    struct vtab {
-        sqlite3_vtab base;
-
-        operator sqlite3_vtab *() {
-            return &this->base;
-        };
-    };
-
     struct cursor {
         sqlite3_vtab_cursor base;
         string c_pattern;
