@@ -29,13 +29,11 @@
 
 #include "config.h"
 
-#include <numeric>
 #include <algorithm>
+#include <regex>
 
-#include <pcrecpp.h>
-
+#include "base/string_util.hh"
 #include "fmt/format.h"
-#include "lnav_util.hh"
 #include "ansi_scrubber.hh"
 #include "help_text_formatter.hh"
 #include "readline_highlighters.hh"
@@ -176,16 +174,52 @@ void format_help_text_for_term(const help_text &ht, size_t width, attr_line_t &o
                 .append("\n");
             break;
         }
+        case help_context_t::HC_SQL_COMMAND: {
+            out.append("Synopsis", &view_curses::VC_STYLE, A_UNDERLINE)
+                .append("\n")
+                .append(body_indent, ' ')
+                .append(";")
+                .append(ht.ht_name, &view_curses::VC_STYLE, A_BOLD);
+            for (auto &param : ht.ht_parameters) {
+                out.append(" ");
+                if (param.ht_nargs == help_nargs_t::HN_OPTIONAL) {
+                    out.append("[");
+                }
+                out.append(param.ht_name, &view_curses::VC_STYLE, A_UNDERLINE);
+                if (param.ht_nargs == help_nargs_t::HN_OPTIONAL) {
+                    out.append("]");
+                }
+                if (param.ht_nargs == help_nargs_t::HN_ONE_OR_MORE) {
+                    out.append("1", &view_curses::VC_STYLE, A_UNDERLINE);
+                    out.append(" [");
+                    out.append("...", &view_curses::VC_STYLE, A_UNDERLINE);
+                    out.append(" ");
+                    out.append(param.ht_name, &view_curses::VC_STYLE,
+                               A_UNDERLINE);
+                    out.append("N", &view_curses::VC_STYLE, A_UNDERLINE);
+                    out.append("]");
+                }
+            }
+            out.append(" - ")
+                .append(attr_line_t::from_ansi_str(ht.ht_summary),
+                        &tws.with_indent(body_indent + 2))
+                .append("\n");
+            break;
+        }
+        case help_context_t::HC_SQL_INFIX:
         case help_context_t::HC_SQL_KEYWORD: {
             size_t line_start = body_indent;
             bool break_all = false;
+            bool is_infix = ht.ht_context == help_context_t::HC_SQL_INFIX;
 
             if (!synopsis_only) {
                 out.append("Synopsis", &view_curses::VC_STYLE, A_UNDERLINE)
                    .append("\n");
             }
             out.append(body_indent, ' ')
-               .append(ht.ht_name, &view_curses::VC_STYLE, A_BOLD);
+               .append(ht.ht_name,
+                       &view_curses::VC_STYLE,
+                       is_infix ? 0 : A_BOLD);
             for (auto &param : ht.ht_parameters) {
                 if (break_all ||
                     (int)(out.get_string().length() - start_index - line_start + 10) >=
@@ -424,6 +458,7 @@ void format_example_text_for_term(const help_text &ht,
                     ex_line.insert(1, ht.ht_name);
                     readline_command_highlighter(ex_line, 0);
                     break;
+                case help_context_t::HC_SQL_INFIX:
                 case help_context_t::HC_SQL_KEYWORD:
                 case help_context_t::HC_SQL_FUNCTION:
                 case help_context_t::HC_SQL_TABLE_VALUED_FUNCTION:
@@ -437,7 +472,7 @@ void format_example_text_for_term(const help_text &ht,
             out.append("#")
                .append(to_string(count))
                .append(" ")
-               .append(ex.he_description)
+               .append(ex.he_description, &tws.with_indent(3))
                .append(":\n   ")
                .append(prompt)
                .append(ex_line, &tws.with_indent(3 + keyword_offset + 1))
@@ -453,18 +488,31 @@ void format_example_text_for_term(const help_text &ht,
 
 static std::string link_name(const help_text &ht)
 {
-    const static pcrecpp::RE SCRUBBER("[^\\w_]");
+    const static std::regex SCRUBBER("[^\\w_]");
 
-    auto scrubbed_name = string(ht.ht_name);
+    bool is_sql_infix = ht.ht_context == help_context_t::HC_SQL_INFIX;
+    string scrubbed_name;
+
+    if (is_sql_infix) {
+        scrubbed_name = "infix";
+    } else {
+        scrubbed_name = ht.ht_name;
+    }
+    if (ht.ht_function_type == help_function_type_t::HFT_AGGREGATE) {
+        scrubbed_name += "_agg";
+    }
     for (auto &param : ht.ht_parameters) {
-        if (param.ht_name[0]) {
+        if (!is_sql_infix && param.ht_name[0]) {
+            continue;
+        }
+        if (!param.ht_flag_name) {
             continue;
         }
 
         scrubbed_name += "_";
         scrubbed_name += param.ht_flag_name;
     }
-    SCRUBBER.GlobalReplace("_", &scrubbed_name);
+    scrubbed_name = std::regex_replace(scrubbed_name, SCRUBBER, "_");
 
     return tolower(scrubbed_name);
 }
@@ -485,11 +533,15 @@ void format_help_text_for_rst(const help_text &ht,
         case help_context_t::HC_COMMAND:
             prefix = ":";
             break;
+        case help_context_t::HC_SQL_COMMAND:
+            prefix = ";";
+            break;
         case help_context_t::HC_SQL_FUNCTION:
         case help_context_t::HC_SQL_TABLE_VALUED_FUNCTION:
             is_sql = is_sql_func = true;
             prefix = "";
             break;
+        case help_context_t::HC_SQL_INFIX:
         case help_context_t::HC_SQL_KEYWORD:
             is_sql = true;
             prefix = "";
@@ -521,7 +573,15 @@ void format_help_text_for_rst(const help_text &ht,
             out_count += fprintf(rst_file, "%s ", param.ht_flag_name);
         }
         if (param.ht_name[0]) {
-            out_count += fprintf(rst_file, "*%s*", param.ht_name);
+            out_count += fprintf(rst_file, "*");
+            if (param.ht_nargs == help_nargs_t::HN_OPTIONAL) {
+                out_count += fprintf(rst_file, "\\[");
+            }
+            out_count += fprintf(rst_file, "%s", param.ht_name);
+            if (param.ht_nargs == help_nargs_t::HN_OPTIONAL) {
+                out_count += fprintf(rst_file, "\\]");
+            }
+            out_count += fprintf(rst_file, "*");
         }
         if (is_sql_func) {
             needs_comma = true;
@@ -547,11 +607,12 @@ void format_help_text_for_rst(const help_text &ht,
     }
 
     if (param_count > 0) {
-        fprintf(rst_file, "  **Parameters:**\n\n");
+        fprintf(rst_file, "  **Parameters**\n");
         for (auto &param: ht.ht_parameters) {
             if (param.ht_summary && param.ht_summary[0]) {
-                fprintf(rst_file, "    * **%s** --- %s\n",
+                fprintf(rst_file, "    * **%s%s** --- %s\n",
                         param.ht_name,
+                        param.ht_nargs == help_nargs_t::HN_REQUIRED ? "\\*" : "",
                         param.ht_summary);
             }
         }
@@ -561,7 +622,7 @@ void format_help_text_for_rst(const help_text &ht,
         prefix = ";";
     }
     if (!ht.ht_example.empty()) {
-        fprintf(rst_file, "  **Examples:**\n\n");
+        fprintf(rst_file, "  **Examples**\n");
         for (auto &example: ht.ht_example) {
             fprintf(rst_file, "    %s:\n\n", example.he_description);
             fprintf(rst_file, "    .. code-block::  %s\n\n",
@@ -593,8 +654,8 @@ void format_help_text_for_rst(const help_text &ht,
         }
         stable_sort(related_refs.begin(), related_refs.end());
 
-        fprintf(rst_file, "  **See Also:**\n\n    %s\n",
-            join(related_refs.begin(), related_refs.end(), ", ").c_str());
+        fmt::print(rst_file, "  **See Also**\n    {}\n",
+                   fmt::join(related_refs, ", "));
     }
 
     fprintf(rst_file, "\n----\n\n");
