@@ -29,6 +29,8 @@
 
 #include "config.h"
 
+#include "base/injector.hh"
+#include "base/math_util.hh"
 #include "lnav.hh"
 #include "bookmarks.hh"
 #include "sql_util.hh"
@@ -37,17 +39,17 @@
 #include "pretty_printer.hh"
 #include "sysclip.hh"
 #include "log_data_helper.hh"
-#include "session_data.hh"
 #include "command_executor.hh"
 #include "termios_guard.hh"
-#include "readline_callbacks.hh"
-#include "readline_possibilities.hh"
 #include "readline_highlighters.hh"
 #include "field_overlay_source.hh"
 #include "hotkeys.hh"
-#include "log_format_loader.hh"
 #include "base/opt_util.hh"
 #include "shlex.hh"
+#include "lnav_util.hh"
+#include "lnav_config.hh"
+#include "bound_tags.hh"
+#include "xterm_mouse.hh"
 
 using namespace std;
 
@@ -62,7 +64,7 @@ public:
         content_line_t cl = this->lh_sub_source.at(this->lh_current_line);
         std::shared_ptr<logfile> lf = this->lh_sub_source.find(cl);
         auto ll = lf->begin() + cl;
-        while (ll->is_continued()) {
+        while (!ll->is_message()) {
             --ll;
             --this->lh_current_line;
         }
@@ -83,7 +85,7 @@ public:
         content_line_t cl = this->lh_sub_source.at(this->lh_current_line);
         std::shared_ptr<logfile> lf = this->lh_sub_source.find(cl);
         auto ll = lf->begin() + cl;
-        log_format *format = lf->get_format();
+        auto format = lf->get_format();
         lf->read_full_message(ll, this->lh_msg_buffer);
         format->annotate(cl,
                          this->lh_msg_buffer,
@@ -160,7 +162,7 @@ bool handle_keyseq(const char *keyseq)
         string expanded_msg;
 
         if (lexer.eval(expanded_msg, {
-            &ec.ec_local_vars.top(),
+            &vars,
             &ec.ec_global_vars,
         })) {
             lnav_data.ld_rl_view->set_alt_value(expanded_msg);
@@ -172,7 +174,7 @@ bool handle_keyseq(const char *keyseq)
 
 bool handle_paging_key(int ch)
 {
-    if (lnav_data.ld_view_stack.vs_views.empty()) {
+    if (lnav_data.ld_view_stack.empty()) {
         return false;
     }
 
@@ -226,9 +228,11 @@ bool handle_paging_key(int ch)
 
                 lnav_data.ld_last_view = nullptr;
                 if (src_view != nullptr && dst_view != nullptr) {
-                    struct timeval top_time = src_view->time_for_row(top_tc->get_top());
-
-                    tc->set_top(vis_line_t(dst_view->row_for_time(top_time)));
+                    src_view->time_for_row(top_tc->get_top()) | [dst_view, tc](auto top_time) {
+                        dst_view->row_for_time(top_time) | [tc](auto row) {
+                            tc->set_top(row);
+                        };
+                    };
                 }
                 ensure_view(tc);
             }
@@ -236,10 +240,11 @@ bool handle_paging_key(int ch)
 
         case KEY_F(2):
             if (xterm_mouse::is_available()) {
-                lnav_data.ld_mouse.set_enabled(!lnav_data.ld_mouse.is_enabled());
+                auto& mouse_i = injector::get<xterm_mouse&>();
+                mouse_i.set_enabled(!mouse_i.is_enabled());
                 lnav_data.ld_rl_view->set_value(
                     ok_prefix("info: mouse mode -- ") +
-                    (lnav_data.ld_mouse.is_enabled() ?
+                    (mouse_i.is_enabled() ?
                      ANSI_BOLD("enabled") : ANSI_BOLD("disabled")));
             }
             else {
@@ -347,53 +352,6 @@ bool handle_paging_key(int ch)
             }
             break;
 
-        case 'u': {
-            vis_line_t user_top, meta_top;
-
-            lnav_data.ld_rl_view->set_alt_value(
-                    HELP_MSG_1(c, "to copy marked lines to the clipboard; ")
-            HELP_MSG_1(C, "to clear marked lines"));
-
-            user_top = next_cluster(&bookmark_vector<vis_line_t>::next,
-                                    &textview_curses::BM_USER,
-                                    tc->get_top());
-            meta_top = next_cluster(&bookmark_vector<vis_line_t>::next,
-                                    &textview_curses::BM_META,
-                                    tc->get_top());
-            if (user_top == -1 && meta_top == -1) {
-                alerter::singleton().chime();
-            }
-            else {
-                if (user_top == -1) {
-                    user_top = vis_line_t(INT_MAX);
-                }
-                if (meta_top == -1) {
-                    meta_top = vis_line_t(INT_MAX);
-                }
-
-                tc->set_top(min(user_top, meta_top));
-            }
-            break;
-        }
-
-        case 'U': {
-            vis_line_t user_top, meta_top;
-
-            user_top = next_cluster(&bookmark_vector<vis_line_t>::prev,
-                                    &textview_curses::BM_USER,
-                                    tc->get_top());
-            meta_top = next_cluster(&bookmark_vector<vis_line_t>::prev,
-                                    &textview_curses::BM_META,
-                                    tc->get_top());
-            if (user_top == -1 && meta_top == -1) {
-                alerter::singleton().chime();
-            }
-            else {
-                tc->set_top(max(user_top, meta_top));
-            }
-            break;
-        }
-
         case 'J':
             if (lnav_data.ld_last_user_mark.find(tc) ==
                 lnav_data.ld_last_user_mark.end() ||
@@ -408,7 +366,7 @@ bool handle_paging_key(int ch)
                 tc->get_dimensions(height, width);
                 if (lnav_data.ld_last_user_mark[tc] > (tc->get_bottom() - 2) &&
                     tc->get_top() + height < tc->get_inner_height()) {
-                    tc->shift_top(vis_line_t(1));
+                    tc->shift_top(1_vl);
                 }
                 if (lnav_data.ld_last_user_mark[tc] + 1 >=
                     tc->get_inner_height()) {
@@ -441,7 +399,7 @@ bool handle_paging_key(int ch)
             tc->toggle_user_mark(&textview_curses::BM_USER,
                                  vis_line_t(new_mark));
             if (new_mark == tc->get_top()) {
-                tc->shift_top(vis_line_t(-1));
+                tc->shift_top(-1_vl);
             }
             if (new_mark > 0) {
                 lnav_data.ld_last_user_mark[tc] = new_mark - 1;
@@ -504,7 +462,7 @@ bool handle_paging_key(int ch)
                 }
                 lss->set_time_offset(true);
                 while (next_top < tc->get_inner_height()) {
-                    if (lss->find_line(lss->at(next_top))->is_continued()) {
+                    if (!lss->find_line(lss->at(next_top))->is_message()) {
                     }
                     else if (lss->get_line_accel_direction(next_top) ==
                              log_accel::A_DECEL) {
@@ -528,7 +486,7 @@ bool handle_paging_key(int ch)
                 }
                 lss->set_time_offset(true);
                 while (0 <= next_top && next_top < tc->get_inner_height()) {
-                    if (lss->find_line(lss->at(next_top))->is_continued()) {
+                    if (!lss->find_line(lss->at(next_top))->is_message()) {
                     }
                     else if (lss->get_line_accel_direction(next_top) ==
                              log_accel::A_DECEL) {
@@ -560,23 +518,26 @@ bool handle_paging_key(int ch)
 
         case '0':
             if (lss) {
-                struct timeval first_time = lss->time_for_row(tc->get_top());
-                int        step       = 24 * 60 * 60;
-                vis_line_t line       =
-                        lss->find_from_time(roundup_size(first_time.tv_sec, step));
-
-                tc->set_top(line);
+                const int step = 24 * 60 * 60;
+                lss->time_for_row(tc->get_top()) | [lss, tc](auto first_time) {
+                    lss->find_from_time(roundup_size(first_time.tv_sec, step)) | [tc](auto line) {
+                        tc->set_top(line);
+                    };
+                };
             }
             break;
 
         case ')':
             if (lss) {
-                struct timeval first_time = lss->time_for_row(tc->get_top());
-                time_t     day  = rounddown(first_time.tv_sec, 24 * 60 * 60);
-                vis_line_t line = lss->find_from_time(day);
-
-                --line;
-                tc->set_top(line);
+                lss->time_for_row(tc->get_top()) | [lss, tc](auto first_time) {
+                    time_t day = rounddown(first_time.tv_sec, 24 * 60 * 60);
+                    lss->find_from_time(day) | [tc](auto line) {
+                        if (line != 0_vl) {
+                            --line;
+                        }
+                        tc->set_top(line);
+                    };
+                };
             }
             break;
 
@@ -585,15 +546,16 @@ bool handle_paging_key(int ch)
                 alerter::singleton().chime();
             }
             else if (lss) {
-                struct timeval first_time = lss->time_for_row(tc->get_top());
-                int        step     = ch == 'D' ? (24 * 60 * 60) : (60 * 60);
-                time_t     top_time = first_time.tv_sec;
-                vis_line_t line     = lss->find_from_time(top_time - step);
-
-                if (line != 0) {
-                    --line;
-                }
-                tc->set_top(line);
+                lss->time_for_row(tc->get_top()) | [lss, ch, tc](auto first_time) {
+                    int step = ch == 'D' ? (24 * 60 * 60) : (60 * 60);
+                    time_t     top_time = first_time.tv_sec;
+                    lss->find_from_time(top_time - step) | [tc](auto line) {
+                        if (line != 0_vl) {
+                            --line;
+                        }
+                        tc->set_top(line);
+                    };
+                };
 
                 lnav_data.ld_rl_view->set_alt_value(HELP_MSG_1(/, "to search"));
             }
@@ -601,12 +563,12 @@ bool handle_paging_key(int ch)
 
         case 'd':
             if (lss) {
-                struct timeval first_time = lss->time_for_row(tc->get_top());
-                int        step = ch == 'd' ? (24 * 60 * 60) : (60 * 60);
-                vis_line_t line =
-                        lss->find_from_time(first_time.tv_sec + step);
-
-                tc->set_top(line);
+                lss->time_for_row(tc->get_top()) | [ch, lss, tc](auto first_time) {
+                    int step = ch == 'd' ? (24 * 60 * 60) : (60 * 60);
+                    lss->find_from_time(first_time.tv_sec + step) | [tc](auto line) {
+                        tc->set_top(line);
+                    };
+                };
 
                 lnav_data.ld_rl_view->set_alt_value(HELP_MSG_1(/, "to search"));
             }
@@ -646,7 +608,7 @@ bool handle_paging_key(int ch)
                             }
                         }
                         logline &next_line = next_helper.current_line();
-                        if (next_line.is_continued()) {
+                        if (!next_line.is_message()) {
                             continue;
                         }
                         if (next_line.get_opid() != opid_hash) {
@@ -679,183 +641,12 @@ bool handle_paging_key(int ch)
             }
             break;
 
-        case ':':
-            if (lnav_data.ld_views[LNV_LOG].get_inner_height() > 0) {
-                static const char *MOVE_TIMES[] = {
-                        "here",
-                        "now",
-                        "today",
-                        "yesterday",
-                        nullptr
-                };
-
-                logfile_sub_source &lss      = lnav_data.ld_log_source;
-                textview_curses &   log_view = lnav_data.ld_views[LNV_LOG];
-                content_line_t      cl       = lss.at(log_view.get_top());
-                std::shared_ptr<logfile>           lf       = lss.find(cl);
-                auto ll = lf->begin() + cl;
-                log_data_helper ldh(lss);
-
-                lnav_data.ld_exec_context.ec_top_line = tc->get_top();
-
-                lnav_data.ld_rl_view->clear_possibilities(LNM_COMMAND, "numeric-colname");
-                lnav_data.ld_rl_view->clear_possibilities(LNM_COMMAND, "colname");
-
-                ldh.parse_line(log_view.get_top(), true);
-
-                if (tc == &lnav_data.ld_views[LNV_DB]) {
-                    db_label_source &dls = lnav_data.ld_db_row_source;
-
-                    for (auto &dls_header : dls.dls_headers) {
-                        if (!dls_header.hm_graphable) {
-                            continue;
-                        }
-
-                        lnav_data.ld_rl_view->add_possibility(LNM_COMMAND,
-                                                              "numeric-colname",
-                                                              dls_header.hm_name);
-                    }
-                }
-                else {
-                    for (auto &ldh_line_value : ldh.ldh_line_values) {
-                        const logline_value_stats *stats = ldh_line_value.lv_format->stats_for_value(
-                            ldh_line_value.lv_name);
-
-                        if (stats == nullptr) {
-                            continue;
-                        }
-
-                        lnav_data.ld_rl_view->add_possibility(LNM_COMMAND,
-                                                              "numeric-colname",
-                                                              ldh_line_value.lv_name.to_string());
-                    }
-                }
-
-                for (auto &cn_name : ldh.ldh_namer->cn_names) {
-                    lnav_data.ld_rl_view->add_possibility(LNM_COMMAND, "colname",
-                                                          cn_name);
-                }
-                for (auto iter : ldh.ldh_namer->cn_builtin_names) {
-                    if (iter == "col") {
-                        continue;
-                    }
-                    lnav_data.ld_rl_view->add_possibility(LNM_COMMAND, "colname", iter);
-                }
-
-                ldh.clear();
-
-                readline_curses *rlc = lnav_data.ld_rl_view;
-
-                rlc->clear_possibilities(LNM_COMMAND, "move-time");
-                rlc->add_possibility(LNM_COMMAND, "move-time", MOVE_TIMES);
-                rlc->clear_possibilities(LNM_COMMAND, "line-time");
-                {
-                    struct timeval tv = lf->get_time_offset();
-                    char buffer[64];
-
-                    sql_strftime(buffer, sizeof(buffer),
-                                 ll->get_time(), ll->get_millis(), 'T');
-                    rlc->add_possibility(LNM_COMMAND,
-                                         "line-time",
-                                         buffer);
-                    rlc->add_possibility(LNM_COMMAND,
-                                         "move-time",
-                                         buffer);
-                    sql_strftime(buffer, sizeof(buffer),
-                                 ll->get_time() - tv.tv_sec,
-                                 ll->get_millis() - (tv.tv_usec / 1000),
-                                 'T');
-                    rlc->add_possibility(LNM_COMMAND,
-                                         "line-time",
-                                         buffer);
-                    rlc->add_possibility(LNM_COMMAND,
-                                         "move-time",
-                                         buffer);
-                }
-            }
-
-            rollback_lnav_config = lnav_config;
-            lnav_data.ld_doc_status_source.set_title("Command Help");
-            add_view_text_possibilities(lnav_data.ld_rl_view, LNM_COMMAND, "filter", tc);
-            lnav_data.ld_rl_view->add_possibility(LNM_COMMAND, "filter", tc->get_last_search());
-            add_filter_possibilities(tc);
-            add_mark_possibilities();
-            add_config_possibilities();
-            add_env_possibilities(LNM_COMMAND);
-            add_tag_possibilities();
-            lnav_data.ld_mode = LNM_COMMAND;
-            lnav_data.ld_rl_view->focus(LNM_COMMAND, ":");
-            break;
-
-        case '/':
-            lnav_data.ld_mode = LNM_SEARCH;
-            lnav_data.ld_previous_search = tc->get_last_search();
-            lnav_data.ld_search_start_line = tc->get_top();
-            add_view_text_possibilities(lnav_data.ld_rl_view, LNM_SEARCH, "*", tc);
-            lnav_data.ld_rl_view->focus(LNM_SEARCH, "/");
-            lnav_data.ld_doc_status_source.set_title("Syntax Help");
-            rl_set_help();
-            lnav_data.ld_bottom_source.set_prompt(
-                    "Enter a regular expression to search for: "
-                            "(Press " ANSI_BOLD("CTRL+]") " to abort)");
-            break;
-
-        case ';':
-            if (tc == &lnav_data.ld_views[LNV_LOG] ||
-                tc == &lnav_data.ld_views[LNV_DB] ||
-                tc == &lnav_data.ld_views[LNV_SCHEMA]) {
-                textview_curses &log_view = lnav_data.ld_views[LNV_LOG];
-
-                lnav_data.ld_exec_context.ec_top_line = tc->get_top();
-
-                lnav_data.ld_mode = LNM_SQL;
-                setup_logline_table(lnav_data.ld_exec_context);
-                lnav_data.ld_rl_view->focus(LNM_SQL, ";");
-
-                lnav_data.ld_doc_status_source.set_title("Query Help");
-                rl_set_help();
-                lnav_data.ld_bottom_source.update_loading(0, 0);
-                lnav_data.ld_status[LNS_BOTTOM].do_update();
-
-                field_overlay_source *fos;
-
-                fos = (field_overlay_source *) log_view.get_overlay_source();
-                fos->fos_active_prev = fos->fos_active;
-                if (!fos->fos_active) {
-                    fos->fos_active = true;
-                    tc->reload_data();
-                }
-                lnav_data.ld_bottom_source.set_prompt(
-                    "Enter an SQL query: (Press " ANSI_BOLD("CTRL+]") " to abort)");
-            }
-            break;
-
-        case '|': {
-            map<string, vector<script_metadata> > &scripts = lnav_data.ld_scripts;
-
-            lnav_data.ld_mode = LNM_EXEC;
-
-            lnav_data.ld_exec_context.ec_top_line = tc->get_top();
-            lnav_data.ld_rl_view->clear_possibilities(LNM_EXEC, "__command");
-            find_format_scripts(lnav_data.ld_config_paths, scripts);
-            for (const auto &iter : scripts) {
-                lnav_data.ld_rl_view->add_possibility(LNM_EXEC, "__command", iter.first);
-            }
-            add_view_text_possibilities(lnav_data.ld_rl_view, LNM_EXEC, "*", tc);
-            add_env_possibilities(LNM_EXEC);
-            lnav_data.ld_rl_view->focus(LNM_EXEC, "|");
-            lnav_data.ld_bottom_source.set_prompt(
-                    "Enter a script to execute: (Press "
-                            ANSI_BOLD("CTRL+]") " to abort)");
-            break;
-        }
-
         case 'p':
             if (tc == &lnav_data.ld_views[LNV_LOG]) {
                 field_overlay_source *fos;
 
                 fos = (field_overlay_source *) tc->get_overlay_source();
-                fos->fos_active = !fos->fos_active;
+                fos->fos_contexts.top().c_show = !fos->fos_contexts.top().c_show;
                 tc->reload_data();
             }
             else if (tc == &lnav_data.ld_views[LNV_DB]) {
@@ -890,16 +681,17 @@ bool handle_paging_key(int ch)
 
         case 'I':
         {
-            auto &hs = lnav_data.ld_hist_source2;
             auto &hist_tc = lnav_data.ld_views[LNV_HISTOGRAM];
 
             if (toggle_view(&hist_tc)) {
                 auto *src_view = dynamic_cast<text_time_translator *>(tc->get_sub_source());
 
                 if (src_view != nullptr) {
-                    struct timeval log_top = src_view->time_for_row(tc->get_top());
-
-                    hist_tc.set_top(vis_line_t(hs.row_for_time(log_top)));
+                    src_view->time_for_row(tc->get_top()) | [](auto log_top) {
+                        lnav_data.ld_hist_source2.row_for_time(log_top) | [](auto row) {
+                            lnav_data.ld_views[LNV_HISTOGRAM].set_top(row);
+                        };
+                    };
                 }
             }
             else {
@@ -907,12 +699,16 @@ bool handle_paging_key(int ch)
                     auto *dst_view = dynamic_cast<text_time_translator *>(top_tc->get_sub_source());
 
                     if (dst_view != nullptr) {
-                        struct timeval hist_top_time = hs.time_for_row(hist_tc.get_top());
-                        struct timeval curr_top_time = dst_view->time_for_row(top_tc->get_top());
-                        if (hs.row_for_time(hist_top_time) != hs.row_for_time(curr_top_time)) {
-                            vis_line_t new_top = vis_line_t(dst_view->row_for_time(hist_top_time));
-                            top_tc->set_top(new_top);
-                            top_tc->set_needs_update();
+                        auto& hs = lnav_data.ld_hist_source2;
+                        auto hist_top_time_opt = hs.time_for_row(hist_tc.get_top());
+                        auto curr_top_time_opt = dst_view->time_for_row(top_tc->get_top());
+                        if (hist_top_time_opt && curr_top_time_opt &&
+                            hs.row_for_time(hist_top_time_opt.value()) !=
+                            hs.row_for_time(curr_top_time_opt.value())) {
+                            dst_view->row_for_time(hist_top_time_opt.value()) | [top_tc](auto new_top) {
+                                top_tc->set_top(new_top);
+                                top_tc->set_needs_update();
+                            };
                         }
                     }
                 };
@@ -979,11 +775,10 @@ bool handle_paging_key(int ch)
                         size_t col_len = strlen(col_value);
 
                         if (dts.scan(col_value, col_len, nullptr, &tm, tv) != nullptr) {
-                            vis_line_t vl;
-
-                            vl = lnav_data.ld_log_source.find_from_time(tv);
-                            tc->set_top(vl);
-                            tc->set_needs_update();
+                            lnav_data.ld_log_source.find_from_time(tv) | [tc](auto vl) {
+                                tc->set_top(vl);
+                                tc->set_needs_update();
+                            };
                             break;
                         }
                     }
@@ -1021,8 +816,33 @@ bool handle_paging_key(int ch)
 
                 tc->reload_data();
             } else if (tc_tss != nullptr && tc_tss->tss_supports_filtering) {
-                lnav_data.ld_mode = LNM_FILTER;
+                lnav_data.ld_mode = lnav_data.ld_last_config_mode;
                 lnav_data.ld_filter_view.reload_data();
+                lnav_data.ld_files_view.reload_data();
+                if (tc->get_inner_height() > 0_vl) {
+                    string_attrs_t::const_iterator line_attr;
+                    std::vector<attr_line_t> rows(1);
+
+                    tc->get_data_source()->
+                        listview_value_for_rows(*tc, tc->get_top(), rows);
+                    string_attrs_t &sa = rows[0].get_attrs();
+                    line_attr = find_string_attr(sa, &logline::L_FILE);
+                    if (line_attr != sa.end()) {
+                        const auto &fc = lnav_data.ld_active_files;
+                        auto lf = ((logfile *)line_attr->sa_value.sav_ptr)->
+                            shared_from_this();
+                        auto iter = find(fc.fc_files.begin(),
+                                         fc.fc_files.end(),
+                                         lf);
+                        if (iter != fc.fc_files.end()) {
+                            auto index = distance(fc.fc_files.begin(), iter);
+                            auto index_vl = vis_line_t(index);
+
+                            lnav_data.ld_files_view.set_top(index_vl);
+                            lnav_data.ld_files_view.set_selection(index_vl);
+                        }
+                    }
+                }
             } else {
                 alerter::singleton().chime();
             }
@@ -1040,13 +860,16 @@ bool handle_paging_key(int ch)
         case 'r':
         case 'R':
             if (lss) {
-                if (lnav_data.ld_last_relative_time.empty()) {
+                auto& last_time =
+                    injector::get<const relative_time&, last_relative_time_tag>();
+
+                if (last_time.empty()) {
                     lnav_data.ld_rl_view->set_value(
                             "Use the 'goto' command to set the relative time to move by");
                 }
                 else {
                     vis_line_t vl = tc->get_top(), new_vl;
-                    relative_time rt = lnav_data.ld_last_relative_time;
+                    relative_time rt = last_time;
                     content_line_t cl;
                     struct exttm tm;
                     bool done = false;
@@ -1065,8 +888,12 @@ bool handle_paging_key(int ch)
                     logline *ll = lnav_data.ld_log_source.find_line(cl);
                     ll->to_exttm(tm);
                     do {
-                        rt.add(tm);
-                        new_vl = lnav_data.ld_log_source.find_from_time(tm);
+                        tm = rt.adjust(tm);
+                        auto new_vl_opt = lnav_data.ld_log_source.find_from_time(tm);
+                        if (!new_vl_opt) {
+                            break;
+                        }
+                        new_vl = new_vl_opt.value();
 
                         if (new_vl == 0_vl || new_vl != vl || !rt.is_relative()) {
                             vl = new_vl;
