@@ -31,6 +31,7 @@
 
 #include "config.h"
 
+#include <stdio.h>
 #include <string.h>
 #include <sys/time.h>
 #include <array>
@@ -50,11 +51,14 @@
 #endif
 
 #include "base/lnav_log.hh"
+#include "base/time_util.hh"
+#include "ww898/cp_utf8.hpp"
 #include "input_dispatcher.hh"
-#include "lnav_util.hh"
+
+using namespace ww898;
 
 template<typename A>
-static void to_key_seq(A &dst, char *src)
+static void to_key_seq(A &dst, const char *src)
 {
     dst[0] = '\0';
     for (size_t lpc = 0; src[lpc]; lpc++) {
@@ -82,6 +86,7 @@ void input_dispatcher::new_input(const struct timeval &current_time, int ch)
                 this->append_to_escape_buffer(ch);
 
                 if (strcmp("\x1b[", this->id_escape_buffer) == 0) {
+                } else if (strcmp("\x1b[<", this->id_escape_buffer) == 0) {
                     this->id_mouse_handler();
                     this->id_escape_index = 0;
                 } else if (this->id_escape_expected_size == -1 ||
@@ -90,13 +95,9 @@ void input_dispatcher::new_input(const struct timeval &current_time, int ch)
                     to_key_seq(keyseq, this->id_escape_buffer);
                     switch (this->id_escape_matcher(keyseq.data())) {
                         case escape_match_t::NONE: {
-                            if (this->id_escape_expected_size == -1) {
-                                for (int lpc = 0; this->id_escape_buffer[lpc]; lpc++) {
-                                    handled = this->id_key_handler(
-                                        this->id_escape_buffer[lpc]);
-                                }
-                            } else {
-                                handled = false;
+                            for (int lpc = 0; this->id_escape_buffer[lpc]; lpc++) {
+                                handled = this->id_key_handler(
+                                    this->id_escape_buffer[lpc]);
                             }
                             this->id_escape_index = 0;
                             break;
@@ -113,15 +114,18 @@ void input_dispatcher::new_input(const struct timeval &current_time, int ch)
                     this->id_escape_index == this->id_escape_expected_size) {
                     this->id_escape_index = 0;
                 }
-            } else if ((ch & 0xf8) == 0xf0) {
-                this->reset_escape_buffer(ch, current_time, 4);
-            } else if ((ch & 0xf0) == 0xe0) {
-                this->reset_escape_buffer(ch, current_time, 3);
-            } else if ((ch & 0xe0) == 0xc0) {
-                this->reset_escape_buffer(ch, current_time, 2);
             } else {
-                snprintf(keyseq.data(), keyseq.size(), "x%02x", ch & 0xff);
-                handled = this->id_key_handler(ch);
+                auto seq_size = utf::utf8::char_size([ch]() {
+                    return std::make_pair(ch, 16);
+                })
+                    .unwrapOr(size_t{1});
+
+                if (seq_size == 1) {
+                    snprintf(keyseq.data(), keyseq.size(), "x%02x", ch & 0xff);
+                    handled = this->id_key_handler(ch);
+                } else {
+                    this->reset_escape_buffer(ch, current_time, seq_size);
+                }
             }
             break;
     }
