@@ -30,6 +30,7 @@
 #include "config.h"
 
 #include "lnav.hh"
+#include "lnav_util.hh"
 #include "log_actions.hh"
 
 using namespace std;
@@ -41,7 +42,7 @@ static string execute_action(log_data_helper &ldh,
     std::map<string, log_format::action_def>::const_iterator iter;
     logline_value &lv = ldh.ldh_line_values[value_index];
     shared_ptr<logfile> lf = ldh.ldh_file;
-    const log_format *format = lf->get_format();
+    const auto format = lf->get_format();
     pid_t child_pid;
     string retval;
 
@@ -81,7 +82,7 @@ static string execute_action(log_data_helper &ldh,
             setenv("LNAV_ACTION_FILE_LINE", env_buffer, 1);
             snprintf(env_buffer, sizeof(env_buffer), "%d", ldh.ldh_y_offset + 1);
             setenv("LNAV_ACTION_MSG_LINE", env_buffer, 1);
-            setenv("LNAV_ACTION_VALUE_NAME", lv.lv_name.get(), 1);
+            setenv("LNAV_ACTION_VALUE_NAME", lv.lv_meta.lvm_name.get(), 1);
             value_line = ldh.ldh_y_offset - ldh.get_value_line(lv) + 1;
             snprintf(env_buffer, sizeof(env_buffer), "%d", value_line);
             setenv("LNAV_ACTION_VALUE_LINE", env_buffer, 1);
@@ -97,7 +98,7 @@ static string execute_action(log_data_helper &ldh,
             for (size_t lpc = 0; lpc < action.ad_cmdline.size(); lpc++) {
                 args[lpc] = action.ad_cmdline[lpc].c_str();
             }
-            args[action.ad_cmdline.size()] = NULL;
+            args[action.ad_cmdline.size()] = nullptr;
             execvp(args[0], (char *const *) args);
             fprintf(stderr,
                     "error: could not exec process -- %s:%s\n",
@@ -122,8 +123,13 @@ static string execute_action(log_data_helper &ldh,
                 auto pp = make_shared<piper_proc>(
                     out_pipe.read_end(),
                     false,
-                    open_temp_file(system_tmpdir() / "lnav.action.XXXXXX")
-                        .then([](auto pair) { pair.first.remove_file(); })
+                    open_temp_file(ghc::filesystem::temp_directory_path() /
+                                   "lnav.action.XXXXXX")
+                        .map([](auto pair) {
+                            ghc::filesystem::remove(pair.first);
+
+                            return pair;
+                        })
                         .expect("Cannot create temporary file for action")
                         .second);
                 char desc[128];
@@ -133,9 +139,9 @@ static string execute_action(log_data_helper &ldh,
                          sizeof(desc), "[%d] Output of %s",
                          exec_count++,
                          action.ad_cmdline[0].c_str());
-                lnav_data.ld_file_names[desc]
+                lnav_data.ld_active_files.fc_file_names[desc]
                     .with_fd(pp->get_fd());
-                lnav_data.ld_files_to_front.push_back({ desc, 0 });
+                lnav_data.ld_files_to_front.emplace_back( desc, 0 );
             }
 
             return "";
@@ -150,7 +156,7 @@ bool action_delegate::text_handle_mouse(textview_curses &tc, mouse_event &me)
 {
     bool retval = false;
 
-    if (me.me_button != BUTTON_LEFT) {
+    if (me.me_button != mouse_button_t::BUTTON_LEFT) {
         return false;
     }
 
@@ -158,7 +164,7 @@ bool action_delegate::text_handle_mouse(textview_curses &tc, mouse_event &me)
     int mouse_left = tc.get_left() + me.me_x;
 
     switch (me.me_state) {
-        case BUTTON_STATE_PRESSED:
+        case mouse_button_state_t::BUTTON_STATE_PRESSED:
             if (mouse_line >= vis_line_t(0) && mouse_line <= tc.get_bottom()) {
                 size_t line_end_index = 0;
                 int x_offset;
@@ -187,7 +193,7 @@ bool action_delegate::text_handle_mouse(textview_curses &tc, mouse_event &me)
                 }
             }
             break;
-        case BUTTON_STATE_DRAGGED:
+        case mouse_button_state_t::BUTTON_STATE_DRAGGED:
             if (mouse_line != this->ad_press_line) {
                 this->ad_press_value = -1;
             }
@@ -195,7 +201,7 @@ bool action_delegate::text_handle_mouse(textview_curses &tc, mouse_event &me)
                 retval = true;
             }
             break;
-        case BUTTON_STATE_RELEASED:
+        case mouse_button_state_t::BUTTON_STATE_RELEASED:
             if (this->ad_press_value != -1 && this->ad_press_line == mouse_line) {
                 logline_value &lv = this->ad_log_helper.ldh_line_values[this->ad_press_value];
                 int x_offset = this->ad_line_index + mouse_left;
@@ -205,7 +211,7 @@ bool action_delegate::text_handle_mouse(textview_curses &tc, mouse_event &me)
                     const vector<string> *actions;
 
                     actions = lf->get_format()->get_actions(lv);
-                    if (actions != NULL && !actions->empty()) {
+                    if (actions != nullptr && !actions->empty()) {
                         string rc = execute_action(
                             this->ad_log_helper, this->ad_press_value, actions->at(0));
 
