@@ -39,33 +39,36 @@
 
 #include <map>
 #include <string>
-#include <fstream>
 
 #include "fmt/format.h"
 
+#include "base/paths.hh"
+#include "base/string_util.hh"
 #include "yajlpp/yajlpp.hh"
 #include "yajlpp/yajlpp_def.hh"
 #include "lnav_config.hh"
-#include "log_format.hh"
+#include "log_format_ext.hh"
 #include "auto_fd.hh"
 #include "sql_util.hh"
+#include "lnav_util.hh"
 #include "builtin-scripts.h"
 #include "builtin-sh-scripts.h"
 #include "default-formats.h"
 
 #include "log_format_loader.hh"
-#include "bin2c.h"
+#include "bin2c.hh"
 
 using namespace std;
 
 static void extract_metadata(const char *contents, size_t len, struct script_metadata &meta_out);
 
-typedef map<intern_string_t, external_log_format *> log_formats_map_t;
+typedef map<intern_string_t, std::shared_ptr<external_log_format>> log_formats_map_t;
 
+static auto intern_lifetime = intern_string::get_table_lifetime();
 static log_formats_map_t LOG_FORMATS;
 
 struct userdata {
-    std::string ud_format_path;
+    ghc::filesystem::path ud_format_path;
     vector<intern_string_t> *ud_format_names{nullptr};
     std::vector<std::string> *ud_errors{nullptr};
 };
@@ -76,12 +79,13 @@ static external_log_format *ensure_format(const yajlpp_provider_context &ypc, us
     vector<intern_string_t> *formats = ud->ud_format_names;
     external_log_format *retval;
 
-    retval = LOG_FORMATS[name];
+    retval = LOG_FORMATS[name].get();
     if (retval == nullptr) {
-        LOG_FORMATS[name] = retval = new external_log_format(name);
+        LOG_FORMATS[name] = std::make_shared<external_log_format>(name);
+        retval = LOG_FORMATS[name].get();
         log_debug("Loading format -- %s", name.get());
     }
-    retval->elf_source_path.insert(ud->ud_format_path.substr(0, ud->ud_format_path.rfind('/')));
+    retval->elf_source_path.insert(ud->ud_format_path.filename().string());
 
     if (find(formats->begin(), formats->end(), name) == formats->end()) {
         formats->push_back(name);
@@ -118,8 +122,8 @@ static external_log_format::value_def *value_def_provider(const yajlpp_provider_
     shared_ptr<external_log_format::value_def> retval;
 
     if (iter == elf->elf_value_defs.end()) {
-        retval = make_shared<external_log_format::value_def>();
-        retval->vd_name = value_name;
+        retval = make_shared<external_log_format::value_def>(
+            value_name, value_kind_t::VALUE_TEXT, -1, elf);
         elf->elf_value_defs[value_name] = retval;
         elf->elf_value_def_order.emplace_back(retval);
     } else {
@@ -131,18 +135,9 @@ static external_log_format::value_def *value_def_provider(const yajlpp_provider_
 
 static scaling_factor *scaling_factor_provider(const yajlpp_provider_context &ypc, external_log_format::value_def *value_def)
 {
-    string scale_spec = ypc.get_substr(0);
-
-    const intern_string_t scale_name = intern_string::lookup(scale_spec.substr(1));
+    auto scale_name = ypc.get_substr_i(0);
     scaling_factor &retval = value_def->vd_unit_scaling[scale_name];
 
-    if (scale_spec[0] == '/') {
-        retval.sf_op = scale_op_t::SO_DIVIDE;
-    }
-    else if (scale_spec[0] == '*') {
-        retval.sf_op = scale_op_t::SO_MULTIPLY;
-    }
-
     return &retval;
 }
 
@@ -223,8 +218,10 @@ static int read_format_int(yajlpp_parse_context *ypc, long long val)
 static int read_format_field(yajlpp_parse_context *ypc, const unsigned char *str, size_t len)
 {
     auto elf = (external_log_format *) ypc->ypc_obj_stack.top();
-    string value = string((const char *)str, len);
-    string field_name = ypc->get_path_fragment(1);
+    auto leading_slash = len > 0 && str[0] == '/';
+    auto value = string((const char *) (leading_slash ? str + 1 : str),
+                        leading_slash ? len - 1 : len);
+    auto field_name = ypc->get_path_fragment(1);
 
     if (field_name == "file-pattern") {
         elf->elf_file_pattern = value;
@@ -353,18 +350,18 @@ static int create_search_table(yajlpp_parse_context *ypc, const unsigned char *s
 
 
 static struct json_path_container pattern_handlers = {
-    json_path_handler("pattern")
+    yajlpp::property_handler("pattern")
         .with_synopsis("<message-regex>")
         .with_description(
             "The regular expression to match a log message and capture fields.")
         .with_min_length(1)
         .FOR_FIELD(external_log_format::pattern, p_string),
-    json_path_handler("module-format")
+    yajlpp::property_handler("module-format")
         .with_synopsis("<bool>")
         .with_description(
             "If true, this pattern will only be used to parse message bodies "
                 "of container formats, like syslog")
-        .FOR_FIELD(external_log_format::pattern, p_module_format)
+        .for_field(&external_log_format::pattern::p_module_format)
 };
 
 static const json_path_handler_base::enum_value_t ALIGN_ENUM[] = {
@@ -392,48 +389,48 @@ static const json_path_handler_base::enum_value_t TRANSFORM_ENUM[] = {
 };
 
 static struct json_path_container line_format_handlers = {
-    json_path_handler("field")
+    yajlpp::property_handler("field")
         .with_synopsis("<field-name>")
         .with_description("The name of the field to substitute at this position")
         .with_min_length(1)
         .FOR_FIELD(external_log_format::json_format_element, jfe_value),
 
-    json_path_handler("default-value")
+    yajlpp::property_handler("default-value")
         .with_synopsis("<string>")
         .with_description("The default value for this position if the field is null")
         .FOR_FIELD(external_log_format::json_format_element, jfe_default_value),
 
-    json_path_handler("timestamp-format")
+    yajlpp::property_handler("timestamp-format")
         .with_synopsis("<string>")
         .with_min_length(1)
         .with_description("The strftime(3) format for this field")
         .FOR_FIELD(external_log_format::json_format_element, jfe_ts_format),
 
-    json_path_handler("min-width")
+    yajlpp::property_handler("min-width")
         .with_min_value(0)
         .with_synopsis("<size>")
         .with_description("The minimum width of the field")
         .FOR_FIELD(external_log_format::json_format_element, jfe_min_width),
 
-    json_path_handler("max-width")
+    yajlpp::property_handler("max-width")
         .with_min_value(0)
         .with_synopsis("<size>")
         .with_description("The maximum width of the field")
         .FOR_FIELD(external_log_format::json_format_element, jfe_max_width),
 
-    json_path_handler("align")
+    yajlpp::property_handler("align")
         .with_synopsis("left|right")
         .with_description("Align the text in the column to the left or right side")
         .with_enum_values(ALIGN_ENUM)
         .FOR_FIELD(external_log_format::json_format_element, jfe_align),
 
-    json_path_handler("overflow")
+    yajlpp::property_handler("overflow")
         .with_synopsis("abbrev|truncate|dot-dot")
         .with_description("Overflow style")
         .with_enum_values(OVERFLOW_ENUM)
         .FOR_FIELD(external_log_format::json_format_element, jfe_overflow),
 
-    json_path_handler("text-transform")
+    yajlpp::property_handler("text-transform")
         .with_synopsis("none|uppercase|lowercase|capitalize")
         .with_description("Text transformation")
         .with_enum_values(TRANSFORM_ENUM)
@@ -441,107 +438,127 @@ static struct json_path_container line_format_handlers = {
 };
 
 static const json_path_handler_base::enum_value_t KIND_ENUM[] = {
-    {"string", logline_value::VALUE_TEXT},
-    {"integer", logline_value::VALUE_INTEGER},
-    {"float", logline_value::VALUE_FLOAT},
-    {"boolean", logline_value::VALUE_BOOLEAN},
-    {"json", logline_value::VALUE_JSON},
-    {"struct", logline_value::VALUE_STRUCT},
-    {"quoted", logline_value::VALUE_QUOTED},
+    {"string", value_kind_t::VALUE_TEXT},
+    {"integer", value_kind_t::VALUE_INTEGER},
+    {"float", value_kind_t::VALUE_FLOAT},
+    {"boolean", value_kind_t::VALUE_BOOLEAN},
+    {"json", value_kind_t::VALUE_JSON},
+    {"struct", value_kind_t::VALUE_STRUCT},
+    {"quoted", value_kind_t::VALUE_QUOTED},
+    {"xml", value_kind_t::VALUE_XML},
 
     json_path_handler_base::ENUM_TERMINATOR
 };
 
-static struct json_path_container scale_handlers = {
-    json_path_handler(pcrepp("(?<scale>.*)"))
-        .with_synopsis("[*,/]<unit>")
+static const json_path_handler_base::enum_value_t SCALE_OP_ENUM[] = {
+    {"identity", scale_op_t::SO_IDENTITY},
+    {"multiply", scale_op_t::SO_MULTIPLY},
+    {"divide", scale_op_t::SO_DIVIDE},
+
+    json_path_handler_base::ENUM_TERMINATOR
+};
+
+static struct json_path_container scaling_factor_handlers = {
+    yajlpp::pattern_property_handler("op")
+        .with_enum_values(SCALE_OP_ENUM)
+        .FOR_FIELD(scaling_factor, sf_op),
+
+    yajlpp::pattern_property_handler("value")
         .FOR_FIELD(scaling_factor, sf_value)
 };
 
+static struct json_path_container scale_handlers = {
+    yajlpp::pattern_property_handler("(?<scale>[^/]+)")
+        .with_obj_provider(scaling_factor_provider)
+        .with_children(scaling_factor_handlers),
+};
+
 static struct json_path_container unit_handlers = {
-    json_path_handler("field")
+    yajlpp::property_handler("field")
         .with_synopsis("<field-name>")
         .with_description("The name of the field that contains the units for this field")
         .FOR_FIELD(external_log_format::value_def, vd_unit_field),
 
-    json_path_handler("scaling-factor")
+    yajlpp::property_handler("scaling-factor")
         .with_description("Transforms the numeric value by the given factor")
-        .with_obj_provider(scaling_factor_provider)
         .with_children(scale_handlers),
 };
 
 static struct json_path_container value_def_handlers = {
-    json_path_handler("kind")
+    yajlpp::property_handler("kind")
         .with_synopsis("string|integer|float|boolean|json|quoted")
         .with_description("The type of data in the field")
         .with_enum_values(KIND_ENUM)
-        .FOR_FIELD(external_log_format::value_def, vd_kind),
+        .for_field(&external_log_format::value_def::vd_meta,
+                   &logline_value_meta::lvm_kind),
 
-    json_path_handler("collate")
+    yajlpp::property_handler("collate")
         .with_synopsis("<function>")
         .with_description("The collating function to use for this column")
         .FOR_FIELD(external_log_format::value_def, vd_collate),
 
-    json_path_handler("unit")
+    yajlpp::property_handler("unit")
         .with_description("Unit definitions for this field")
         .with_children(unit_handlers),
 
-    json_path_handler("identifier")
+    yajlpp::property_handler("identifier")
         .with_synopsis("<bool>")
         .with_description("Indicates whether or not this field contains an identifier that should be highlighted")
-        .FOR_FIELD(external_log_format::value_def, vd_identifier),
+        .for_field(&external_log_format::value_def::vd_meta,
+                   &logline_value_meta::lvm_identifier),
 
-    json_path_handler("foreign-key")
+    yajlpp::property_handler("foreign-key")
         .with_synopsis("<bool>")
         .with_description("Indicates whether or not this field should be treated as a foreign key for row in another table")
-        .FOR_FIELD(external_log_format::value_def, vd_foreign_key),
+        .for_field(&external_log_format::value_def::vd_foreign_key),
 
-    json_path_handler("hidden")
+    yajlpp::property_handler("hidden")
         .with_synopsis("<bool>")
         .with_description("Indicates whether or not this field should be hidden")
-        .FOR_FIELD(external_log_format::value_def, vd_hidden),
+        .for_field(&external_log_format::value_def::vd_meta,
+                   &logline_value_meta::lvm_hidden),
 
-    json_path_handler("action-list#")
+    yajlpp::property_handler("action-list#")
         .with_synopsis("<string>")
         .with_description("Actions to execute when this field is clicked on")
         .FOR_FIELD(external_log_format::value_def, vd_action_list),
 
-    json_path_handler("rewriter")
+    yajlpp::property_handler("rewriter")
         .with_synopsis("<command>")
         .with_description("A command that will rewrite this field when pretty-printing")
         .FOR_FIELD(external_log_format::value_def, vd_rewriter),
 
-    json_path_handler("description")
+    yajlpp::property_handler("description")
         .with_synopsis("<string>")
         .with_description("A description of the field")
         .FOR_FIELD(external_log_format::value_def, vd_description)
 };
 
 static struct json_path_container highlighter_def_handlers = {
-    json_path_handler("pattern")
+    yajlpp::property_handler("pattern")
         .with_synopsis("<regex>")
         .with_description("A regular expression to highlight in logs of this format.")
         .FOR_FIELD(external_log_format::highlighter_def, hd_pattern),
 
-    json_path_handler("color")
+    yajlpp::property_handler("color")
         .with_synopsis("#<hex>|<name>")
         .with_description("The color to use when highlighting this pattern.")
         .FOR_FIELD(external_log_format::highlighter_def, hd_color),
 
-    json_path_handler("background-color")
+    yajlpp::property_handler("background-color")
         .with_synopsis("#<hex>|<name>")
         .with_description("The background color to use when highlighting this pattern.")
         .FOR_FIELD(external_log_format::highlighter_def, hd_background_color),
 
-    json_path_handler("underline")
+    yajlpp::property_handler("underline")
         .with_synopsis("<enabled>")
         .with_description("Highlight this pattern with an underline.")
-        .FOR_FIELD(external_log_format::highlighter_def, hd_underline),
+        .for_field(&external_log_format::highlighter_def::hd_underline),
 
-    json_path_handler("blink")
+    yajlpp::property_handler("blink")
         .with_synopsis("<enabled>")
         .with_description("Highlight this pattern by blinking.")
-        .FOR_FIELD(external_log_format::highlighter_def, hd_blink)
+        .for_field(&external_log_format::highlighter_def::hd_blink)
 };
 
 static const json_path_handler_base::enum_value_t LEVEL_ENUM[] = {
@@ -563,12 +580,12 @@ static const json_path_handler_base::enum_value_t LEVEL_ENUM[] = {
 };
 
 static struct json_path_container sample_handlers = {
-    json_path_handler("line")
+    yajlpp::property_handler("line")
         .with_synopsis("<log-line>")
         .with_description("A sample log line that should match a pattern in this format.")
         .FOR_FIELD(external_log_format::sample, s_line),
 
-    json_path_handler("level")
+    yajlpp::property_handler("level")
         .with_enum_values(LEVEL_ENUM)
         .with_description("The expected level for this sample log line.")
         .FOR_FIELD(external_log_format::sample, s_level)
@@ -583,15 +600,15 @@ static const json_path_handler_base::enum_value_t TYPE_ENUM[] = {
 };
 
 static struct json_path_container regex_handlers = {
-    json_path_handler(pcrepp(R"((?<pattern_name>[^/]+))"))
+    yajlpp::pattern_property_handler(R"((?<pattern_name>[^/]+))")
         .with_description("The set of patterns used to match log messages")
         .with_obj_provider(pattern_provider)
         .with_children(pattern_handlers),
 };
 
 static struct json_path_container level_handlers = {
-    json_path_handler(
-        pcrepp("(?<level>trace|debug[2345]?|info|stats|notice|warning|error|critical|fatal)"))
+    yajlpp::pattern_property_handler(
+        "(?<level>trace|debug[2345]?|info|stats|notice|warning|error|critical|fatal)")
         .add_cb(read_levels)
         .add_cb(read_level_int)
         .with_synopsis("<pattern|integer>")
@@ -600,14 +617,14 @@ static struct json_path_container level_handlers = {
 };
 
 static struct json_path_container value_handlers = {
-    json_path_handler(pcrepp("(?<value_name>[^/]+)"))
+    yajlpp::pattern_property_handler("(?<value_name>[^/]+)")
         .with_description("The set of values captured by the log message patterns")
         .with_obj_provider(value_def_provider)
         .with_children(value_def_handlers)
 };
 
 static struct json_path_container highlight_handlers = {
-    json_path_handler(pcrepp(R"((?<highlight_name>[^/]+))"))
+    yajlpp::pattern_property_handler(R"((?<highlight_name>[^/]+))")
         .with_description("The definition of a highlight")
         .with_obj_provider<external_log_format::highlighter_def, external_log_format>([](const yajlpp_provider_context &ypc, external_log_format *root) {
             return &(root->elf_highlighter_patterns[ypc.get_substr_i(0)]);
@@ -633,13 +650,13 @@ static struct json_path_container search_table_def_handlers = {
 };
 
 static struct json_path_container search_table_handlers = {
-    json_path_handler(pcrepp("\\w+"))
+    yajlpp::pattern_property_handler("\\w+")
         .with_description("The set of search tables to be automatically defined")
         .with_children(search_table_def_handlers)
 };
 
 struct json_path_container format_handlers = {
-    json_path_handler("regex")
+    yajlpp::property_handler("regex")
         .with_description("The set of regular expressions used to match log messages")
         .with_children(regex_handlers),
 
@@ -676,26 +693,26 @@ struct json_path_container format_handlers = {
         .with_description("The name of the module field in the log message pattern"),
     json_path_handler("opid-field", read_format_field)
         .with_description("The name of the operation-id field in the log message pattern"),
-    json_path_handler("ordered-by-time")
+    yajlpp::property_handler("ordered-by-time")
         .with_synopsis("<bool>")
         .with_description("Indicates that the order of messages in the file is time-based.")
-        .FOR_FIELD(log_format, lf_time_ordered),
-    json_path_handler("level")
+        .for_field(&log_format::lf_time_ordered),
+    yajlpp::property_handler("level")
         .with_description("The map of level names to patterns or integer values")
         .with_children(level_handlers),
 
-    json_path_handler("value")
+    yajlpp::property_handler("value")
         .with_description("The set of value definitions")
         .with_children(value_handlers),
 
-    json_path_handler("action")
+    yajlpp::property_handler("action")
         .with_children(action_handlers),
-    json_path_handler("sample#")
+    yajlpp::property_handler("sample#")
         .with_description("An array of sample log messages to be tested against the log message patterns")
         .with_obj_provider(sample_provider)
         .with_children(sample_handlers),
 
-    json_path_handler("line-format#")
+    yajlpp::property_handler("line-format#")
         .with_description("The display format for JSON-encoded log messages")
         .with_obj_provider(line_format_provider)
         .add_cb(read_json_constant)
@@ -704,21 +721,17 @@ struct json_path_container format_handlers = {
         .with_description("Search tables to automatically define for this log format")
         .with_children(search_table_handlers),
 
-    json_path_handler("highlights")
+    yajlpp::property_handler("highlights")
         .with_description("The set of highlight definitions")
         .with_children(highlight_handlers),
 
-    json_path_handler("file-type")
+    yajlpp::property_handler("file-type")
         .with_synopsis("text|json|csv")
         .with_description("The type of file that contains the log messages")
         .with_enum_values(TYPE_ENUM)
         .FOR_FIELD(external_log_format, elf_type)
 };
 
-static vector<string> SUPPORTED_FORMAT_SCHEMAS = {
-    "https://lnav.org/schemas/format-v1.schema.json",
-};
-
 static int read_id(yajlpp_parse_context *ypc, const unsigned char *str, size_t len)
 {
     auto file_id = string((const char *) str, len);
@@ -739,60 +752,66 @@ struct json_path_container root_format_handler = json_path_container {
         .with_synopsis("The URI of the schema for this file")
         .with_description("Specifies the type of this file"),
 
-    json_path_handler(pcrepp("(?<format_name>\\w+)"))
+    yajlpp::pattern_property_handler("(?<format_name>\\w+)")
         .with_description("The definition of a log file format.")
         .with_obj_provider(ensure_format)
         .with_children(format_handlers)
 }
-    .with_schema_id(SUPPORTED_FORMAT_SCHEMAS.back());
+    .with_schema_id(DEFAULT_FORMAT_SCHEMA);
 
 static void write_sample_file()
 {
-    for (int lpc = 0; lnav_format_json[lpc].bsf_name; lpc++) {
-        auto &bsf = lnav_format_json[lpc];
-        auto sample_path = dotlnav_path() /
-            fmt::format("formats/default/{}.sample", bsf.bsf_name);
+    for (const auto& bsf : lnav_format_json) {
+        auto sample_path = lnav::paths::dotlnav() /
+            fmt::format("formats/default/{}.sample", bsf.get_name());
+        auto sf = bsf.to_string_fragment();
         auto_fd sample_fd;
 
         if ((sample_fd = openp(sample_path,
                                O_WRONLY | O_TRUNC | O_CREAT,
                                0644)) == -1 ||
-            (write(sample_fd.get(), bsf.bsf_data, bsf.bsf_size) == -1)) {
-            perror("error: unable to write default format file");
+            (write(sample_fd.get(), sf.data(), sf.length()) == -1)) {
+            fprintf(stderr,
+                    "error:unable to write default format file: %s -- %s\n",
+                    sample_path.c_str(),
+                    strerror(errno));
         }
     }
 
-    for (int lpc = 0; lnav_sh_scripts[lpc].bsf_name; lpc++) {
-        struct bin_src_file &bsf = lnav_sh_scripts[lpc];
-        auto sh_path = dotlnav_path() / fmt::format("formats/default/{}", bsf.bsf_name);
+    for (const auto& bsf : lnav_sh_scripts) {
+        auto sh_path = lnav::paths::dotlnav() / fmt::format("formats/default/{}", bsf.get_name());
+        auto sf = bsf.to_string_fragment();
         auto_fd sh_fd;
 
         if ((sh_fd = openp(sh_path, O_WRONLY|O_TRUNC|O_CREAT, 0755)) == -1 ||
-            write(sh_fd.get(), bsf.bsf_data, strlen((const char *) bsf.bsf_data)) == -1) {
-            perror("error: unable to write default text file");
+            write(sh_fd.get(), sf.data(), sf.length()) == -1) {
+            fprintf(stderr,
+                    "error:unable to write default text file: %s -- %s\n",
+                    sh_path.c_str(),
+                    strerror(errno));
         }
     }
 
-    for (int lpc = 0; lnav_scripts[lpc].bsf_name; lpc++) {
+    for (const auto& bsf : lnav_scripts) {
         struct script_metadata meta;
-        struct bin_src_file &bsf = lnav_scripts[lpc];
-        const char *script_content = reinterpret_cast<const char *>(bsf.bsf_data);
+        auto sf = bsf.to_string_fragment();
         auto_fd script_fd;
         char path[2048];
-        size_t script_len;
         struct stat st;
 
-        script_len = strlen(script_content);
-        extract_metadata(script_content, script_len, meta);
+        extract_metadata(sf.data(), sf.length(), meta);
         snprintf(path, sizeof(path), "formats/default/%s.lnav", meta.sm_name.c_str());
-        auto script_path = dotlnav_path() / path;
-        if (statp(script_path, &st) == 0 && (size_t) st.st_size == script_len) {
+        auto script_path = lnav::paths::dotlnav() / path;
+        if (statp(script_path, &st) == 0 && st.st_size == sf.length()) {
             // Assume it's the right contents and move on...
             continue;
         }
         if ((script_fd = openp(script_path, O_WRONLY|O_TRUNC|O_CREAT, 0755)) == -1 ||
-            write(script_fd.get(), script_content, script_len) == -1) {
-            perror("error: unable to write default text file");
+            write(script_fd.get(), sf.data(), sf.length()) == -1) {
+            fprintf(stderr,
+                    "error:unable to write default text file: %s -- %s\n",
+                    script_path.c_str(),
+                    strerror(errno));
         }
     }
 }
@@ -810,7 +829,9 @@ static void format_error_reporter(const yajlpp_parse_context &ypc,
     }
 }
 
-std::vector<intern_string_t> load_format_file(const string &filename, std::vector<string> &errors)
+std::vector<intern_string_t>
+load_format_file(const ghc::filesystem::path &filename,
+                 std::vector<string> &errors)
 {
     std::vector<intern_string_t> retval;
     struct userdata ud;
@@ -823,14 +844,10 @@ std::vector<intern_string_t> load_format_file(const string &filename, std::vecto
     yajlpp_parse_context ypc(filename, &root_format_handler);
     ypc.ypc_userdata = &ud;
     ypc.with_obj(ud);
-    if ((fd = open(filename.c_str(), O_RDONLY)) == -1) {
-        char errmsg[1024];
-
-        snprintf(errmsg, sizeof(errmsg),
-                 "error:unable to open format file '%s' -- %s",
-                 filename.c_str(),
-                 strerror(errno));
-        errors.emplace_back(errmsg);
+    if ((fd = openp(filename, O_RDONLY)) == -1) {
+        errors.emplace_back(fmt::format(
+            "error: unable to open format file '{}' -- {}",
+            filename.string(), strerror(errno)));
     }
     else {
         auto_mem<yajl_handle_t> handle(yajl_free);
@@ -838,7 +855,7 @@ std::vector<intern_string_t> load_format_file(const string &filename, std::vecto
         off_t offset = 0;
         ssize_t rc = -1;
 
-        handle = yajl_alloc(&ypc.ypc_callbacks, NULL, &ypc);
+        handle = yajl_alloc(&ypc.ypc_callbacks, nullptr, &ypc);
         ypc.with_handle(handle)
             .with_error_reporter(format_error_reporter);
         yajl_config(handle, yajl_allow_comments, 1);
@@ -848,11 +865,10 @@ std::vector<intern_string_t> load_format_file(const string &filename, std::vecto
                 break;
             }
             else if (rc == -1) {
-                errors.push_back(
-                    "error:" +
-                    filename +
-                    ":unable to read file -- " +
-                    string(strerror(errno)));
+                errors.push_back(fmt::format(
+                    "error:{}:unable to read file -- {}",
+                    filename.string(),
+                    strerror(errno)));
                 break;
             }
             if (offset == 0 && (rc > 2) &&
@@ -873,13 +889,13 @@ std::vector<intern_string_t> load_format_file(const string &filename, std::vecto
     return retval;
 }
 
-static void load_from_path(const filesystem::path &path, std::vector<string> &errors)
+static void load_from_path(const ghc::filesystem::path &path, std::vector<string> &errors)
 {
     auto format_path = path / "formats/*/*.json";
     static_root_mem<glob_t, globfree> gl;
 
-    log_info("loading formats from path: %s", format_path.str().c_str());
-    if (glob(format_path.str().c_str(), 0, nullptr, gl.inout()) == 0) {
+    log_info("loading formats from path: %s", format_path.c_str());
+    if (glob(format_path.c_str(), 0, nullptr, gl.inout()) == 0) {
         for (int lpc = 0; lpc < (int)gl->gl_pathc; lpc++) {
             const char *base = basename(gl->gl_pathv[lpc]);
 
@@ -905,11 +921,11 @@ static void load_from_path(const filesystem::path &path, std::vector<string> &er
     }
 }
 
-void load_formats(const std::vector<filesystem::path> &extra_paths,
+void load_formats(const std::vector<ghc::filesystem::path> &extra_paths,
                   std::vector<std::string> &errors)
 {
-    auto default_source = dotlnav_path() / "default";
-    yajlpp_parse_context ypc_builtin(default_source.str(), &root_format_handler);
+    auto default_source = lnav::paths::dotlnav() / "default";
+    yajlpp_parse_context ypc_builtin(default_source.string(), &root_format_handler);
     std::vector<intern_string_t> retval;
     struct userdata ud;
     yajl_handle handle;
@@ -917,8 +933,7 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
     write_sample_file();
 
     log_debug("Loading default formats");
-    for (int lpc = 0; lnav_format_json[lpc].bsf_name; lpc++) {
-        auto &bsf = lnav_format_json[lpc];
+    for (const auto& bsf : lnav_format_json) {
         handle = yajl_alloc(&ypc_builtin.ypc_callbacks, nullptr, &ypc_builtin);
         ud.ud_format_names = &retval;
         ud.ud_errors = &errors;
@@ -928,11 +943,15 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
             .with_error_reporter(format_error_reporter)
             .ypc_userdata = &ud;
         yajl_config(handle, yajl_allow_comments, 1);
-        if (ypc_builtin.parse(bsf.bsf_data, bsf.bsf_size) != yajl_status_ok) {
-            errors.push_back("builtin: invalid json -- " +
-                             string((char *) yajl_get_error(handle, 1,
-                                                            bsf.bsf_data,
-                                                            bsf.bsf_size)));
+        auto sf = bsf.to_string_fragment();
+        if (ypc_builtin.parse(sf) != yajl_status_ok) {
+            unsigned char *msg = yajl_get_error(handle, 1,
+                                                (const unsigned char *) sf.data(),
+                                                sf.length());
+
+            errors.push_back(fmt::format(
+                FMT_STRING("builtin: invalid json -- {}"), msg));
+            yajl_free_error(handle, msg);
         }
         ypc_builtin.complete_parse();
         yajl_free(handle);
@@ -942,15 +961,11 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
         load_from_path(extra_path, errors);
     }
 
-    if (!errors.empty()) {
-        return;
-    }
-
     uint8_t mod_counter = 0;
 
-    vector<external_log_format *> alpha_ordered_formats;
+    vector<std::shared_ptr<external_log_format>> alpha_ordered_formats;
     for (auto iter = LOG_FORMATS.begin(); iter != LOG_FORMATS.end(); ++iter) {
-        external_log_format *elf = iter->second;
+        auto& elf = iter->second;
         elf->build(errors);
 
         if (elf->elf_has_module_format) {
@@ -963,7 +978,7 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
                 continue;
             }
 
-            external_log_format *check_elf = check_iter.second;
+            auto& check_elf = check_iter.second;
             if (elf->match_samples(check_elf->elf_samples)) {
                 log_warning("Format collision, format '%s' matches sample from '%s'",
                         elf->get_name().get(),
@@ -977,15 +992,18 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
         }
     }
 
-    vector<external_log_format *> &graph_ordered_formats =
-            external_log_format::GRAPH_ORDERED_FORMATS;
+    if (!errors.empty()) {
+        return;
+    }
+
+    auto& graph_ordered_formats = external_log_format::GRAPH_ORDERED_FORMATS;
 
     while (!alpha_ordered_formats.empty()) {
         vector<intern_string_t> popped_formats;
 
         for (auto iter = alpha_ordered_formats.begin();
              iter != alpha_ordered_formats.end();) {
-            external_log_format *elf = *iter;
+            auto elf = *iter;
             if (elf->elf_collision.empty()) {
                 iter = alpha_ordered_formats.erase(iter);
                 popped_formats.push_back(elf->get_name());
@@ -1000,7 +1018,7 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
             bool broke_cycle = false;
 
             log_warning("Detected a cycle...");
-            for (auto elf : alpha_ordered_formats) {
+            for (const auto& elf : alpha_ordered_formats) {
                 if (elf->elf_builtin_format) {
                     log_warning("  Skipping builtin format -- %s",
                                 elf->get_name().get());
@@ -1017,7 +1035,7 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
             }
         }
 
-        for (auto elf : alpha_ordered_formats) {
+        for (const auto& elf : alpha_ordered_formats) {
             for (auto & popped_format : popped_formats) {
                 elf->elf_collision.remove(popped_format);
             }
@@ -1029,38 +1047,41 @@ void load_formats(const std::vector<filesystem::path> &extra_paths,
         log_info("  %s", graph_ordered_format->get_name().get());
     }
 
-    vector<log_format *> &roots = log_format::get_root_formats();
-    roots.insert(roots.begin(), graph_ordered_formats.begin(), graph_ordered_formats.end());
+    auto &roots = log_format::get_root_formats();
+    auto iter = std::find_if(roots.begin(), roots.end(), [](const auto& elem) {
+        return elem->get_name() == "generic_log";
+    });
+    roots.insert(iter, graph_ordered_formats.begin(), graph_ordered_formats.end());
 }
 
-static void exec_sql_in_path(sqlite3 *db, const filesystem::path &path, std::vector<string> &errors)
+static void exec_sql_in_path(sqlite3 *db, const ghc::filesystem::path &path, std::vector<string> &errors)
 {
     auto format_path = path / "formats/*/*.sql";
     static_root_mem<glob_t, globfree> gl;
 
-    log_info("executing SQL files in path: %s", format_path.str().c_str());
-    if (glob(format_path.str().c_str(), 0, nullptr, gl.inout()) == 0) {
+    log_info("executing SQL files in path: %s", format_path.c_str());
+    if (glob(format_path.c_str(), 0, nullptr, gl.inout()) == 0) {
         for (int lpc = 0; lpc < (int)gl->gl_pathc; lpc++) {
-            string filename(gl->gl_pathv[lpc]);
-            string content;
+            auto filename = ghc::filesystem::path(gl->gl_pathv[lpc]);
+            auto read_res = read_file(filename);
 
-            if (read_file(filename, content)) {
+            if (read_res.isOk()) {
                 log_info("Executing SQL file: %s", filename.c_str());
+                auto content = read_res.unwrap();
+
                 sql_execute_script(db, filename.c_str(), content.c_str(), errors);
             }
             else {
-                errors.push_back(
-                    "error:unable to read file '" +
-                    filename +
-                    "' -- " +
-                    string(strerror(errno)));
+                errors.push_back(fmt::format(
+                    "error:unable to read file '{}' -- {}",
+                    filename.string(), read_res.unwrapErr()));
             }
         }
     }
 }
 
 void load_format_extra(sqlite3 *db,
-                       const std::vector<filesystem::path> &extra_paths,
+                       const std::vector<ghc::filesystem::path> &extra_paths,
                        std::vector<std::string> &errors)
 {
     for (const auto & extra_path : extra_paths) {
@@ -1102,10 +1123,10 @@ void extract_metadata_from_file(struct script_metadata &meta_inout)
     struct stat st;
 
     if (statp(meta_inout.sm_path, &st) == -1) {
-        log_warning("unable to open script -- %s", meta_inout.sm_path.str().c_str());
+        log_warning("unable to open script -- %s", meta_inout.sm_path.c_str());
     } else if (!S_ISREG(st.st_mode)) {
-        log_warning("not a regular file -- %s", meta_inout.sm_path.str().c_str());
-    } else if ((fp = fopen(meta_inout.sm_path.str().c_str(), "r")) != NULL) {
+        log_warning("not a regular file -- %s", meta_inout.sm_path.c_str());
+    } else if ((fp = fopen(meta_inout.sm_path.c_str(), "r")) != nullptr) {
         size_t len;
 
         len = fread(buffer, 1, sizeof(buffer), fp.in());
@@ -1113,14 +1134,14 @@ void extract_metadata_from_file(struct script_metadata &meta_inout)
     }
 }
 
-static void find_format_in_path(const filesystem::path &path,
-                                map<string, vector<script_metadata> > &scripts)
+static void find_format_in_path(const ghc::filesystem::path &path,
+                                available_scripts& scripts)
 {
     auto format_path = path / "formats/*/*.lnav";
     static_root_mem<glob_t, globfree> gl;
 
-    log_debug("Searching for script in path: %s", format_path.str().c_str());
-    if (glob(format_path.str().c_str(), 0, nullptr, gl.inout()) == 0) {
+    log_debug("Searching for script in path: %s", format_path.c_str());
+    if (glob(format_path.c_str(), 0, nullptr, gl.inout()) == 0) {
         for (int lpc = 0; lpc < (int)gl->gl_pathc; lpc++) {
             const char *filename = basename(gl->gl_pathv[lpc]);
             string script_name = string(filename, strlen(filename) - 5);
@@ -1129,15 +1150,15 @@ static void find_format_in_path(const filesystem::path &path,
             meta.sm_path = gl->gl_pathv[lpc];
             meta.sm_name = script_name;
             extract_metadata_from_file(meta);
-            scripts[script_name].push_back(meta);
+            scripts.as_scripts[script_name].push_back(meta);
 
-            log_debug("  found script: %s", meta.sm_path.str().c_str());
+            log_debug("  found script: %s", meta.sm_path.c_str());
         }
     }
 }
 
-void find_format_scripts(const vector<filesystem::path> &extra_paths,
-                         map<string, vector<script_metadata> > &scripts)
+void find_format_scripts(const vector<ghc::filesystem::path> &extra_paths,
+                         available_scripts& scripts)
 {
     for (const auto &extra_path : extra_paths) {
         find_format_in_path(extra_path, scripts);
@@ -1147,7 +1168,7 @@ void find_format_scripts(const vector<filesystem::path> &extra_paths,
 void load_format_vtabs(log_vtab_manager *vtab_manager,
                        std::vector<std::string> &errors)
 {
-    map<intern_string_t, external_log_format *> &root_formats = LOG_FORMATS;
+    auto &root_formats = LOG_FORMATS;
 
     for (auto & root_format : root_formats) {
         root_format.second->register_vtabs(vtab_manager, errors);
