@@ -32,12 +32,10 @@
 #include "config.h"
 
 #include <errno.h>
-#include <stdarg.h>
 #include <string.h>
 #include <signal.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <sys/types.h>
 #include <sys/wait.h>
 #include <sys/socket.h>
 
@@ -54,9 +52,11 @@
 #endif
 
 #include <string>
+#include <utility>
 
-#include "lnav_config.hh"
-#include "pcrepp/pcrepp.hh"
+#include "base/paths.hh"
+#include "base/string_util.hh"
+#include "fmt/format.h"
 #include "shlex.hh"
 #include "auto_mem.hh"
 #include "base/lnav_log.hh"
@@ -69,6 +69,7 @@
 using namespace std;
 
 static int              got_line    = 0;
+static bool             alt_done    = 0;
 static sig_atomic_t     got_timeout = 0;
 static sig_atomic_t     got_winch   = 0;
 static readline_curses *child_this;
@@ -89,14 +90,14 @@ static const char *RL_INIT[] = {
     "set menu-complete-display-prefix on",
     "TAB: menu-complete",
     "\"\\e[Z\": menu-complete-backward",
-
-    NULL
 };
 
 readline_context *readline_context::loaded_context;
 set<string> *     readline_context::arg_possibilities;
 static string last_match_str;
 static bool last_match_str_valid;
+static bool arg_needs_shlex;
+static nonstd::optional<std::string> rewrite_line_start;
 
 static void sigalrm(int sig)
 {
@@ -219,15 +220,25 @@ static ssize_t recvstring(int sock, char *buf, size_t len)
     return retval;
 }
 
-char *readline_context::completion_generator(const char *text, int state)
+char *readline_context::completion_generator(const char *text_in, int state)
 {
     static vector<string> matches;
-    vector<string> long_matches;
 
+    vector<string> long_matches;
     char *retval = nullptr;
 
     if (state == 0) {
-        auto len = strlen(text);
+        auto text_str = std::string(text_in);
+
+        if (arg_needs_shlex) {
+            shlex arg_lexer(text_str);
+            map<string, string> scope;
+            std::string result;
+
+            if (arg_lexer.eval(result, scope)) {
+                text_str = result;
+            }
+        }
 
         matches.clear();
         if (arg_possibilities != nullptr) {
@@ -237,16 +248,16 @@ char *readline_context::completion_generator(const char *text, int state)
                 auto poss_str = poss.c_str();
 
                 // Check for an exact match and for the quoted version.
-                if (cmpfunc(text, poss_str, len) == 0 ||
+                if (cmpfunc(text_str.c_str(), poss_str, text_str.length()) == 0 ||
                     ((strchr(loaded_context->rc_quote_chars, poss_str[0]) !=
                         nullptr) &&
-                     cmpfunc(text, &poss_str[1], len) == 0)) {
+                     cmpfunc(text_str.c_str(), &poss_str[1], text_str.length()) == 0)) {
                     auto poss_slash_count = std::count(poss.begin(), poss.end(), '/');
 
-                    if (endswith(poss.c_str(), "/")) {
+                    if (endswith(poss, "/")) {
                         poss_slash_count -= 1;
                     }
-                    if (std::count(&text[0], &text[len], '/') == poss_slash_count) {
+                    if (std::count(text_str.begin(), text_str.end(), '/') == poss_slash_count) {
                         matches.emplace_back(poss);
                     } else {
                         long_matches.emplace_back(poss);
@@ -266,16 +277,16 @@ char *readline_context::completion_generator(const char *text, int state)
                     string poss_str = tolower(poss);
                     int score;
 
-                    if (fts::fuzzy_match(text, poss_str.c_str(), score) && score > 0) {
+                    if (fts::fuzzy_match(text_str.c_str(), poss_str.c_str(), score) && score > 0) {
                         if (score <= 0) {
                             continue;
                         }
 
                         auto poss_slash_count = std::count(poss_str.begin(), poss_str.end(), '/');
-                        if (endswith(poss.c_str(), "/")) {
+                        if (endswith(poss, "/")) {
                             poss_slash_count -= 1;
                         }
-                        if (std::count(&text[0], &text[len], '/') == poss_slash_count) {
+                        if (std::count(text_str.begin(), text_str.end(), '/') == poss_slash_count) {
                             fuzzy_matches.emplace_back(score, poss);
                         } else {
                             fuzzy_long_matches.emplace_back(score, poss);
@@ -305,7 +316,7 @@ char *readline_context::completion_generator(const char *text, int state)
         }
 
         if (matches.size() == 1) {
-            if (strcmp(text, matches[0].c_str()) == 0) {
+            if (text_str == matches[0]) {
                 matches.pop_back();
             }
 
@@ -336,6 +347,7 @@ char **readline_context::attempted_completion(const char *text,
     if (start == 0 && loaded_context->rc_possibilities.find("__command") !=
             loaded_context->rc_possibilities.end()) {
         arg_possibilities = &loaded_context->rc_possibilities["__command"];
+        arg_needs_shlex = false;
         rl_completion_append_character = loaded_context->rc_append_character;
     }
     else {
@@ -352,11 +364,12 @@ char **readline_context::attempted_completion(const char *text,
         arg_possibilities = nullptr;
         rl_completion_append_character = 0;
         if (lexer.split(prefix, scope)) {
-            string prefix2 = join(prefix.begin(), prefix.end(), "\x1f");
+            auto prefix2 = fmt::format("{}", fmt::join(prefix, "\x1f"));
             auto prefix_iter = loaded_context->rc_prefixes.find(prefix2);
 
             if (prefix_iter != loaded_context->rc_prefixes.end()) {
                 arg_possibilities = &(loaded_context->rc_possibilities[prefix_iter->second]);
+                arg_needs_shlex = false;
             }
         }
 
@@ -373,6 +386,7 @@ char **readline_context::attempted_completion(const char *text,
                 if (loaded_context->rc_possibilities.find("*") !=
                     loaded_context->rc_possibilities.end()) {
                     arg_possibilities = &loaded_context->rc_possibilities["*"];
+                    arg_needs_shlex = false;
                     rl_completion_append_character = loaded_context->rc_append_character;
                 }
             } else {
@@ -381,16 +395,72 @@ char **readline_context::attempted_completion(const char *text,
                 if (proto.empty()) {
                     arg_possibilities = nullptr;
                 } else if (proto[0] == "filename") {
-                    return nullptr; /* XXX */
+                    shlex fn_lexer(rl_line_buffer, rl_point);
+                    std::vector<std::string> fn_list;
+                    int found = 0;
+
+                    fn_lexer.split(fn_list, scope);
+
+                    const auto& last_fn =
+                        fn_list.size() <= 1 ? "" : fn_list.back();
+
+                    if (last_fn.find(':') != string::npos) {
+                        auto rp_iter = loaded_context->rc_possibilities.find("remote-path");
+                        if (rp_iter != loaded_context->rc_possibilities.end()) {
+                            for (const auto& poss : rp_iter->second) {
+                                if (startswith(poss, last_fn.c_str())) {
+                                    found += 1;
+                                }
+                            }
+                            if (found) {
+                                arg_possibilities = &rp_iter->second;
+                                arg_needs_shlex = false;
+                            }
+                        }
+                        if (!found ||
+                            (endswith(last_fn, "/") && found == 1)) {
+                            char msg[2048];
+
+                            snprintf(msg, sizeof(msg),
+                                     "\t:%s",
+                                     last_fn.c_str());
+                            sendstring(child_this->rc_command_pipe[1],
+                                       msg, strlen(msg));
+                        }
+                    }
+                    if (!found) {
+                        static std::set<std::string> file_name_set;
+
+                        file_name_set.clear();
+                        auto_mem<char> completed_fn;
+                        int fn_state = 0;
+                        auto recent_netlocs_iter = loaded_context->
+                            rc_possibilities.find("recent-netlocs");
+
+                        if (recent_netlocs_iter !=
+                            loaded_context->rc_possibilities.end()) {
+                            file_name_set.insert(
+                                recent_netlocs_iter->second.begin(),
+                                recent_netlocs_iter->second.end());
+                        }
+                        while ((completed_fn = rl_filename_completion_function(
+                            last_fn.c_str(), fn_state)) != nullptr) {
+                            file_name_set.insert(completed_fn.in());
+                            fn_state += 1;
+                        }
+                        arg_possibilities = &file_name_set;
+                        arg_needs_shlex = true;
+                    }
                 } else {
                     arg_possibilities = &(loaded_context->rc_possibilities[proto[0]]);
+                    arg_needs_shlex = false;
                 }
             }
         }
     }
 
     retval = rl_completion_matches(text, completion_generator);
-    if (retval == NULL) {
+    if (retval == nullptr) {
         rl_attempted_completion_over = 1;
     }
 
@@ -407,6 +477,13 @@ static int rubout_char_or_abort(int count, int key)
     }
 }
 
+static int alt_done_func(int count, int key)
+{
+    alt_done = true;
+    rl_newline(count, key);
+    return 0;
+}
+
 int readline_context::command_complete(int count, int key)
 {
     if (loaded_context->rc_possibilities.find("__command") !=
@@ -420,10 +497,10 @@ int readline_context::command_complete(int count, int key)
     return rl_insert(count, key);
 }
 
-readline_context::readline_context(const std::string &name,
+readline_context::readline_context(std::string name,
                                    readline_context::command_map_t *commands,
                                    bool case_sensitive)
-    : rc_name(name),
+    : rc_name(std::move(name)),
       rc_case_sensitive(case_sensitive),
       rc_quote_chars("\"'"),
       rc_highlighter(nullptr)
@@ -442,38 +519,77 @@ readline_context::readline_context(const std::string &name,
     memset(&this->rc_history, 0, sizeof(this->rc_history));
     history_set_history_state(&this->rc_history);
 
-    auto config_dir = dotlnav_path();
-    auto hpath = (config_dir / this->rc_name) + ".history";
-    read_history(hpath.str().c_str());
+    auto config_dir = lnav::paths::dotlnav();
+    auto hpath = (config_dir / this->rc_name).string() + ".history";
+    read_history(hpath.c_str());
     this->save();
 
     this->rc_append_character = ' ';
 }
 
+void readline_context::load()
+{
+    char buffer[128];
+
+    rl_completer_word_break_characters = (char *)" \t\n|()"; /* XXX */
+    /*
+     * XXX Need to keep the input on a single line since the display screws
+     * up if it wraps around.
+     */
+    snprintf(buffer, sizeof(buffer),
+             "set completion-ignore-case %s",
+             this->rc_case_sensitive ? "off" : "on");
+    rl_parse_and_bind(buffer); /* NOTE: buffer is modified */
+
+    loaded_context = this;
+    rl_attempted_completion_function = attempted_completion;
+    history_set_history_state(&this->rc_history);
+    for (auto &rc_var : this->rc_vars) {
+        *(rc_var.rv_dst.ch) = (char *) rc_var.rv_val.ch;
+    }
+}
+
+void readline_context::save()
+{
+    HISTORY_STATE *hs = history_get_history_state();
+
+    this->rc_history = *hs;
+    free(hs);
+    hs = nullptr;
+}
+
 readline_curses::readline_curses()
-    : rc_active_context(-1),
-      rc_child(-1),
-      rc_value_expiration(0),
-      rc_matches_remaining(0),
-      rc_max_match_length(0)
+    : rc_focus(noop_func{}),
+      rc_change(noop_func{}),
+      rc_perform(noop_func{}),
+      rc_alt_perform(noop_func{}),
+      rc_timeout(noop_func{}),
+      rc_abort(noop_func{}),
+      rc_display_match(noop_func{}),
+      rc_display_next(noop_func{}),
+      rc_blur(noop_func{}),
+      rc_completion_request(noop_func{})
 {
 }
 
 readline_curses::~readline_curses()
 {
+    this->rc_pty[RCF_MASTER].reset();
+    this->rc_command_pipe[RCF_MASTER].reset();
     if (this->rc_child == 0) {
         _exit(0);
     }
     else if (this->rc_child > 0) {
         int status;
 
-        log_debug("term child %d", this->rc_child);
-        kill(this->rc_child, SIGTERM);
+        log_debug("terminating readline child %d", this->rc_child);
+        log_perror(kill(this->rc_child, SIGTERM));
         this->rc_child = -1;
 
         while (wait(&status) < 0 && (errno == EINTR)) {
             ;
         }
+        log_debug("  child %d has exited", this->rc_child);
     }
 }
 
@@ -540,8 +656,8 @@ void readline_curses::start()
 
     if (openpty(this->rc_pty[RCF_MASTER].out(),
                 this->rc_pty[RCF_SLAVE].out(),
-                NULL,
-                NULL,
+                nullptr,
+                nullptr,
                 &ws) < 0) {
         perror("error: failed to open terminal(openpty)");
         throw error(errno);
@@ -578,10 +694,11 @@ void readline_curses::start()
         stifle_history(HISTORY_SIZE);
 
         rl_add_defun("rubout-char-or-abort", rubout_char_or_abort, '\b');
+        rl_add_defun("alt-done", alt_done_func, '\x0a');
         // rl_add_defun("command-complete", readline_context::command_complete, ' ');
 
-        for (int lpc = 0; RL_INIT[lpc]; lpc++) {
-            snprintf(buffer, sizeof(buffer), "%s", RL_INIT[lpc]);
+        for (const auto* init_cmd : RL_INIT) {
+            snprintf(buffer, sizeof(buffer), "%s", init_cmd);
             rl_parse_and_bind(buffer); /* NOTE: buffer is modified */
         }
 
@@ -626,7 +743,7 @@ void readline_curses::start()
                 itv.it_value.tv_usec    = KEY_TIMEOUT;
                 itv.it_interval.tv_sec  = 0;
                 itv.it_interval.tv_usec = 0;
-                setitimer(ITIMER_REAL, &itv, NULL);
+                setitimer(ITIMER_REAL, &itv, nullptr);
 
                 rl_callback_read_char();
                 if (RL_ISSTATE(RL_STATE_DONE) && !got_line) {
@@ -637,8 +754,6 @@ void readline_curses::start()
                 else {
                     uint64_t h1 = 1, h2 = 2;
 
-                    SpookyHash::Hash128(rl_line_buffer, rl_end, &h1, &h2);
-
                     if (rl_last_func == readline_context::command_complete) {
                         rl_last_func = rl_menu_complete;
                     }
@@ -649,7 +764,31 @@ void readline_curses::start()
 
                     if (complete_done) {
                         last_match_str_valid = false;
+                    } else if (rewrite_line_start &&
+                               !startswith(rl_line_buffer,
+                                           rewrite_line_start->c_str())) {
+                        // If the line was rewritten, the extra text stays on
+                        // the screen, so we need to delete it, make sure the
+                        // append character is there, and redisplay.  For
+                        // example, ':co<TAB>' will complete ':comment' and
+                        // append the current comment.  Pressing '<TAB>' again
+                        // would switch to ':config' and the comment text would
+                        // be left on the display.
+                        rl_delete_text(rl_point, rl_end);
+                        if (rl_completion_append_character &&
+                            rl_line_buffer[rl_point] !=
+                            rl_completion_append_character) {
+                            char buf[2] = {
+                                (char) rl_completion_append_character, '\0'
+                            };
+
+                            rl_insert_text(buf);
+                        }
+                        rl_redisplay();
                     }
+                    rewrite_line_start = nonstd::nullopt;
+
+                    SpookyHash::Hash128(rl_line_buffer, rl_end, &h1, &h2);
 
                     if (h1 == last_h1 && h2 == last_h2) {
                         // do nothing
@@ -662,6 +801,13 @@ void readline_curses::start()
                     }
                     last_h1 = h1;
                     last_h2 = h2;
+                    if (sendcmd(this->rc_command_pipe[RCF_SLAVE],
+                                'w',
+                                "",
+                                0) != 0) {
+                        perror("line: write failed");
+                        _exit(1);
+                    }
                 }
             }
             if (FD_ISSET(this->rc_command_pipe[RCF_SLAVE], &ready_rfds)) {
@@ -683,7 +829,15 @@ void readline_curses::start()
                         rl_extend_line_buffer(strlen(initial) + 1);
                         strcpy(rl_line_buffer, initial);
                         rl_end = strlen(initial);
+                        rewrite_line_start = std::string(rl_line_buffer, rl_point);
                         rl_redisplay();
+                        if (sendcmd(this->rc_command_pipe[RCF_SLAVE],
+                                    'c',
+                                    rl_line_buffer,
+                                    rl_end) != 0) {
+                            perror("line: write failed");
+                            _exit(1);
+                        }
                     }
                     else if (sscanf(msg, "f:%d:%n", &context, &prompt_start) == 1 &&
                              prompt_start != 0 &&
@@ -700,6 +854,13 @@ void readline_curses::start()
                             perror("line: write failed");
                             _exit(1);
                         }
+                        if (sendcmd(this->rc_command_pipe[RCF_SLAVE],
+                                    'w',
+                                    "",
+                                    0) != 0) {
+                            perror("line: write failed");
+                            _exit(1);
+                        }
                     }
                     else if (strcmp(msg, "a") == 0) {
                         char reply[4];
@@ -723,7 +884,7 @@ void readline_curses::start()
                                     &context,
                                     type,
                                     &prompt_start) == 2) {
-                        require(this->rc_contexts[context] != NULL);
+                        require(this->rc_contexts[context] != nullptr);
 
                         this->rc_contexts[context]->rc_prefixes[string(type)] =
                             string(&msg[prompt_start]);
@@ -733,18 +894,22 @@ void readline_curses::start()
                                     &context,
                                     type,
                                     &prompt_start) == 2) {
-                        require(this->rc_contexts[context] != NULL);
+                        require(this->rc_contexts[context] != nullptr);
 
                         this->rc_contexts[context]->
                                                       add_possibility(string(type),
                                                                       string(&msg[prompt_start]));
+                        if (rl_last_func == rl_complete ||
+                            rl_last_func == rl_menu_complete) {
+                            rl_last_func = NULL;
+                        }
                     }
                     else if (sscanf(msg,
                                     "rp:%d:%31[^:]:%n",
                                     &context,
                                     type,
                                     &prompt_start) == 2) {
-                        require(this->rc_contexts[context] != NULL);
+                        require(this->rc_contexts[context] != nullptr);
 
                         this->rc_contexts[context]->
                                                       rem_possibility(string(type),
@@ -780,33 +945,32 @@ void readline_curses::start()
             itv.it_value.tv_usec    = 0;
             itv.it_interval.tv_sec  = 0;
             itv.it_interval.tv_usec = 0;
-            if (setitimer(ITIMER_REAL, &itv, NULL) < 0) {
+            if (setitimer(ITIMER_REAL, &itv, nullptr) < 0) {
                 log_error("setitimer: %s", strerror(errno));
             }
-            current_context->second->save();
-            current_context = this->rc_contexts.end();
+            if (current_context != this->rc_contexts.end()) {
+                current_context->second->save();
+                current_context = this->rc_contexts.end();
+            }
         }
         if (got_winch) {
-            struct winsize ws;
+            struct winsize new_ws;
 
-            if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) == -1) {
+            if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &new_ws) == -1) {
                 throw error(errno);
             }
             got_winch = 0;
-            rl_set_screen_size(ws.ws_row, ws.ws_col);
+            rl_set_screen_size(new_ws.ws_row, new_ws.ws_col);
         }
     }
 
-    auto config_dir = dotlnav_path();
-    std::map<int, readline_context *>::iterator citer;
-    for (citer = this->rc_contexts.begin();
-         citer != this->rc_contexts.end();
-         ++citer) {
-        citer->second->load();
+    auto config_dir = lnav::paths::dotlnav();
+    for (auto& pair : this->rc_contexts) {
+        pair.second->load();
 
-        auto hpath = (config_dir / citer->second->get_name()) + ".history";
-        write_history(hpath.str().c_str());
-        citer->second->save();
+        auto hpath = (config_dir / pair.second->get_name()).string() + ".history";
+        write_history(hpath.c_str());
+        pair.second->save();
     }
 
     _exit(0);
@@ -817,7 +981,9 @@ void readline_curses::line_ready(const char *line)
     auto_mem<char> expanded;
     char           msg[1024] = {0};
     int            rc;
+    const char *cmd_ch = alt_done ? "D" : "d";
 
+    alt_done = false;
     if (line == nullptr) {
         snprintf(msg, sizeof(msg), "a");
 
@@ -848,13 +1014,13 @@ void readline_curses::line_ready(const char *line)
 #endif
 
     case -1:
-        snprintf(msg, sizeof(msg), "d:%s", line);
+        snprintf(msg, sizeof(msg), "%s:%s", cmd_ch, line);
         break;
 
     case 0:
     case 1:
     case 2: /* XXX */
-        snprintf(msg, sizeof(msg), "d:%s", expanded.in());
+        snprintf(msg, sizeof(msg), "%s:%s", cmd_ch, expanded.in());
         break;
     }
 
@@ -868,9 +1034,9 @@ void readline_curses::line_ready(const char *line)
     {
         HIST_ENTRY *entry;
 
-        if (line != nullptr && line[0] != '\0' && (
+        if (line[0] != '\0' && (
             history_length == 0 ||
-            (entry = history_get(history_base + history_length - 1)) == NULL ||
+            (entry = history_get(history_base + history_length - 1)) == nullptr ||
             strcmp(entry->line, line) != 0)) {
             add_history(line);
         }
@@ -890,7 +1056,7 @@ void readline_curses::check_poll_set(const vector<struct pollfd> &pollfds)
 
             this->map_output(buffer, rc);
             if (this->vc_x != old_x) {
-                this->rc_change.invoke(this);
+                this->rc_change(this);
             }
         }
     }
@@ -906,7 +1072,7 @@ void readline_curses::check_poll_set(const vector<struct pollfd> &pollfds)
                 this->rc_matches.emplace_back(msg);
                 this->rc_matches_remaining -= 1;
                 if (this->rc_matches_remaining == 0) {
-                    this->rc_display_match.invoke(this);
+                    this->rc_display_match(this);
                 }
             }
             else if (msg[0] == 'm') {
@@ -918,58 +1084,80 @@ void readline_curses::check_poll_set(const vector<struct pollfd> &pollfds)
                 }
                 this->rc_matches.clear();
                 if (this->rc_matches_remaining == 0) {
-                    this->rc_display_match.invoke(this);
+                    this->rc_display_match(this);
                 }
                 this->rc_match_index = 0;
             }
+            else if (msg[0] == '\t') {
+                char path[2048];
+
+                if (sscanf(msg, "\t:%s", path) != 1) {
+                    require(0);
+                }
+                this->rc_remote_complete_path = path;
+                this->rc_completion_request(this);
+            }
             else if (msg[0] == 'n') {
                 if (sscanf(msg, "n:%d", &this->rc_match_index) != 1) {
                     require(0);
                 }
-                this->rc_display_next.invoke(this);
+                this->rc_display_next(this);
             }
             else {
                 switch (msg[0]) {
                 case 't':
                 case 'd':
+                case 'D':
                     this->rc_value = string(&msg[2]);
                     break;
                 }
                 switch (msg[0]) {
                 case 'a':
+                    curs_set(0);
                     this->vc_line.clear();
                     this->rc_active_context = -1;
                     this->rc_matches.clear();
-                    this->rc_abort.invoke(this);
-                    this->rc_display_match.invoke(this);
-                    this->rc_blur.invoke(this);
-                    curs_set(0);
+                    this->rc_abort(this);
+                    this->rc_display_match(this);
+                    this->rc_blur(this);
                     break;
 
                 case 't':
-                    this->rc_timeout.invoke(this);
+                    this->rc_timeout(this);
                     break;
 
                 case 'd':
+                case 'D':
+                    curs_set(0);
                     this->rc_active_context = -1;
                     this->rc_matches.clear();
-                    this->rc_perform.invoke(this);
-                    this->rc_display_match.invoke(this);
-                    this->rc_blur.invoke(this);
-                    curs_set(0);
+                    if (msg[0] == 'D' || this->rc_is_alt_focus) {
+                        this->rc_alt_perform(this);
+                    } else {
+                        this->rc_perform(this);
+                    }
+                    this->rc_display_match(this);
+                    this->rc_blur(this);
                     break;
 
                 case 'l':
                     this->rc_line_buffer = &msg[2];
-                    this->rc_change.invoke(this);
+                    if (this->rc_active_context != -1) {
+                        this->rc_change(this);
+                    }
                     this->rc_matches.clear();
-                    this->rc_display_match.invoke(this);
+                    if (this->rc_active_context != -1) {
+                        this->rc_display_match(this);
+                    }
                     break;
 
                 case 'c':
                     this->rc_line_buffer = &msg[2];
-                    this->rc_change.invoke(this);
-                    this->rc_display_match.invoke(this);
+                    this->rc_change(this);
+                    this->rc_display_match(this);
+                    break;
+                case 'w':
+                    this->rc_ready_for_input = true;
                     break;
                 }
             }
@@ -988,7 +1176,7 @@ void readline_curses::handle_key(int ch)
     }
 }
 
-void readline_curses::focus(int context, const char *prompt, const char *initial)
+void readline_curses::focus(int context, const std::string& prompt, const std::string& initial)
 {
     char buffer[1024];
 
@@ -996,7 +1184,7 @@ void readline_curses::focus(int context, const char *prompt, const char *initial
 
     this->rc_active_context = context;
 
-    snprintf(buffer, sizeof(buffer), "f:%d:%s", context, prompt);
+    snprintf(buffer, sizeof(buffer), "f:%d:%s", context, prompt.c_str());
     if (sendstring(this->rc_command_pipe[RCF_MASTER],
                    buffer,
                    strlen(buffer) + 1) == -1) {
@@ -1004,17 +1192,14 @@ void readline_curses::focus(int context, const char *prompt, const char *initial
     }
     wmove(this->vc_window, this->get_actual_y(), this->vc_left);
     wclrtoeol(this->vc_window);
-    if (initial != nullptr) {
-        snprintf(buffer, sizeof(buffer), "i:0:%s", initial);
-        if (sendstring(this->rc_command_pipe[RCF_MASTER],
-                       buffer,
-                       strlen(buffer) + 1) == -1) {
-            perror("focus: write failed");
-        }
+    if (!initial.empty()) {
+        this->rewrite_line(initial.size(), initial);
     }
+    this->rc_is_alt_focus = false;
+    this->rc_focus(this);
 }
 
-void readline_curses::rewrite_line(int pos, std::string value)
+void readline_curses::rewrite_line(int pos, const std::string& value)
 {
     char buffer[1024];
 
@@ -1022,7 +1207,7 @@ void readline_curses::rewrite_line(int pos, std::string value)
     if (sendstring(this->rc_command_pipe[RCF_MASTER],
                    buffer,
                    strlen(buffer) + 1) == -1) {
-        perror("focus: write failed");
+        perror("rewrite_line: write failed");
     }
 }
 
@@ -1044,7 +1229,7 @@ void readline_curses::add_prefix(int context,
                                  const string &value)
 {
     char buffer[1024];
-    string prefix_wire = join(prefix.begin(), prefix.end(), "\x1f");
+    auto prefix_wire = fmt::format("{}", fmt::join(prefix, "\x1f"));
 
     snprintf(buffer, sizeof(buffer),
              "apre:%d:%s\x1d%s",
@@ -1190,10 +1375,17 @@ std::string readline_curses::get_match_string() const
     auto len = ::min((size_t) this->vc_x, this->rc_line_buffer.size()) - this->rc_match_start;
     auto context = this->get_active_context();
 
-    if (context->get_append_character() != 0 &&
-        this->rc_line_buffer[this->rc_match_start + len - 2] ==
-        context->get_append_character()) {
-        len -= 2;
+    if (context->get_append_character() != 0) {
+        if (this->rc_line_buffer.length() > (this->rc_match_start + len - 1) &&
+            this->rc_line_buffer[this->rc_match_start + len - 1] ==
+            context->get_append_character()) {
+            len -= 1;
+        } else if (this->rc_line_buffer.length() >
+                   (this->rc_match_start + len - 2) &&
+                   this->rc_line_buffer[this->rc_match_start + len - 2] ==
+                   context->get_append_character()) {
+            len -= 2;
+        }
     }
 
     return this->rc_line_buffer.substr(this->rc_match_start, len);
