@@ -31,15 +31,17 @@
 
 #include "config.h"
 
+#include "base/string_util.hh"
 #include "pcrepp/pcrepp.hh"
 #include "sql_util.hh"
 #include "shlex.hh"
-#include "lnav_util.hh"
 
 #include "readline_highlighters.hh"
 
 using namespace std;
 
+static void readline_sqlite_highlighter_int(attr_line_t &al, int x, int skip);
+
 static bool check_re_prev(const string &line, int x)
 {
     bool retval = false;
@@ -182,7 +184,7 @@ static void readline_regex_highlighter_int(attr_line_t &al, int x, int skip)
         "()",
         "QE",
 
-        NULL
+        nullptr
     };
 
     string &line = al.get_string();
@@ -358,9 +360,10 @@ void readline_command_highlighter(attr_line_t &al, int x)
     static const pcrepp RE_PREFIXES(
         R"(^:(filter-in|filter-out|delete-filter|enable-filter|disable-filter|highlight|clear-highlight|create-search-table\s+[^\s]+\s+))");
     static const pcrepp SH_PREFIXES("^:(eval|open|append-to|write-to|write-csv-to|write-json-to)");
+    static const pcrepp SQL_PREFIXES("^:(filter-expr|mark-expr)");
     static const pcrepp IDENT_PREFIXES("^:(tag|untag|delete-tags)");
     static const pcrepp COLOR_PREFIXES("^:(config)");
-    static const pcrepp COLOR_RE("(#(?:[a-fA-F0-9]{3}|[a-fA-F0-9]{6}))");
+    static const pcrepp COLOR_RE("(#(?:[a-fA-F0-9]{6}|[a-fA-F0-9]{3}))");
 
     view_colors &vc = view_colors::singleton();
     int keyword_attrs = (
@@ -388,25 +391,22 @@ void readline_command_highlighter(attr_line_t &al, int x)
         readline_shlex_highlighter(al, x);
     }
     pi.reset(line);
+    if (SQL_PREFIXES.match(pc, pi)) {
+        readline_sqlite_highlighter_int(al, x, 1 + pc[0]->length());
+    }
+    pi.reset(line);
     if (COLOR_PREFIXES.match(pc, pi)) {
         pi.reset(line);
         if (COLOR_RE.match(pc, pi)) {
             pcre_context::capture_t *cap = pc[0];
             string hash_color = pi.get_substr(cap);
-            string errmsg;
-            rgb_color rgb_fg, rgb_bg;
-            attr_t color_hint_attrs = vc.attrs_for_role(view_colors::VCR_COLOR_HINT);
-            int pnum = PAIR_NUMBER(color_hint_attrs);
-
-            if (rgb_color::from_str(hash_color, rgb_bg, errmsg)) {
-                pnum -= 1;
-                vc.ensure_color_pair(pnum, rgb_fg, rgb_bg);
 
+            styling::color_unit::from_str(hash_color).then([&](const auto& rgb_fg) {
                 al.get_attrs().emplace_back(
                     line_range{cap->c_begin, cap->c_begin + 1},
                     &view_curses::VC_ROLE,
                     view_colors::VCR_COLOR_HINT);
-            }
+            });
         }
     }
     pi.reset(line);
@@ -437,21 +437,21 @@ void readline_command_highlighter(attr_line_t &al, int x)
     }
 }
 
-void readline_sqlite_highlighter(attr_line_t &al, int x)
+static void readline_sqlite_highlighter_int(attr_line_t &al, int x, int skip)
 {
-    static string keyword_re_str = sql_keyword_re() + "|\\.schema|\\.msgformats";
+    static string keyword_re_str = sql_keyword_re();
     static pcrepp keyword_pcre(keyword_re_str.c_str(), PCRE_CASELESS);
     static pcrepp string_literal_pcre("'[^']*('(?:'[^']*')*|$)");
-    static pcrepp ident_pcre("(\\$?\\b[a-z_]\\w*)|\"([^\"]+)\"|\\[([^\\]]+)]", PCRE_CASELESS);
+    static pcrepp ident_pcre("(?:\\$|:)?(\\b[a-z_]\\w*)|\"([^\"]+)\"|\\[([^\\]]+)]", PCRE_CASELESS);
 
     static const char *brackets[] = {
         "[]",
         "()",
 
-        NULL
+        nullptr
     };
 
-    view_colors &vc = view_colors::singleton();
+    auto &vc = view_colors::singleton();
 
     int keyword_attrs = vc.attrs_for_role(view_colors::VCR_KEYWORD);
     int symbol_attrs = vc.attrs_for_role(view_colors::VCR_SYMBOL);
@@ -459,8 +459,19 @@ void readline_sqlite_highlighter(attr_line_t &al, int x)
     int error_attrs = vc.attrs_for_role(view_colors::VCR_ERROR) | A_REVERSE;
 
     pcre_context_static<30> pc;
-    pcre_input pi(al.get_string());
-    string &line = al.get_string();
+    pcre_input pi(al.get_string(), skip);
+    auto &line = al.get_string();
+
+    if (startswith(line, ";.")) {
+        auto space = line.find(' ');
+        struct line_range lr{2, -1};
+
+        if (space != std::string::npos) {
+            lr.lr_end = space;
+        }
+        al.get_attrs().emplace_back(lr, &view_curses::VC_STYLE, keyword_attrs);
+        return;
+    }
 
     while (ident_pcre.match(pc, pi)) {
         pcre_context::capture_t *cap = pc.first_valid();
@@ -476,7 +487,7 @@ void readline_sqlite_highlighter(attr_line_t &al, int x)
         }
     }
 
-    pi.reset(line);
+    pi.reset(line, skip);
 
     while (keyword_pcre.match(pc, pi)) {
         pcre_context::capture_t *cap = pc.all();
@@ -487,7 +498,7 @@ void readline_sqlite_highlighter(attr_line_t &al, int x)
             keyword_attrs);
     }
 
-    for (size_t lpc = 0; lpc < line.length(); lpc++) {
+    for (size_t lpc = skip; lpc < line.length(); lpc++) {
         switch (line[lpc]) {
         case '*':
         case '<':
@@ -504,7 +515,7 @@ void readline_sqlite_highlighter(attr_line_t &al, int x)
         }
     }
 
-    pi.reset(line);
+    pi.reset(line, skip);
 
     while (string_literal_pcre.match(pc, pi)) {
         pcre_context::capture_t *cap = pc.all();
@@ -528,6 +539,11 @@ void readline_sqlite_highlighter(attr_line_t &al, int x)
     }
 }
 
+void readline_sqlite_highlighter(attr_line_t &al, int x)
+{
+    readline_sqlite_highlighter_int(al, x, 0);
+}
+
 void readline_shlex_highlighter(attr_line_t &al, int x)
 {
     view_colors &vc = view_colors::singleton();
@@ -572,7 +588,7 @@ void readline_shlex_highlighter(attr_line_t &al, int x)
             case shlex_token_t::ST_QUOTED_VARIABLE_REF: {
                 int extra = token == shlex_token_t::ST_VARIABLE_REF ? 0 : 1;
                 string ident = str.substr(cap.c_begin + 1 + extra, cap.length() - 1 - extra * 2);
-                int attrs = vc.attrs_for_ident(ident.c_str(), ident.size());
+                int attrs = vc.attrs_for_ident(ident);
 
                 al.with_attr(string_attr(
                         line_range(cap.c_begin, cap.c_begin + 1 + extra),
