@@ -29,15 +29,9 @@
 
 #include "config.h"
 
-#include <stdlib.h>
-#include <unistd.h>
-#include <string.h>
-
-#include "logfile.hh"
-#include "auto_mem.hh"
 #include "base/lnav_log.hh"
+#include "pcrepp/pcrepp.hh"
 #include "sql_util.hh"
-#include "file_vtab.hh"
 #include "vtab_module.hh"
 
 using namespace std;
@@ -55,8 +49,7 @@ enum {
 };
 
 struct regexp_capture {
-    using iterator = vector<logfile *>::iterator;
-
+    static constexpr const char *NAME = "regexp_capture";
     static constexpr const char *CREATE_STMT = R"(
 -- The regexp_capture() table-valued function allows you to execute a regular-
 -- expression over a given string and get the captured data as rows in a table.
@@ -73,24 +66,16 @@ CREATE TABLE regexp_capture (
 );
 )";
 
-    struct vtab {
-        sqlite3_vtab base;
-
-        operator sqlite3_vtab *() {
-            return &this->base;
-        };
-    };
-
     struct cursor {
         sqlite3_vtab_cursor base;
-        unique_ptr<pcrepp> c_pattern;
+        pcrepp c_pattern;
         pcre_context_static<30> c_context;
         unique_ptr<pcre_input> c_input;
-        string c_pattern_string;
         string c_content;
+        bool c_content_as_blob{false};
         int c_index;
         int c_start_index;
-        bool c_matched;
+        bool c_matched{false};
         int c_match_index;
         sqlite3_int64 c_rowid;
 
@@ -110,12 +95,12 @@ CREATE TABLE regexp_capture (
         int next() {
             if (this->c_index >= (this->c_context.get_count() - 1)) {
                 this->c_input->pi_offset = this->c_input->pi_next_offset;
-                this->c_matched = this->c_pattern->match(this->c_context, *(this->c_input));
+                this->c_matched = this->c_pattern.match(this->c_context, *(this->c_input));
                 this->c_index = -1;
                 this->c_match_index += 1;
             }
 
-            if (!this->c_pattern || !this->c_matched) {
+            if (this->c_pattern.empty() || !this->c_matched) {
                 return SQLITE_OK;
             }
 
@@ -125,7 +110,7 @@ CREATE TABLE regexp_capture (
         };
 
         int eof() {
-            return !this->c_pattern || !this->c_matched;
+            return this->c_pattern.empty() || !this->c_matched;
         };
 
         int get_rowid(sqlite3_int64 &rowid_out) {
@@ -149,7 +134,7 @@ CREATE TABLE regexp_capture (
                 if (vc.c_index == 0) {
                     sqlite3_result_null(ctx);
                 } else {
-                    sqlite3_result_text(ctx, vc.c_pattern->name_for_capture(
+                    sqlite3_result_text(ctx, vc.c_pattern.name_for_capture(
                         vc.c_index - 1), -1, SQLITE_TRANSIENT);
                 }
                 break;
@@ -157,10 +142,10 @@ CREATE TABLE regexp_capture (
                 sqlite3_result_int64(ctx, vc.c_context.get_count());
                 break;
             case RC_COL_RANGE_START:
-                sqlite3_result_int64(ctx, cap.c_begin);
+                sqlite3_result_int64(ctx, cap.c_begin + 1);
                 break;
             case RC_COL_RANGE_STOP:
-                sqlite3_result_int64(ctx, cap.c_end);
+                sqlite3_result_int64(ctx, cap.c_end + 1);
                 break;
             case RC_COL_CONTENT:
                 if (cap.is_valid()) {
@@ -173,17 +158,25 @@ CREATE TABLE regexp_capture (
                 }
                 break;
             case RC_COL_VALUE:
-                sqlite3_result_text(ctx,
-                                    vc.c_content.c_str(),
-                                    vc.c_content.length(),
-                                    SQLITE_TRANSIENT);
+                if (vc.c_content_as_blob) {
+                    sqlite3_result_blob64(ctx,
+                                          vc.c_content.c_str(),
+                                          vc.c_content.length(),
+                                          SQLITE_STATIC);
+                } else {
+                    sqlite3_result_text(ctx,
+                                        vc.c_content.c_str(),
+                                        vc.c_content.length(),
+                                        SQLITE_STATIC);
+                }
                 break;
-            case RC_COL_PATTERN:
-                sqlite3_result_text(ctx,
-                                    vc.c_pattern_string.c_str(),
-                                    vc.c_pattern_string.length(),
+            case RC_COL_PATTERN: {
+                auto str = vc.c_pattern.get_pattern();
+
+                sqlite3_result_text(ctx, str.c_str(), str.length(),
                                     SQLITE_TRANSIENT);
                 break;
+            }
         }
 
         return SQLITE_OK;
@@ -220,29 +213,33 @@ static int rcFilter(sqlite3_vtab_cursor *pVtabCursor,
 
     if (argc != 2) {
         pCur->c_content.clear();
-        pCur->c_pattern.reset();
+        pCur->c_pattern.clear();
         return SQLITE_OK;
     }
 
-    const char *value = (const char *) sqlite3_value_text(argv[0]);
+    auto byte_count = sqlite3_value_bytes(argv[0]);
+    auto blob = (const char *) sqlite3_value_blob(argv[0]);
+
+    pCur->c_content_as_blob = (sqlite3_value_type(argv[0]) == SQLITE_BLOB);
+    pCur->c_content.assign(blob, byte_count);
+
     const char *pattern = (const char *) sqlite3_value_text(argv[1]);
-
-    pCur->c_content = value;
-
-    try {
-        pCur->c_pattern = make_unique<pcrepp>(pattern);
-        pCur->c_pattern_string = pattern;
-    } catch (const pcrepp::error &e) {
+    auto re_res = pcrepp::from_str(pattern);
+    if (re_res.isErr()) {
         pVtabCursor->pVtab->zErrMsg = sqlite3_mprintf(
-            "Invalid regular expression: %s", e.e_msg.c_str());
+            "Invalid regular expression: %s", re_res.unwrapErr().ce_msg);
         return SQLITE_ERROR;
     }
 
+    pCur->c_pattern = re_res.unwrap();
+
     pCur->c_index = 0;
     pCur->c_context.set_count(0);
 
     pCur->c_input = make_unique<pcre_input>(pCur->c_content);
-    pCur->c_matched = pCur->c_pattern->match(pCur->c_context, *(pCur->c_input));
+    pCur->c_matched = pCur->c_pattern.match(pCur->c_context, *(pCur->c_input));
+
+    log_debug("matched %d", pCur->c_matched);
 
     return SQLITE_OK;
 }
