@@ -35,26 +35,29 @@
 #include <glob.h>
 #include <fcntl.h>
 #include <sys/types.h>
-#include <dirent.h>
-
-#include "spookyhash/SpookyV2.h"
 
 #include <algorithm>
 #include <utility>
 #include <yajl/api/yajl_tree.h>
 
+#include "base/isc.hh"
+#include "base/paths.hh"
+#include "tailer/tailer.looper.hh"
 #include "yajlpp/yajlpp.hh"
 #include "yajlpp/yajlpp_def.hh"
 #include "lnav.hh"
 #include "logfile.hh"
 #include "sql_util.hh"
 #include "lnav_util.hh"
-#include "lnav_config.hh"
 #include "session_data.hh"
 #include "command_executor.hh"
+#include "log_format_ext.hh"
+#include "service_tags.hh"
 
 using namespace std;
 
+struct session_data_t session_data;
+
 static const char *LOG_METADATA_NAME = "log_metadata.db";
 
 static const char *BOOKMARK_TABLE_DEF = R"(
@@ -82,6 +85,14 @@ CREATE TABLE IF NOT EXISTS time_offset (
 
     PRIMARY KEY (log_time, log_format, log_hash, session_time)
 );
+
+CREATE TABLE IF NOT EXISTS recent_netlocs (
+    netloc text,
+
+    access_time datetime DEFAULT CURRENT_TIMESTAMP,
+
+    PRIMARY KEY (netloc)
+);
 )";
 
 static const char *BOOKMARK_LRU_STMT =
@@ -89,6 +100,11 @@ static const char *BOOKMARK_LRU_STMT =
     "  (SELECT access_time FROM bookmarks "
     "   ORDER BY access_time DESC LIMIT 1 OFFSET 50000)";
 
+static const char *NETLOC_LRU_STMT =
+    "DELETE FROM recent_netlocs WHERE access_time <= "
+    "  (SELECT access_time FROM bookmarks "
+    "   ORDER BY access_time DESC LIMIT 1 OFFSET 10)";
+
 static const char *UPGRADE_STMTS[] = {
     R"(ALTER TABLE bookmarks ADD COLUMN comment text DEFAULT '';)",
     R"(ALTER TABLE bookmarks ADD COLUMN tags text DEFAULT '';)",
@@ -172,9 +188,10 @@ static bool bind_line(sqlite3 *db,
     }
 
     auto line_hash = read_result.map([cl](auto sbr) {
-        return hash_bytes(sbr.get_data(), sbr.length(),
-                          &cl, sizeof(cl),
-                          nullptr);
+        return hasher()
+            .update(sbr.get_data(), sbr.length())
+            .update(cl)
+            .to_string();
     }).unwrap();
 
     return bind_values(stmt,
@@ -214,9 +231,9 @@ static void cleanup_session_data()
     static_root_mem<glob_t, globfree>   session_file_list;
     std::list<struct session_file_info> session_info_list;
     map<string, int> session_count;
-    auto session_file_pattern = dotlnav_path() / "*-*.ts*.json";
+    auto session_file_pattern = lnav::paths::dotlnav() / "*-*.ts*.json";
 
-    if (glob(session_file_pattern.str().c_str(),
+    if (glob(session_file_pattern.c_str(),
              0,
              nullptr,
              session_file_list.inout()) == 0) {
@@ -263,7 +280,7 @@ static void cleanup_session_data()
         else {
             if (remove(front.sfi_path.c_str()) != 0) {
                 log_error(
-                        "Unable to remove session file: %s -- %s\n",
+                        "Unable to remove session file: %s -- %s",
                         front.sfi_path.c_str(),
                         strerror(errno));
             }
@@ -279,7 +296,7 @@ static void cleanup_session_data()
 
         if (remove(front.sfi_path.c_str()) != 0) {
             log_error(
-                    "Unable to remove session file: %s -- %s\n",
+                    "Unable to remove session file: %s -- %s",
                     front.sfi_path.c_str(),
                     strerror(errno));
         }
@@ -296,25 +313,21 @@ void init_session()
 
 static nonstd::optional<std::string> compute_session_id()
 {
-    byte_array<2, uint64> hash;
-    SpookyHash context;
     bool has_files = false;
+    hasher h;
 
-    context.Init(0, 0);
-    hash_updater updater(&context);
-    for (auto &ld_file_name : lnav_data.ld_file_names) {
+    for (auto &ld_file_name : lnav_data.ld_active_files.fc_file_names) {
         if (!ld_file_name.second.loo_include_in_session) {
             continue;
         }
         has_files = true;
-        updater(ld_file_name.first);
+        h.update(ld_file_name.first);
     }
     if (!has_files) {
         return nonstd::nullopt;
     }
-    context.Final(hash.out(0), hash.out(1));
 
-    return hash.to_string();
+    return h.to_string();
 }
 
 nonstd::optional<session_pair_t> scan_sessions()
@@ -336,8 +349,8 @@ nonstd::optional<session_pair_t> scan_sessions()
     snprintf(view_info_pattern_base, sizeof(view_info_pattern_base),
              "view-info-%s.*.json",
              session_id.value().c_str());
-    auto view_info_pattern = dotlnav_path() / view_info_pattern_base;
-    if (glob(view_info_pattern.str().c_str(), 0, nullptr,
+    auto view_info_pattern = lnav::paths::dotlnav() / view_info_pattern_base;
+    if (glob(view_info_pattern.c_str(), 0, nullptr,
              view_info_list.inout()) == 0) {
         for (size_t lpc = 0; lpc < view_info_list->gl_pathc; lpc++) {
             const char *path = view_info_list->gl_pathv[lpc];
@@ -369,7 +382,7 @@ nonstd::optional<session_pair_t> scan_sessions()
 
         if (remove(name.c_str()) != 0) {
             log_error(
-                    "Unable to remove session: %s -- %s\n",
+                    "Unable to remove session: %s -- %s",
                     name.c_str(),
                     strerror(errno));
         }
@@ -383,26 +396,52 @@ nonstd::optional<session_pair_t> scan_sessions()
     return nonstd::make_optional(session_file_names.back());
 }
 
-static void load_time_bookmarks()
+void load_time_bookmarks()
 {
     logfile_sub_source &lss = lnav_data.ld_log_source;
     std::map<content_line_t, bookmark_metadata> &bm_meta = lss.get_user_bookmark_metadata();
     auto_mem<sqlite3, sqlite_close_wrapper> db;
-    auto db_path = dotlnav_path() / LOG_METADATA_NAME;
+    auto db_path = lnav::paths::dotlnav() / LOG_METADATA_NAME;
     auto_mem<sqlite3_stmt> stmt(sqlite3_finalize);
     logfile_sub_source::iterator file_iter;
     bool reload_needed = false;
     auto_mem<char, sqlite3_free> errmsg;
 
-    log_info("loading bookmark db: %s", db_path.str().c_str());
+    log_info("loading bookmark db: %s", db_path.c_str());
 
-    if (sqlite3_open(db_path.str().c_str(), db.out()) != SQLITE_OK) {
+    if (sqlite3_open(db_path.c_str(), db.out()) != SQLITE_OK) {
         return;
     }
 
     for (const char *upgrade_stmt : UPGRADE_STMTS) {
         if (sqlite3_exec(db.in(), upgrade_stmt, nullptr, nullptr, errmsg.out()) != SQLITE_OK) {
-            log_error("unable to upgrade bookmark table -- %s\n", errmsg.in());
+            log_error("unable to upgrade bookmark table -- %s", errmsg.in());
+        }
+    }
+
+    {
+        if (sqlite3_prepare_v2(db.in(),
+                               "SELECT netloc FROM recent_netlocs",
+                               -1,
+                               stmt.out(),
+                               nullptr) != SQLITE_OK) {
+            return;
+        }
+
+        bool done = false;
+
+        while (!done) {
+            switch (sqlite3_step(stmt.in())) {
+                case SQLITE_ROW: {
+                    auto netloc = sqlite3_column_text(stmt.in(), 0);
+
+                    session_data.sd_recent_netlocs.insert((const char *) netloc);
+                    break;
+                }
+                default:
+                    done = true;
+                    break;
+            }
         }
     }
 
@@ -415,7 +454,7 @@ static void load_time_bookmarks()
                            stmt.out(),
                            nullptr) != SQLITE_OK) {
         log_error(
-                "could not prepare bookmark select statement -- %s\n",
+                "could not prepare bookmark select statement -- %s",
                 sqlite3_errmsg(db));
         return;
     }
@@ -494,7 +533,7 @@ static void load_time_bookmarks()
                         break;
                     }
 
-                    content_line_t cl = content_line_t(std::distance(lf->begin(), line_iter));
+                    auto cl = content_line_t(std::distance(lf->begin(), line_iter));
                     auto read_result = lf->read_line(line_iter);
 
                     if(read_result.isErr()) {
@@ -503,9 +542,10 @@ static void load_time_bookmarks()
 
                     auto sbr = read_result.unwrap();
 
-                    string line_hash = hash_bytes(sbr.get_data(), sbr.length(),
-                                                  &cl, sizeof(cl),
-                                                  nullptr);
+                    string line_hash = hasher()
+                        .update(sbr.get_data(), sbr.length())
+                        .update(cl)
+                        .to_string();
 
                     if (line_hash == log_hash) {
                         content_line_t line_cl = content_line_t(
@@ -564,7 +604,7 @@ static void load_time_bookmarks()
 
                     errmsg = sqlite3_errmsg(lnav_data.ld_db);
                     log_error(
-                            "bookmark select error: code %d -- %s\n",
+                            "bookmark select error: code %d -- %s",
                             rc,
                             errmsg);
                     done = true;
@@ -584,7 +624,7 @@ static void load_time_bookmarks()
                            stmt.out(),
                            nullptr) != SQLITE_OK) {
         log_error(
-                "could not prepare time_offset select statement -- %s\n",
+                "could not prepare time_offset select statement -- %s",
                 sqlite3_errmsg(db));
         return;
     }
@@ -662,16 +702,7 @@ static void load_time_bookmarks()
                         break;
                     }
 
-                    auto read_result = lf->read_line(line_iter);
-
-                    if (read_result.isErr()) {
-                        break;
-                    }
-
-                    auto sbr = read_result.unwrap();
-
-                    string line_hash = hash_bytes(sbr.get_data(), sbr.length(), nullptr);
-                    if (line_hash == log_hash) {
+                    if (lf->get_content_id() == log_hash) {
                         int file_line = std::distance(lf->begin(), line_iter);
                         content_line_t line_cl = content_line_t(
                             base_content_line + file_line);
@@ -696,7 +727,7 @@ static void load_time_bookmarks()
 
                     errmsg = sqlite3_errmsg(lnav_data.ld_db);
                     log_error(
-                            "bookmark select error: code %d -- %s\n",
+                            "bookmark select error: code %d -- %s",
                             rc,
                             errmsg);
                     done = true;
@@ -713,24 +744,12 @@ static void load_time_bookmarks()
     }
 }
 
-static int read_save_time(yajlpp_parse_context *ypc, long long value)
-{
-    lnav_data.ld_session_save_time = value;
-
-    return 1;
-}
-
-static int read_time_offset(yajlpp_parse_context *ypc, int value)
-{
-    return 1;
-}
-
 static int read_files(yajlpp_parse_context *ypc, const unsigned char *str, size_t len)
 {
     return 1;
 }
 
-static int read_last_search(yajlpp_parse_context *ypc, const unsigned char *str, size_t len)
+static int read_current_search(yajlpp_parse_context *ypc, const unsigned char *str, size_t len)
 {
     string       regex        = std::string((const char *)str, len);
     const char **view_name;
@@ -743,7 +762,7 @@ static int read_last_search(yajlpp_parse_context *ypc, const unsigned char *str,
 
     if (view_index < LNV__MAX && !regex.empty()) {
         lnav_data.ld_views[view_index].execute_search(regex);
-        lnav_data.ld_views[view_index].set_follow_search_for(-1);
+        lnav_data.ld_views[view_index].set_follow_search_for(-1, {});
     }
 
     return 1;
@@ -759,11 +778,7 @@ static int read_top_line(yajlpp_parse_context *ypc, long long value)
                      ypc->get_path_fragment(-2));
     view_index = view_name - lnav_view_strings;
     if (view_index < LNV__MAX) {
-        textview_curses &tc = lnav_data.ld_views[view_index];
-
-        if (value != -1 && value < tc.get_inner_height()) {
-            tc.set_top(vis_line_t(value));
-        }
+        session_data.sd_view_states[view_index].vs_top = value;
     }
 
     return 1;
@@ -790,6 +805,29 @@ static int read_word_wrap(yajlpp_parse_context *ypc, int value)
     return 1;
 }
 
+static int read_filtering(yajlpp_parse_context *ypc, int value)
+{
+    const char **view_name;
+    int view_index;
+
+    view_name = find(lnav_view_strings,
+                     lnav_view_strings + LNV__MAX,
+                     ypc->get_path_fragment(-2));
+    view_index = view_name - lnav_view_strings;
+    if (view_index == LNV_HELP) {
+
+    }
+    else if (view_index < LNV__MAX) {
+        textview_curses &tc = lnav_data.ld_views[view_index];
+
+        if (tc.get_sub_source() != nullptr) {
+            tc.get_sub_source()->tss_apply_filters = value;
+        }
+    }
+
+    return 1;
+}
+
 static int read_commands(yajlpp_parse_context *ypc, const unsigned char *str, size_t len)
 {
     std::string cmdline = std::string((const char *)str, len);
@@ -803,7 +841,7 @@ static int read_commands(yajlpp_parse_context *ypc, const unsigned char *str, si
     bool active = ensure_view(&lnav_data.ld_views[view_index]);
     execute_command(lnav_data.ld_exec_context, cmdline);
     if (!active) {
-        lnav_data.ld_view_stack.vs_views.pop_back();
+        lnav_data.ld_view_stack.pop_back();
     }
 
     return 1;
@@ -811,22 +849,43 @@ static int read_commands(yajlpp_parse_context *ypc, const unsigned char *str, si
 
 static struct json_path_container view_def_handlers = {
     json_path_handler("top_line",  read_top_line),
-    json_path_handler("search",    read_last_search),
+    json_path_handler("search",    read_current_search),
     json_path_handler("word_wrap", read_word_wrap),
+    json_path_handler("filtering", read_filtering),
     json_path_handler("commands#", read_commands)
 };
 
 static struct json_path_container view_handlers = {
-    json_path_handler(pcrepp("([^/]+)"))
+    yajlpp::pattern_property_handler("([^/]+)")
         .with_children(view_def_handlers)
 };
 
+static struct json_path_container file_state_handlers = {
+    yajlpp::property_handler("visible")
+        .with_description("Indicates whether the file is visible or not")
+        .FOR_FIELD(file_state, fs_is_visible),
+};
+
+static struct json_path_container file_states_handlers = {
+    yajlpp::pattern_property_handler(R"((?<filename>[^/]+))")
+        .with_description("Map of file names to file state objects")
+        .with_obj_provider<file_state, void>([](const auto& ypc, auto *root) {
+            auto fn = ypc.get_substr("filename");
+            return &session_data.sd_file_states[fn];
+        })
+        .with_children(file_state_handlers),
+};
+
 static struct json_path_container view_info_handlers = {
-    json_path_handler("save-time",               read_save_time),
-    json_path_handler("time-offset",             read_time_offset),
+    yajlpp::property_handler("save-time")
+        .FOR_FIELD(session_data_t, sd_save_time),
+    yajlpp::property_handler("time-offset")
+        .FOR_FIELD(session_data_t, sd_time_offset),
     json_path_handler("files#",                  read_files),
-    json_path_handler("views")
-        .with_children(view_handlers)
+    yajlpp::property_handler("file-states")
+        .with_children(file_states_handlers),
+    yajlpp::property_handler("views")
+        .with_children(view_handlers),
 };
 
 void load_session()
@@ -837,28 +896,60 @@ void load_session()
         auto_fd fd;
 
         lnav_data.ld_session_load_time = pair.first.second;
-        lnav_data.ld_session_save_time = pair.first.second;
-        const string &view_info_name = pair.second;
+        session_data.sd_save_time = pair.first.second;
+        const auto& view_info_path = pair.second;
 
-        yajlpp_parse_context ypc(view_info_name, &view_info_handlers);
+        yajlpp_parse_context ypc(view_info_path, &view_info_handlers);
+        ypc.with_obj(session_data);
         handle = yajl_alloc(&ypc.ypc_callbacks, nullptr, &ypc);
 
         load_time_bookmarks();
 
-        if ((fd = open(view_info_name.c_str(), O_RDONLY)) < 0) {
+        if ((fd = openp(view_info_path, O_RDONLY)) < 0) {
             perror("cannot open session file");
         }
         else {
             unsigned char buffer[1024];
             ssize_t        rc;
 
-            log_info("loading session file: %s", view_info_name.c_str());
+            log_info("loading session file: %s", view_info_path.c_str());
             while ((rc = read(fd, buffer, sizeof(buffer))) > 0) {
                 yajl_parse(handle, buffer, rc);
             }
             yajl_complete_parse(handle);
         }
         yajl_free(handle);
+
+        bool log_changes = false, text_changes = false;
+
+        for (auto& lf : lnav_data.ld_active_files.fc_files) {
+            auto iter = session_data.sd_file_states.find(lf->get_filename());
+
+            if (iter == session_data.sd_file_states.end()) {
+                continue;
+            }
+
+            log_debug("found state for file: %s %d",
+                      lf->get_content_id().c_str(),
+                      iter->second.fs_is_visible);
+            lnav_data.ld_log_source.find_data(lf) | [iter](auto ld) {
+                ld->set_visibility(iter->second.fs_is_visible);
+            };
+            if (!iter->second.fs_is_visible) {
+                if (lf->get_format() != nullptr) {
+                    log_changes = true;
+                } else {
+                    text_changes = true;
+                }
+            }
+        }
+
+        if (log_changes) {
+            lnav_data.ld_log_source.text_filters_changed();
+        }
+        if (text_changes) {
+            lnav_data.ld_text_source.text_filters_changed();
+        }
     };
 }
 
@@ -889,7 +980,7 @@ static void save_user_bookmarks(
 
         if (meta_iter == bm_meta.end()) {
             if (sqlite3_bind_text(stmt, 5, "", 0, SQLITE_TRANSIENT) != SQLITE_OK) {
-                log_error("could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s",
                         sqlite3_errmsg(db));
                 return;
             }
@@ -903,7 +994,7 @@ static void save_user_bookmarks(
                                   meta_iter->second.bm_name.c_str(),
                                   meta_iter->second.bm_name.length(),
                                   SQLITE_TRANSIENT) != SQLITE_OK) {
-                log_error("could not bind part name -- %s\n",
+                log_error("could not bind part name -- %s",
                         sqlite3_errmsg(db));
                 return;
             }
@@ -913,7 +1004,7 @@ static void save_user_bookmarks(
                                   meta_iter->second.bm_comment.c_str(),
                                   meta_iter->second.bm_comment.length(),
                                   SQLITE_TRANSIENT) != SQLITE_OK) {
-                log_error("could not bind comment -- %s\n",
+                log_error("could not bind comment -- %s",
                           sqlite3_errmsg(db));
                 return;
             }
@@ -940,7 +1031,7 @@ static void save_user_bookmarks(
                                   tags.c_str(),
                                   tags.length(),
                                   SQLITE_TRANSIENT) != SQLITE_OK) {
-                log_error("could not bind tags -- %s\n",
+                log_error("could not bind tags -- %s",
                           sqlite3_errmsg(db));
                 return;
             }
@@ -948,7 +1039,7 @@ static void save_user_bookmarks(
 
         if (sqlite3_step(stmt) != SQLITE_DONE) {
             log_error(
-                    "could not execute bookmark insert statement -- %s\n",
+                    "could not execute bookmark insert statement -- %s",
                     sqlite3_errmsg(db));
             return;
         }
@@ -963,25 +1054,61 @@ static void save_user_bookmarks(
 static void save_time_bookmarks()
 {
     auto_mem<sqlite3, sqlite_close_wrapper> db;
-    auto db_path = dotlnav_path() / LOG_METADATA_NAME;
+    auto db_path = lnav::paths::dotlnav() / LOG_METADATA_NAME;
     auto_mem<char, sqlite3_free> errmsg;
     auto_mem<sqlite3_stmt> stmt(sqlite3_finalize);
 
-    if (sqlite3_open(db_path.str().c_str(), db.out()) != SQLITE_OK) {
-        log_error("unable to open bookmark DB -- %s\n", db_path.str().c_str());
+    if (sqlite3_open(db_path.c_str(), db.out()) != SQLITE_OK) {
+        log_error("unable to open bookmark DB -- %s", db_path.c_str());
         return;
     }
 
-    if (sqlite3_exec(db.in(), BOOKMARK_TABLE_DEF, NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        log_error("unable to make bookmark table -- %s\n", errmsg.in());
+    if (sqlite3_exec(db.in(), BOOKMARK_TABLE_DEF, nullptr, nullptr, errmsg.out()) != SQLITE_OK) {
+        log_error("unable to make bookmark table -- %s", errmsg.in());
         return;
     }
 
-    if (sqlite3_exec(db.in(), "BEGIN TRANSACTION", NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        log_error("unable to begin transaction -- %s\n", errmsg.in());
+    if (sqlite3_exec(db.in(), "BEGIN TRANSACTION", nullptr, nullptr, errmsg.out()) != SQLITE_OK) {
+        log_error("unable to begin transaction -- %s", errmsg.in());
         return;
     }
 
+    {
+        static const char *UPDATE_NETLOCS_STMT =
+            R"(REPLACE INTO recent_netlocs (netloc) VALUES (?))";
+
+        std::set<std::string> netlocs;
+
+        isc::to<tailer::looper&, services::remote_tailer_t>()
+            .send_and_wait([&netlocs](auto& tlooper) {
+                netlocs = tlooper.active_netlocs();
+            });
+
+        if (sqlite3_prepare_v2(db.in(),
+                               UPDATE_NETLOCS_STMT,
+                               -1,
+                               stmt.out(),
+                               nullptr) != SQLITE_OK) {
+            log_error("could not prepare recent_netlocs statement -- %s",
+                      sqlite3_errmsg(db));
+            return;
+        }
+
+        for (const auto& netloc : netlocs) {
+            bind_to_sqlite(stmt.in(), 1, netloc);
+
+            if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
+                log_error(
+                    "could not execute bookmark insert statement -- %s",
+                    sqlite3_errmsg(db));
+                return;
+            }
+
+            sqlite3_reset(stmt.in());
+        }
+        session_data.sd_recent_netlocs.insert(netlocs.begin(), netlocs.end());
+    }
+
     logfile_sub_source &lss = lnav_data.ld_log_source;
     bookmarks<content_line_t>::type &bm = lss.get_user_bookmarks();
 
@@ -993,7 +1120,7 @@ static void save_time_bookmarks()
                            stmt.out(),
                            nullptr) != SQLITE_OK) {
         log_error(
-                "could not prepare bookmark delete statement -- %s\n",
+                "could not prepare bookmark delete statement -- %s",
                 sqlite3_errmsg(db));
         return;
     }
@@ -1006,7 +1133,7 @@ static void save_time_bookmarks()
 
         if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
             log_error(
-                    "could not execute bookmark insert statement -- %s\n",
+                    "could not execute bookmark insert statement -- %s",
                     sqlite3_errmsg(db));
             return;
         }
@@ -1024,7 +1151,7 @@ static void save_time_bookmarks()
                            stmt.out(),
                            nullptr) != SQLITE_OK) {
         log_error(
-                "could not prepare bookmark replace statement -- %s\n",
+                "could not prepare bookmark replace statement -- %s",
                 sqlite3_errmsg(db));
         return;
     }
@@ -1036,29 +1163,30 @@ static void save_time_bookmarks()
              file_iter != lnav_data.ld_log_source.end();
              ++file_iter) {
             shared_ptr<logfile> lf = (*file_iter)->get_file();
-            content_line_t base_content_line;
 
-            if (lf == nullptr)
+            if (lf == nullptr) {
                 continue;
+            }
 
+            content_line_t base_content_line;
             base_content_line = lss.get_file_base_content_line(file_iter);
             base_content_line = content_line_t(
                 base_content_line + lf->size() - 1);
 
             if (!bind_line(db.in(), stmt.in(), base_content_line,
-                lnav_data.ld_session_time)) {
+                           lnav_data.ld_session_time)) {
                 continue;
             }
 
             if (sqlite3_bind_null(stmt.in(), 5) != SQLITE_OK) {
-                log_error("could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s",
                         sqlite3_errmsg(db.in()));
                 return;
             }
 
             if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
                 log_error(
-                        "could not execute bookmark insert statement -- %s\n",
+                        "could not execute bookmark insert statement -- %s",
                         sqlite3_errmsg(db));
                 return;
             }
@@ -1078,7 +1206,7 @@ static void save_time_bookmarks()
                            stmt.out(),
                            NULL) != SQLITE_OK) {
         log_error(
-                "could not prepare time_offset delete statement -- %s\n",
+                "could not prepare time_offset delete statement -- %s",
                 sqlite3_errmsg(db));
         return;
     }
@@ -1091,7 +1219,7 @@ static void save_time_bookmarks()
 
         if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
             log_error(
-                    "could not execute bookmark insert statement -- %s\n",
+                    "could not execute bookmark insert statement -- %s",
                     sqlite3_errmsg(db));
             return;
         }
@@ -1109,7 +1237,7 @@ static void save_time_bookmarks()
                            stmt.out(),
                            NULL) != SQLITE_OK) {
         log_error(
-                "could not prepare time_offset replace statement -- %s\n",
+                "could not prepare time_offset replace statement -- %s",
                 sqlite3_errmsg(db));
         return;
     }
@@ -1129,26 +1257,29 @@ static void save_time_bookmarks()
 
             base_content_line = lss.get_file_base_content_line(file_iter);
 
-            if (!bind_line(db.in(), stmt.in(), base_content_line,
-                lnav_data.ld_session_time)) {
+            if (!bind_values(stmt,
+                             lf->original_line_time(lf->begin()),
+                             lf->get_format()->get_name(),
+                             lf->get_content_id(),
+                             lnav_data.ld_session_time)) {
                 continue;
             }
 
             if (sqlite3_bind_null(stmt.in(), 5) != SQLITE_OK) {
-                log_error("could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s",
                         sqlite3_errmsg(db.in()));
                 return;
             }
 
             if (sqlite3_bind_null(stmt.in(), 6) != SQLITE_OK) {
-                log_error("could not bind log hash -- %s\n",
+                log_error("could not bind log hash -- %s",
                         sqlite3_errmsg(db.in()));
                 return;
             }
 
             if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
                 log_error(
-                        "could not execute bookmark insert statement -- %s\n",
+                        "could not execute bookmark insert statement -- %s",
                         sqlite3_errmsg(db));
                 return;
             }
@@ -1160,7 +1291,7 @@ static void save_time_bookmarks()
     for (auto &ls : lss) {
         logfile::iterator line_iter;
 
-        if (ls->get_file() == NULL)
+        if (ls->get_file() == nullptr)
             continue;
 
         shared_ptr<logfile> lf = ls->get_file();
@@ -1181,16 +1312,14 @@ static void save_time_bookmarks()
         bind_values(stmt.in(),
                     lf->original_line_time(line_iter),
                     lf->get_format()->get_name(),
-                    read_result.map([](auto sbr) {
-                        return hash_bytes(sbr.get_data(), sbr.length(), nullptr);
-                    }).unwrap(),
+                    lf->get_content_id(),
                     lnav_data.ld_session_time,
                     offset.tv_sec,
                     offset.tv_usec);
 
         if (sqlite3_step(stmt.in()) != SQLITE_DONE) {
             log_error(
-                    "could not execute bookmark insert statement -- %s\n",
+                    "could not execute bookmark insert statement -- %s",
                     sqlite3_errmsg(db));
             return;
         }
@@ -1198,13 +1327,18 @@ static void save_time_bookmarks()
         sqlite3_reset(stmt.in());
     }
 
-    if (sqlite3_exec(db.in(), "COMMIT", NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        log_error("unable to begin transaction -- %s\n", errmsg.in());
+    if (sqlite3_exec(db.in(), "COMMIT", nullptr, nullptr, errmsg.out()) != SQLITE_OK) {
+        log_error("unable to begin transaction -- %s", errmsg.in());
         return;
     }
 
-    if (sqlite3_exec(db.in(), BOOKMARK_LRU_STMT, NULL, NULL, errmsg.out()) != SQLITE_OK) {
-        log_error("unable to delete old bookmarks -- %s\n", errmsg.in());
+    if (sqlite3_exec(db.in(), BOOKMARK_LRU_STMT, nullptr, nullptr, errmsg.out()) != SQLITE_OK) {
+        log_error("unable to delete old bookmarks -- %s", errmsg.in());
+        return;
+    }
+
+    if (sqlite3_exec(db.in(), NETLOC_LRU_STMT, nullptr, nullptr, errmsg.out()) != SQLITE_OK) {
+        log_error("unable to delete old netlocs -- %s", errmsg.in());
         return;
     }
 }
@@ -1225,10 +1359,12 @@ static void save_session_with_id(const std::string session_id)
              lnav_data.ld_session_time,
              getppid());
 
-    auto view_file_name = dotlnav_path() / view_base_name;
-    auto view_file_tmp_name = view_file_name + ".tmp";
+    auto view_file_name = lnav::paths::dotlnav() / view_base_name;
+    auto view_file_tmp_name = view_file_name.string() + ".tmp";
 
-    if ((file = fopen(view_file_tmp_name.str().c_str(), "w")) == nullptr) {
+
+
+    if ((file = fopen(view_file_tmp_name.c_str(), "w")) == nullptr) {
         perror("Unable to open session file");
     }
     else if (nullptr == (handle = yajl_gen_alloc(nullptr))) {
@@ -1242,7 +1378,7 @@ static void save_session_with_id(const std::string session_id)
             yajlpp_map root_map(handle);
 
             root_map.gen("save-time");
-            root_map.gen((long long) time(NULL));
+            root_map.gen((long long) time(nullptr));
 
             root_map.gen("time-offset");
             root_map.gen(lnav_data.ld_log_source.is_time_offset_enabled());
@@ -1252,11 +1388,30 @@ static void save_session_with_id(const std::string session_id)
             {
                 yajlpp_array file_list(handle);
 
-                for (auto &ld_file_name : lnav_data.ld_file_names) {
+                for (auto &ld_file_name : lnav_data.ld_active_files.fc_file_names) {
                     file_list.gen(ld_file_name.first);
                 }
             }
 
+            root_map.gen("file-states");
+
+            {
+                yajlpp_map file_states(handle);
+
+                for (auto &lf : lnav_data.ld_active_files.fc_files) {
+                    auto ld_opt = lnav_data.ld_log_source.find_data(lf);
+
+                    file_states.gen(lf->get_filename());
+
+                    {
+                        yajlpp_map file_state(handle);
+
+                        file_state.gen("visible");
+                        file_state.gen(!ld_opt || ld_opt.value()->ld_visible);
+                    }
+                }
+            }
+
             root_map.gen("views");
 
             {
@@ -1282,16 +1437,19 @@ static void save_session_with_id(const std::string session_id)
                     }
 
                     view_map.gen("search");
-                    view_map.gen(lnav_data.ld_views[lpc].get_last_search());
+                    view_map.gen(lnav_data.ld_views[lpc].get_current_search());
 
                     view_map.gen("word_wrap");
                     view_map.gen(tc.get_word_wrap());
 
-                    text_sub_source *tss = tc.get_sub_source();
+                    auto tss = tc.get_sub_source();
                     if (tss == nullptr) {
                         continue;
                     }
 
+                    view_map.gen("filtering");
+                    view_map.gen(tss->tss_apply_filters);
+
                     filter_stack &fs = tss->get_filters();
 
                     view_map.gen("commands");
@@ -1311,29 +1469,25 @@ static void save_session_with_id(const std::string session_id)
                         }
                     }
 
-                    textview_curses::highlight_map_t &hmap =
-                        lnav_data.ld_views[lpc].get_highlights();
-                    textview_curses::highlight_map_t::iterator hl_iter;
+                    auto &hmap = lnav_data.ld_views[lpc].get_highlights();
 
-                    for (hl_iter = hmap.begin();
-                         hl_iter != hmap.end();
-                         ++hl_iter) {
-                        if (hl_iter->first.first != highlight_source_t::INTERACTIVE) {
+                    for (auto & hl : hmap) {
+                        if (hl.first.first != highlight_source_t::INTERACTIVE) {
                             continue;
                         }
-                        cmd_array.gen("highlight " + hl_iter->first.second);
+                        cmd_array.gen("highlight " + hl.first.second);
                     }
 
                     if (lpc == LNV_LOG) {
-                        for (auto format : log_format::get_root_formats()) {
-                            auto *elf = dynamic_cast<external_log_format *>(format);
+                        for (const auto& format : log_format::get_root_formats()) {
+                            auto *elf = dynamic_cast<external_log_format *>(format.get());
 
                             if (elf == nullptr) {
                                 continue;
                             }
 
                             for (const auto& vd : elf->elf_value_defs) {
-                                if (!vd.second->vd_user_hidden) {
+                                if (!vd.second->vd_meta.lvm_user_hidden) {
                                     continue;
                                 }
 
@@ -1361,6 +1515,11 @@ static void save_session_with_id(const std::string session_id)
                             cmd_array.gen("hide-lines-after "
                                           + string(max_time_str));
                         }
+
+                        auto mark_expr = lss.get_sql_marker_text();
+                        if (!mark_expr.empty()) {
+                            cmd_array.gen("mark-expr " + mark_expr);
+                        }
                     }
                 }
             }
@@ -1371,27 +1530,29 @@ static void save_session_with_id(const std::string session_id)
 
         fclose(file.release());
 
-        log_perror(rename(view_file_tmp_name.str().c_str(),
-            view_file_name.str().c_str()));
+        log_perror(rename(view_file_tmp_name.c_str(),
+            view_file_name.c_str()));
 
-        log_info("Saved session: %s", view_file_name.str().c_str());
+        log_info("Saved session: %s", view_file_name.c_str());
     }
 }
 
 void save_session()
 {
+    log_debug("BEGIN save_session");
     save_time_bookmarks();
 
     const auto opt_session_id = compute_session_id();
     opt_session_id | [](auto &session_id) {
         save_session_with_id(session_id);
     };
-    for (const auto pair : lnav_data.ld_session_id) {
+    for (const auto& pair : lnav_data.ld_session_id) {
         if (opt_session_id && pair.first == opt_session_id.value()) {
             continue;
         }
         save_session_with_id(pair.first);
     }
+    log_debug("END save_session");
 }
 
 void reset_session()
@@ -1401,6 +1562,7 @@ void reset_session()
     save_session();
 
     lnav_data.ld_session_time = time(nullptr);
+    session_data.sd_file_states.clear();
 
     for (auto &tc : lnav_data.ld_views) {
         auto &hmap = tc.get_highlights();
@@ -1415,15 +1577,15 @@ void reset_session()
         }
     }
 
-    for (auto ld : lnav_data.ld_log_source) {
-        shared_ptr<logfile> lf = ld->get_file();
-
-        lf->clear_time_offset();
+    for (const auto& lf : lnav_data.ld_active_files.fc_files) {
+        lf->reset_state();
     }
 
     lnav_data.ld_log_source.set_marked_only(false);
     lnav_data.ld_log_source.clear_min_max_log_times();
     lnav_data.ld_log_source.set_min_log_level(LEVEL_UNKNOWN);
+    lnav_data.ld_log_source.set_sql_filter("", nullptr);
+    lnav_data.ld_log_source.set_sql_marker("", nullptr);
 
     lnav_data.ld_log_source.get_user_bookmark_metadata().clear();
 
@@ -1434,6 +1596,7 @@ void reset_session()
             continue;
         }
         tss->get_filters().clear_filters();
+        tss->tss_apply_filters = true;
         tss->text_filters_changed();
         tss->text_clear_marks(&textview_curses::BM_USER);
         tc.get_bookmarks()[&textview_curses::BM_USER].clear();
@@ -1443,15 +1606,16 @@ void reset_session()
     }
 
     lnav_data.ld_filter_view.reload_data();
-    for (auto format : log_format::get_root_formats()) {
-        auto *elf = dynamic_cast<external_log_format *>(format);
+    lnav_data.ld_files_view.reload_data();
+    for (const auto& format : log_format::get_root_formats()) {
+        auto *elf = dynamic_cast<external_log_format *>(format.get());
 
         if (elf == nullptr) {
             continue;
         }
 
         for (const auto &vd : elf->elf_value_defs) {
-            vd.second->vd_user_hidden = false;
+            vd.second->vd_meta.lvm_user_hidden = false;
         }
     }
 }
