@@ -139,3 +139,20 @@ void shared_buffer_ref::disown()
     this->sb_data = nullptr;
     this->sb_length = 0;
 }
+
+void shared_buffer_ref::copy_ref(const shared_buffer_ref &other)
+{
+    if (other.sb_data == nullptr) {
+        this->sb_owner = nullptr;
+        this->sb_data = nullptr;
+        this->sb_length = 0;
+    }
+    else if (other.sb_owner != nullptr) {
+        this->share(*other.sb_owner, other.sb_data, other.sb_length);
+    } else {
+        this->sb_owner = nullptr;
+        this->sb_data = (char *)malloc(other.sb_length);
+        memcpy(this->sb_data, other.sb_data, other.sb_length);
+        this->sb_length = other.sb_length;
+    }
+}
