@@ -34,15 +34,20 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <string.h>
-#include <pcrecpp.h>
 
+#include <regex>
+#include <algorithm>
 #include <vector>
 
 #include "auto_mem.hh"
 #include "sql_util.hh"
+#include "base/injector.hh"
+#include "base/string_util.hh"
 #include "base/lnav_log.hh"
-#include "lnav_util.hh"
+#include "base/time_util.hh"
 #include "pcrepp/pcrepp.hh"
+#include "readline_curses.hh"
+#include "bound_tags.hh"
 #include "sqlite-extension-func.hh"
 
 using namespace std;
@@ -251,7 +256,7 @@ const char *sql_function_names[] = {
     "julianday(",
     "strftime(",
 
-    NULL
+    nullptr
 };
 
 multimap<std::string, help_text *> sqlite_function_help;
@@ -266,6 +271,9 @@ static int handle_db_list(void *ptr,
     smc = (struct sqlite_metadata_callbacks *)ptr;
 
     smc->smc_db_list[colvalues[1]] = std::vector<std::string>();
+    if (!smc->smc_database_list) {
+        return 0;
+    }
 
     return smc->smc_database_list(ptr, ncols, colvalues, colnames);
 }
@@ -282,7 +290,10 @@ static int handle_table_list(void *ptr,
 {
     struct table_list_data *tld = (struct table_list_data *)ptr;
 
-    (*tld->tld_iter)->second.push_back(colvalues[0]);
+    (*tld->tld_iter)->second.emplace_back(colvalues[0]);
+    if (!tld->tld_callbacks->smc_table_list) {
+        return 0;
+    }
 
     return tld->tld_callbacks->smc_table_list(tld->tld_callbacks,
                                               ncols,
@@ -295,14 +306,16 @@ int walk_sqlite_metadata(sqlite3 *db, struct sqlite_metadata_callbacks &smc)
     auto_mem<char, sqlite3_free> errmsg;
     int retval;
 
-    retval = sqlite3_exec(db,
-                          "pragma collation_list",
-                          smc.smc_collation_list,
-                          &smc,
-                          errmsg.out());
-    if (retval != SQLITE_OK) {
-        log_error("could not get collation list -- %s", errmsg.in());
-        return retval;
+    if (smc.smc_collation_list) {
+        retval = sqlite3_exec(db,
+                              "pragma collation_list",
+                              smc.smc_collation_list,
+                              &smc,
+                              errmsg.out());
+        if (retval != SQLITE_OK) {
+            log_error("could not get collation list -- %s", errmsg.in());
+            return retval;
+        }
     }
 
     retval = sqlite3_exec(db,
@@ -342,39 +355,44 @@ int walk_sqlite_metadata(sqlite3 *db, struct sqlite_metadata_callbacks &smc)
             std::string &table_name = *table_iter;
 
             table_query = sqlite3_mprintf(
-                "pragma %Q.table_info(%Q)",
+                "pragma %Q.table_xinfo(%Q)",
                 iter->first.c_str(),
                 table_name.c_str());
-            if (table_query == NULL) {
+            if (table_query == nullptr) {
                 return SQLITE_NOMEM;
             }
 
-            retval = sqlite3_exec(db,
-                                  table_query,
-                                  smc.smc_table_info,
-                                  &smc,
-                                  errmsg.out());
-            if (retval != SQLITE_OK) {
-                log_error("could not get table info -- %s", errmsg.in());
-                return retval;
+            if (smc.smc_table_info) {
+                retval = sqlite3_exec(db,
+                                      table_query,
+                                      smc.smc_table_info,
+                                      &smc,
+                                      errmsg.out());
+                if (retval != SQLITE_OK) {
+                    log_error("could not get table info -- %s", errmsg.in());
+                    return retval;
+                }
             }
 
             table_query = sqlite3_mprintf(
                 "pragma %Q.foreign_key_list(%Q)",
                 iter->first.c_str(),
                 table_name.c_str());
-            if (table_query == NULL) {
+            if (table_query == nullptr) {
                 return SQLITE_NOMEM;
             }
 
-            retval = sqlite3_exec(db,
-                                  table_query,
-                                  smc.smc_foreign_key_list,
-                                  &smc,
-                                  errmsg.out());
-            if (retval != SQLITE_OK) {
-                log_error("could not get foreign key list -- %s", errmsg.in());
-                return retval;
+            if (smc.smc_foreign_key_list) {
+                retval = sqlite3_exec(db,
+                                      table_query,
+                                      smc.smc_foreign_key_list,
+                                      &smc,
+                                      errmsg.out());
+                if (retval != SQLITE_OK) {
+                    log_error("could not get foreign key list -- %s",
+                              errmsg.in());
+                    return retval;
+                }
             }
         }
     }
@@ -447,7 +465,8 @@ void dump_sqlite_schema(sqlite3 *db, std::string &schema_out)
         schema_table_list,
         schema_table_info,
         schema_foreign_key_list,
-        &schema_out
+        &schema_out,
+        {}
     };
 
     walk_sqlite_metadata(db, schema_sql_meta_callbacks);
@@ -455,7 +474,7 @@ void dump_sqlite_schema(sqlite3 *db, std::string &schema_out)
 
 void attach_sqlite_db(sqlite3 *db, const std::string &filename)
 {
-    static pcrecpp::RE db_name_converter("[^\\w]");
+    static const std::regex db_name_converter("[^\\w]");
 
     auto_mem<sqlite3_stmt> stmt(sqlite3_finalize);
 
@@ -487,7 +506,7 @@ void attach_sqlite_db(sqlite3 *db, const std::string &filename)
         db_name = filename.substr(base_start + 1);
     }
 
-    db_name_converter.GlobalReplace("_", &db_name);
+    db_name = std::regex_replace(db_name, db_name_converter, "_");
 
     if (sqlite3_bind_text(stmt.in(), 2,
                           db_name.c_str(), db_name.length(),
@@ -504,13 +523,13 @@ void attach_sqlite_db(sqlite3 *db, const std::string &filename)
     }
 }
 
-ssize_t sql_strftime(char *buffer, size_t buffer_size, time_t time, int millis,
+ssize_t sql_strftime(char *buffer, size_t buffer_size, lnav::time64_t tim, int millis,
     char sep)
 {
     struct tm gmtm;
     int year, month, index = 0;
 
-    secs2tm(&time, &gmtm);
+    secs2tm(tim, &gmtm);
     year = gmtm.tm_year + 1900;
     month = gmtm.tm_mon + 1;
     buffer[index++] = '0' + ((year / 1000) % 10);
@@ -642,7 +661,7 @@ string sql_safe_ident(const string_fragment &ident)
         char ch = retval[lpc];
 
         if (isalnum(ch) || ch == '_') {
-            retval[lpc] = ch;
+            retval[lpc] = tolower(ch);
         } else {
             retval[lpc] = '_';
         }
@@ -688,7 +707,7 @@ void sql_compile_script(sqlite3 *db,
                 log_error("unable to allocate error message");
                 break;
             }
-            errors.push_back(full_msg.in());
+            errors.emplace_back(full_msg.in());
             break;
         } else if (script == tail) {
             break;
@@ -727,7 +746,7 @@ void sql_execute_script(sqlite3 *db,
                     sqlite3_bind_text(stmt, lpc + 1,
                                       iter->second.c_str(), -1,
                                       SQLITE_TRANSIENT);
-                } else if ((env_value = getenv(&name[1])) != NULL) {
+                } else if ((env_value = getenv(&name[1])) != nullptr) {
                     sqlite3_bind_text(stmt, lpc + 1,
                                       env_value, -1,
                                       SQLITE_TRANSIENT);
@@ -763,7 +782,7 @@ void sql_execute_script(sqlite3 *db,
                     const char *errmsg;
 
                     errmsg = sqlite3_errmsg(db);
-                    errors.push_back(errmsg);
+                    errors.emplace_back(errmsg);
                     break;
                 }
             }
@@ -799,26 +818,27 @@ static struct {
         { SQLITE_INTEGER, "", "123" },
         { SQLITE_FLOAT, "", "123.0" },
         { SQLITE_TEXT, "ipaddress", "127.0.0.1" },
-
-        { SQLITE_NULL }
 };
 
 int guess_type_from_pcre(const string &pattern, std::string &collator)
 {
     try {
-        pcrepp re(pattern.c_str());
+        pcrepp re(pattern);
         vector<int> matches;
         int retval = SQLITE3_TEXT;
+        int index = 0;
 
         collator.clear();
-        for (int lpc = 0; TYPE_TEST_VALUE[lpc].sqlite_type != SQLITE_NULL; lpc++) {
+        for (const auto& test_value : TYPE_TEST_VALUE) {
             pcre_context_static<30> pc;
-            pcre_input pi(TYPE_TEST_VALUE[lpc].sample);
+            pcre_input pi(test_value.sample);
 
             if (re.match(pc, pi, PCRE_ANCHORED) &&
                 pc[0]->c_begin == 0 && pc[0]->length() == (int) pi.pi_length) {
-                matches.push_back(lpc);
+                matches.push_back(index);
             }
+
+            index += 1;
         }
 
         if (matches.size() == 1) {
@@ -869,6 +889,7 @@ string sql_keyword_re(void)
     return retval;
 }
 
+string_attr_type SQL_COMMAND_ATTR("sql_command");
 string_attr_type SQL_KEYWORD_ATTR("sql_keyword");
 string_attr_type SQL_IDENTIFIER_ATTR("sql_ident");
 string_attr_type SQL_FUNCTION_ATTR("sql_func");
@@ -880,16 +901,16 @@ string_attr_type SQL_GARBAGE_ATTR("sql_garbage");
 
 void annotate_sql_statement(attr_line_t &al)
 {
-    static string keyword_re_str =
-        R"(\A)" + sql_keyword_re() + R"(|\.schema|\.msgformats)";
+    static string keyword_re_str = R"(\A)" + sql_keyword_re();
 
     static struct {
         pcrepp re;
         string_attr_type_t type;
     } PATTERNS[] = {
+        { pcrepp{R"(^(\.\w+))"}, &SQL_COMMAND_ATTR },
         { pcrepp{R"(\A,)"}, &SQL_COMMA_ATTR },
         { pcrepp{R"(\A\(|\A\))"}, &SQL_PAREN_ATTR },
-        { pcrepp{keyword_re_str.c_str(), PCRE_CASELESS}, &SQL_KEYWORD_ATTR },
+        { pcrepp{keyword_re_str, PCRE_CASELESS}, &SQL_KEYWORD_ATTR },
         { pcrepp{R"(\A'[^']*('(?:'[^']*')*|$))"}, &SQL_STRING_ATTR },
         { pcrepp{R"(\A(\$?\b[a-z_]\w*)|\"([^\"]+)\"|\[([^\]]+)])", PCRE_CASELESS}, &SQL_IDENTIFIER_ATTR },
         { pcrepp{R"(\A(\*|<|>|=|!|\-|\+|\|\|))"}, &SQL_OPERATOR_ATTR },
@@ -900,14 +921,14 @@ void annotate_sql_statement(attr_line_t &al)
 
     pcre_context_static<30> pc;
     pcre_input pi(al.get_string());
-    string &line = al.get_string();
-    string_attrs_t &sa = al.get_attrs();
+    auto &line = al.get_string();
+    auto &sa = al.get_attrs();
 
     while (pi.pi_next_offset < line.length()) {
         if (ws_pattern.match(pc, pi, PCRE_ANCHORED)) {
             continue;
         }
-        for (auto &pat : PATTERNS) {
+        for (const auto &pat : PATTERNS) {
             if (pat.re.match(pc, pi, PCRE_ANCHORED)) {
                 pcre_context::capture_t *cap = pc.all();
                 struct line_range lr(cap->c_begin, cap->c_end);
@@ -973,8 +994,24 @@ vector<const help_text *> find_sql_help_for_line(const attr_line_t &al, size_t x
 
     x = al.nearest_text(x);
 
+    {
+        auto sa_opt = get_string_attr(al.get_attrs(), &SQL_COMMAND_ATTR);
+
+        if (sa_opt) {
+            auto sql_cmd_map = injector::get<
+                readline_context::command_map_t*, sql_cmd_map_tag>();
+            auto cmd_name = al.get_substring((*sa_opt)->sa_range);
+            auto cmd_iter = sql_cmd_map->find(cmd_name);
+
+            if (cmd_iter != sql_cmd_map->end()) {
+                return {&cmd_iter->second->c_help};
+            }
+        }
+    }
+
     vector<string> kw;
     auto iter = rfind_string_attr_if(sa, x, [&al, &name, &kw, x](auto sa) {
+
         if (sa.sa_type != &SQL_FUNCTION_ATTR &&
             sa.sa_type != &SQL_KEYWORD_ATTR) {
             return false;
