@@ -31,17 +31,27 @@
 
 #include "config.h"
 
-#include <stdio.h>
 #include <assert.h>
 
-#include "lnav_util.hh"
+#include "base/string_util.hh"
 #include "base/lnav_log.hh"
+#include "auto_mem.hh"
 #include "sql_util.hh"
 
 #include "sqlite-extension-func.hh"
 
 using namespace std;
 
+extern "C" {
+struct sqlite3_api_routines;
+
+int sqlite3_series_init(
+    sqlite3 *db,
+    char **pzErrMsg,
+    const sqlite3_api_routines *pApi
+);
+}
+
 sqlite_registration_func_t sqlite_registration_funcs[] = {
     common_extension_functions,
     state_extension_functions,
@@ -60,7 +70,13 @@ int register_sqlite_funcs(sqlite3 *db, sqlite_registration_func_t *reg_funcs)
 
     assert(db != nullptr);
     assert(reg_funcs != nullptr);
-    
+
+    {
+        auto_mem<char> errmsg(sqlite3_free);
+
+        sqlite3_series_init(db, errmsg.out(), nullptr);
+    }
+
     for (lpc = 0; reg_funcs[lpc]; lpc++) {
         struct FuncDef *basic_funcs = nullptr;
         struct FuncDefAgg *agg_funcs = nullptr;
@@ -99,7 +115,7 @@ int register_sqlite_funcs(sqlite3 *db, sqlite_registration_func_t *reg_funcs)
                                     agg_funcs[i].nArg,
                                     SQLITE_UTF8,
                                     (void *) &agg_funcs[i],
-                                    0,
+                                    nullptr,
                                     agg_funcs[i].xStep,
                                     agg_funcs[i].xFinalize);
 
@@ -680,6 +696,27 @@ int register_sqlite_funcs(sqlite3 *db, sqlite_registration_func_t *reg_funcs)
                 "SELECT total(ex_duration) FROM lnav_example_log"
             }),
 
+        help_text("generate_series",
+                  "A table-valued-function that returns the whole numbers "
+                  "between a lower and upper bound, inclusive")
+            .sql_table_valued_function()
+            .with_parameter({"start", "The starting point of the series"})
+            .with_parameter({"stop", "The stopping point of the series"})
+            .with_parameter(help_text("step", "The increment between each value")
+                                .optional())
+            .with_result({"value", "The number in the series"})
+            .with_example({
+                "To generate the numbers in the range [10, 14]",
+                "SELECT value FROM generate_series(10, 14)"
+            })
+            .with_example({
+                "To generate every other number in the range [10, 14]",
+                "SELECT value FROM generate_series(10, 14, 2)"
+            })
+            .with_example({
+                "To count down from five to 1",
+                "SELECT value FROM generate_series(1, 5, -1)"
+            })
     };
 
     for (auto &ht : builtin_funcs) {
@@ -959,6 +996,70 @@ int register_sqlite_funcs(sqlite3 *db, sqlite_registration_func_t *reg_funcs)
                 "SELECT CAST(1.23 AS INTEGER)"
             }),
 
+        help_text("expr",
+                  "Match an expression against a glob pattern.")
+            .sql_infix()
+            .with_parameter(help_text("NOT")
+                                .optional())
+            .with_parameter(help_text("pattern", "The glob pattern to match against.")
+                                .with_flag_name("GLOB"))
+            .with_example({
+                "To check if a value matches the pattern '*.log'",
+                "SELECT 'foobar.log' GLOB '*.log'"
+            }),
+
+        help_text("expr",
+                  "Match an expression against a text pattern.")
+            .sql_infix()
+            .with_parameter(help_text("NOT")
+                                .optional())
+            .with_parameter(help_text("pattern", "The pattern to match against.")
+                                .with_flag_name("LIKE"))
+            .with_example({
+                "To check if a value matches the pattern 'Hello, %!'",
+                "SELECT 'Hello, World!' LIKE 'Hello, %!'"
+            }),
+
+        help_text("expr",
+                  "Match an expression against a regular expression.")
+            .sql_infix()
+            .with_parameter(help_text("NOT")
+                                .optional())
+            .with_parameter(help_text("pattern", "The regular expression to match against.")
+                                .with_flag_name("REGEXP"))
+            .with_example({
+                "To check if a value matches the pattern 'file-\\d+'",
+                "SELECT 'file-23' REGEXP 'file-\\d+'"
+            }),
+
+        help_text("expr",
+                  "Assign a collating sequence to the expression.")
+            .sql_infix()
+            .with_parameter(help_text("collation-name", "The name of the collator.")
+                                .with_flag_name("COLLATE"))
+            .with_example({
+                "To change the collation method for string comparisons",
+                "SELECT ('a2' < 'a10'), ('a2' < 'a10' COLLATE naturalnocase)"
+            }),
+
+        help_text("expr",
+                  "Test if an expression is between two values.")
+            .sql_infix()
+            .with_parameter(help_text("NOT")
+                                .optional())
+            .with_parameter(help_text("low", "The low point")
+                                .with_flag_name("BETWEEN"))
+            .with_parameter(help_text("hi", "The high point")
+                                .with_flag_name("AND"))
+            .with_example({
+                "To check if 3 is between 5 and 10",
+                "SELECT 3 BETWEEN 5 AND 10"
+            })
+            .with_example({
+                "To check if 10 is between 5 and 10",
+                "SELECT 10 BETWEEN 5 AND 10"
+            }),
+
         help_text("OVER", "Executes the preceding function over a window")
             .sql_keyword()
             .with_parameter({"window-name", "The name of the window definition"}),
