@@ -32,6 +32,7 @@
 #include "config.h"
 
 #include <vector>
+#include <algorithm>
 
 #include "statusview_curses.hh"
 
@@ -44,23 +45,45 @@ void status_field::set_value(std::string value)
     sa.clear();
 
     scrub_ansi_string(value, sa);
-
     this->sf_value.with_string(value);
 
     if (this->sf_cylon) {
-        struct line_range lr(this->sf_cylon_pos, this->sf_width);
-        view_colors &vc = view_colors::singleton();
-
-        sa.emplace_back(lr, &view_curses::VC_STYLE,
-            vc.attrs_for_role(view_colors::VCR_ACTIVE_STATUS) | A_REVERSE);
-
-        this->sf_cylon_pos += 1;
-        if (this->sf_cylon_pos > this->sf_width) {
-            this->sf_cylon_pos = 0;
-        }
+        this->do_cylon();
     }
 }
 
+void status_field::do_cylon()
+{
+    string_attrs_t &sa = this->sf_value.get_attrs();
+
+    remove_string_attr(sa, &view_curses::VC_STYLE);
+
+    struct line_range lr(this->sf_cylon_pos, this->sf_width);
+    view_colors &vc = view_colors::singleton();
+
+    sa.emplace_back(lr, &view_curses::VC_STYLE,
+                    vc.attrs_for_role(view_colors::VCR_ACTIVE_STATUS) | A_REVERSE);
+
+    this->sf_cylon_pos += 1;
+    if (this->sf_cylon_pos > this->sf_width) {
+        this->sf_cylon_pos = 0;
+    }
+}
+
+void status_field::set_stitch_value(view_colors::role_t left,
+                                    view_colors::role_t right)
+{
+    string_attrs_t &sa = this->sf_value.get_attrs();
+    struct line_range lr(0, 1);
+
+    this->sf_value.get_string() = "::";
+    sa.clear();
+    sa.emplace_back(lr, &view_curses::VC_ROLE, left);
+    lr.lr_start = 1;
+    lr.lr_end   = 2;
+    sa.emplace_back(lr, &view_curses::VC_ROLE, right);
+}
+
 void statusview_curses::do_update()
 {
     int           top, attrs, field, field_count, left = 0, right;
@@ -100,9 +123,15 @@ void statusview_curses::do_update()
                     if (sa.sa_type == &view_curses::VC_STYLE) {
                         sa.sa_value.sav_int &= ~(A_REVERSE | A_COLOR);
                     } else if (sa.sa_type == &view_curses::VC_ROLE) {
-                        sa.sa_value.sav_int = view_colors::VCR_NONE;
+                        if (sa.sa_value.sav_int == view_colors::VCR_ALERT_STATUS) {
+                            sa.sa_value.sav_int = view_colors::VCR_INACTIVE_ALERT_STATUS;
+                        } else {
+                            sa.sa_value.sav_int = view_colors::VCR_NONE;
+                        }
                     }
                 }
+            } else if (sf.is_cylon()) {
+                sf.do_cylon();
             }
             if (sf.get_left_pad() > 0) {
                 val.insert(0, sf.get_left_pad(), ' ');
@@ -133,11 +162,20 @@ void statusview_curses::do_update()
                 }
             }
 
+            auto default_role = sf.get_role();
+            if (!this->sc_enabled) {
+                if (default_role == view_colors::VCR_ALERT_STATUS) {
+                    default_role = view_colors::VCR_INACTIVE_ALERT_STATUS;
+                } else {
+                    default_role = view_colors::VCR_INACTIVE_STATUS;
+                }
+            }
+
             mvwattrline(this->sc_window,
                         top, x,
                         val,
                         lr,
-                        this->sc_enabled ? sf.get_role() : view_colors::VCR_INACTIVE_STATUS);
+                        default_role);
         }
     }
     wmove(this->sc_window, top + 1, 0);
@@ -174,7 +212,7 @@ void statusview_curses::window_change()
         remaining = 0;
     }
 
-    stable_sort(begin(resizable), end(resizable), [](auto l, auto r) {
+    std::stable_sort(begin(resizable), end(resizable), [](auto l, auto r) {
         return r->get_share() < l->get_share();
     });
     for (auto sf : resizable) {
