@@ -45,7 +45,7 @@ static pcre *xpcre_compile(const char *pattern, int options = 0)
                                options,
                                &errptr,
                                &eoff,
-                               NULL)) == NULL) {
+                               nullptr)) == nullptr) {
         fprintf(stderr, "internal error: failed to compile -- %s\n", pattern);
         fprintf(stderr, "internal error: %s\n", errptr);
 
@@ -55,12 +55,7 @@ static pcre *xpcre_compile(const char *pattern, int options = 0)
     return retval;
 }
 
-static highlighter static_highlighter(const string &regex) {
-    return highlighter(xpcre_compile(regex.c_str()))
-        .with_attrs(view_colors::singleton().attrs_for_ident(regex));
-}
-
-void setup_highlights(textview_curses::highlight_map_t &hm)
+void setup_highlights(highlight_map_t &hm)
 {
     hm[{highlight_source_t::INTERNAL, "python"}] = highlighter(xpcre_compile(
         "(?:"
@@ -251,6 +246,7 @@ void setup_highlights(textview_curses::highlight_map_t &hm)
         ")"))
         .with_nestable(false)
         .with_text_format(text_format_t::TF_C_LIKE)
+        .with_text_format(text_format_t::TF_JAVA)
         .with_role(view_colors::VCR_KEYWORD);
 
     hm[{highlight_source_t::INTERNAL, "sql.0.comment"}] = highlighter(xpcre_compile(
@@ -399,12 +395,11 @@ void setup_highlights(textview_curses::highlight_map_t &hm)
         "(?:java|a|o|so|c|cc|cpp|cxx|h|hh|hpp|hxx|py|pyc|rb):"
         "\\d+"))
         .with_role(view_colors::VCR_FILE);
-    hm[{highlight_source_t::INTERNAL, "xml"}] = static_highlighter("<(/?[^ >=]+)[^>]*>");
     hm[{highlight_source_t::INTERNAL, "1.stringd"}] = highlighter(xpcre_compile(
-        "\"(?:\\\\.|[^\"])*\""))
+        R"("(?:\\.|[^"])*")"))
         .with_role(view_colors::VCR_STRING);
     hm[{highlight_source_t::INTERNAL, "1.strings"}] = highlighter(xpcre_compile(
-        "(?<![A-WY-Za-qstv-z])\'(?:\\\\.|[^'])*\'"))
+        R"((?<![A-WY-Za-qstv-z])'(?:\\.|[^'])*')"))
         .with_role(view_colors::VCR_STRING);
     hm[{highlight_source_t::INTERNAL, "1.stringb"}] = highlighter(xpcre_compile(
         "`(?:\\\\.|[^`])*`"))
@@ -418,12 +413,12 @@ void setup_highlights(textview_curses::highlight_map_t &hm)
     hm[{highlight_source_t::INTERNAL, "diffs"}] = highlighter(xpcre_compile(
         "^\\@@ .*"))
         .with_role(view_colors::VCR_DIFF_SECTION);
-    hm[{highlight_source_t::INTERNAL, "ip"}] = static_highlighter("\\d+\\.\\d+\\.\\d+\\.\\d+");
     hm[{highlight_source_t::INTERNAL, "0.comment"}] = highlighter(xpcre_compile(
-        "(?<=[\\s;])//.*|/\\*.*\\*/|\\(\\*.*\\*\\)|^#.*|\\s+#.*|dnl.*"))
+        R"((?<=[\s;])//.*|/\*.*\*/|\(\*.*\*\)|^#.*|\s+#.*|dnl.*)"))
         .with_role(view_colors::VCR_COMMENT);
-    hm[{highlight_source_t::INTERNAL, "javadoc"}] = static_highlighter(
-        "@(?:author|deprecated|exception|file|param|return|see|since|throws|todo|version)");
+    hm[{highlight_source_t::INTERNAL, "javadoc"}] = highlighter(xpcre_compile(
+        "@(?:author|deprecated|exception|file|param|return|see|since|throws|todo|version)"))
+        .with_role(view_colors::VCR_DOC_DIRECTIVE);
     hm[{highlight_source_t::INTERNAL, "var"}] = highlighter(xpcre_compile(
         "(?:"
         "(?:var\\s+)?([\\-\\w]+)\\s*[!=+\\-*/|&^]?=|"
@@ -446,10 +441,12 @@ void setup_highlights(textview_curses::highlight_map_t &hm)
         "\\b[A-Z_][A-Z0-9_]+\\b"))
         .with_nestable(false)
         .with_text_format(text_format_t::TF_C_LIKE)
+        .with_text_format(text_format_t::TF_JAVA)
         .with_role(view_colors::VCR_SYMBOL);
     hm[{highlight_source_t::INTERNAL, "num"}] = highlighter(xpcre_compile(
         R"(\b-?(?:\d+|0x[a-zA-Z0-9]+)\b)"))
         .with_nestable(false)
         .with_text_format(text_format_t::TF_C_LIKE)
+        .with_text_format(text_format_t::TF_JAVA)
         .with_role(view_colors::VCR_NUMBER);
 }
