@@ -32,15 +32,15 @@
 #include <vector>
 #include <algorithm>
 
-#include <pcrecpp.h>
-
-#include "pcrepp/pcrepp.hh"
-#include "lnav_util.hh"
+#include "base/time_util.hh"
+#include "shlex.hh"
+#include "fmt/format.h"
 #include "data_parser.hh"
 #include "ansi_scrubber.hh"
 #include "log_format.hh"
 #include "textview_curses.hh"
 #include "view_curses.hh"
+#include "lnav_config.hh"
 
 using namespace std;
 
@@ -75,9 +75,9 @@ text_filter::revert_to_last(logfile_filter_state &lfs, size_t rollback_size)
 
 void text_filter::add_line(
         logfile_filter_state &lfs, logfile::const_iterator ll, shared_buffer_ref &line) {
-    bool match_state = this->matches(*lfs.tfs_logfile, *ll, line);
+    bool match_state = this->matches(*lfs.tfs_logfile, ll, line);
 
-    if (!ll->is_continued()) {
+    if (ll->is_message()) {
         this->end_of_message(lfs);
     }
 
@@ -89,7 +89,7 @@ void text_filter::end_of_message(logfile_filter_state &lfs)
 {
     uint32_t mask = 0;
 
-    mask = ((uint32_t) lfs.tfs_message_matched[this->lf_index] ? 1U : 0) << this->lf_index;
+    mask = ((uint32_t) 1U << this->lf_index);
 
     for (size_t lpc = 0; lpc < lfs.tfs_lines_for_message[this->lf_index]; lpc++) {
         require(lfs.tfs_filter_count[this->lf_index] <=
@@ -97,7 +97,11 @@ void text_filter::end_of_message(logfile_filter_state &lfs)
 
         size_t line_number = lfs.tfs_filter_count[this->lf_index];
 
-        lfs.tfs_mask[line_number] |= mask;
+        if (lfs.tfs_message_matched[this->lf_index]) {
+            lfs.tfs_mask[line_number] |= mask;
+        } else {
+            lfs.tfs_mask[line_number] &= ~mask;
+        }
         lfs.tfs_filter_count[this->lf_index] += 1;
         if (lfs.tfs_message_matched[this->lf_index]) {
             lfs.tfs_filter_hits[this->lf_index] += 1;
@@ -110,30 +114,110 @@ void text_filter::end_of_message(logfile_filter_state &lfs)
 }
 
 bookmark_type_t textview_curses::BM_USER("user");
+bookmark_type_t textview_curses::BM_USER_EXPR("user-expr");
 bookmark_type_t textview_curses::BM_SEARCH("search");
 bookmark_type_t textview_curses::BM_META("meta");
 
-string_attr_type textview_curses::SA_ORIGINAL_LINE("original_line");
-string_attr_type textview_curses::SA_BODY("body");
-string_attr_type textview_curses::SA_HIDDEN("hidden");
-string_attr_type textview_curses::SA_FORMAT("format");
-string_attr_type textview_curses::SA_REMOVED("removed");
-
 textview_curses::textview_curses()
-    : tc_sub_source(NULL),
-      tc_delegate(NULL),
-      tc_selection_start(-1),
-      tc_selection_last(-1),
-      tc_selection_cleared(false),
-      tc_hide_fields(true)
+    : tc_search_action(noop_func{})
 {
     this->set_data_source(this);
 }
 
 textview_curses::~textview_curses()
-{ }
+{
+    this->tc_search_action = noop_func{};
+}
 
-void textview_curses::reload_data(void)
+void textview_curses::reload_config(error_reporter &reporter)
+{
+    static auto DEFAULT_THEME_NAME = string("default");
+
+    for (auto iter = this->tc_highlights.begin();
+         iter != this->tc_highlights.end();) {
+        if (iter->first.first != highlight_source_t::THEME) {
+            ++iter;
+            continue;
+        }
+
+        iter = this->tc_highlights.erase(iter);
+    }
+
+    std::map<std::string, std::string> vars;
+    auto curr_theme_iter = lnav_config.lc_ui_theme_defs.find(lnav_config.lc_ui_theme);
+    if (curr_theme_iter != lnav_config.lc_ui_theme_defs.end()) {
+        vars = curr_theme_iter->second.lt_vars;
+    }
+
+    for (const auto& theme_name : {DEFAULT_THEME_NAME, lnav_config.lc_ui_theme}) {
+        auto theme_iter = lnav_config.lc_ui_theme_defs.find(theme_name);
+
+        if (theme_iter == lnav_config.lc_ui_theme_defs.end()) {
+            continue;
+        }
+
+        for (const auto &hl_pair : theme_iter->second.lt_highlights) {
+            if (hl_pair.second.hc_regex.empty()) {
+                continue;
+            }
+
+            const char *errptr;
+            pcre *code;
+            int eoff;
+
+            if ((code = pcre_compile(hl_pair.second.hc_regex.c_str(),
+                                     0,
+                                     &errptr,
+                                     &eoff,
+                                     nullptr)) == nullptr) {
+                reporter(&hl_pair.second.hc_regex,
+                         fmt::format("invalid highlight regex: {} at {}",
+                                     errptr, eoff));
+                continue;
+            }
+
+            const auto &sc = hl_pair.second.hc_style;
+            string fg1, bg1, fg_color, bg_color, errmsg;
+            bool invalid = false;
+            int attrs = 0;
+
+            fg1 = sc.sc_color;
+            bg1 = sc.sc_background_color;
+            shlex(fg1).eval(fg_color, vars);
+            shlex(bg1).eval(bg_color, vars);
+
+            auto fg = styling::color_unit::from_str(fg_color)
+                .unwrapOrElse([&](const auto& msg) {
+                    reporter(&sc.sc_color, errmsg);
+                    invalid = true;
+                    return styling::color_unit::make_empty();
+                });
+            auto bg = styling::color_unit::from_str(bg_color)
+                .unwrapOrElse([&](const auto& msg) {
+                    reporter(&sc.sc_background_color, errmsg);
+                    invalid = true;
+                    return styling::color_unit::make_empty();
+                });
+            if (invalid) {
+                continue;
+            }
+
+            if (sc.sc_bold) {
+                attrs |= A_BOLD;
+            }
+            if (sc.sc_underline) {
+                attrs |= A_UNDERLINE;
+            }
+            this->tc_highlights[{highlight_source_t::THEME, hl_pair.first}] =
+                highlighter(code)
+                    .with_pattern(hl_pair.second.hc_regex)
+                    .with_attrs(attrs != 0 ? attrs : -1)
+                    .with_color(fg, bg);
+        }
+    }
+}
+
+void textview_curses::reload_data()
 {
     if (this->tc_sub_source != nullptr) {
         this->tc_sub_source->text_update_marks(this->tc_bookmarks);
@@ -154,27 +238,55 @@ void textview_curses::grep_begin(grep_proc<vis_line_t> &gp, vis_line_t start, vi
     require(this->tc_searching >= 0);
 
     this->tc_searching += 1;
-    this->tc_search_action.invoke(this);
+    this->tc_search_action(this);
 
-    bookmark_vector<vis_line_t> &search_bv = this->tc_bookmarks[&BM_SEARCH];
-
-    if (start != -1) {
-        auto pair = search_bv.equal_range(vis_line_t(start), vis_line_t(stop));
+    if (start != -1_vl) {
+        auto& search_bv = this->tc_bookmarks[&BM_SEARCH];
+        auto pair = search_bv.equal_range(start, stop);
 
+        if (pair.first != pair.second) {
+            this->set_needs_update();
+        }
         for (auto mark_iter = pair.first;
              mark_iter != pair.second;
              ++mark_iter) {
-            this->set_user_mark(&BM_SEARCH, *mark_iter, false);
+            if (this->tc_sub_source) {
+                this->tc_sub_source->text_mark(&BM_SEARCH, *mark_iter, false);
+            }
+        }
+        if (pair.first != pair.second) {
+            search_bv.erase(pair.first, pair.second);
         }
     }
 
     listview_curses::reload_data();
 }
 
+void textview_curses::grep_end_batch(grep_proc<vis_line_t> &gp)
+{
+    if (this->tc_follow_deadline.tv_sec &&
+        this->tc_follow_top == this->get_top()) {
+        struct timeval now;
+
+        gettimeofday(&now, nullptr);
+        if (this->tc_follow_deadline < now) {
+        } else {
+            if (this->tc_follow_func) {
+                if (this->tc_follow_func()) {
+                    this->tc_follow_deadline = {0, 0};
+                }
+            } else {
+                this->tc_follow_deadline = {0, 0};
+            }
+        }
+    }
+    this->tc_search_action(this);
+}
+
 void textview_curses::grep_end(grep_proc<vis_line_t> &gp)
 {
     this->tc_searching -= 1;
-    this->tc_search_action.invoke(this);
+    this->grep_end_batch(gp);
 
     ensure(this->tc_searching >= 0);
 }
@@ -209,7 +321,7 @@ bool textview_curses::handle_mouse(mouse_event &me)
     unsigned long width;
     vis_line_t height;
 
-    if (this->tc_selection_start == -1 &&
+    if (this->tc_selection_start == -1_vl &&
         listview_curses::handle_mouse(me)) {
         return true;
     }
@@ -219,7 +331,7 @@ bool textview_curses::handle_mouse(mouse_event &me)
         return true;
     }
 
-    if (me.me_button != BUTTON_LEFT) {
+    if (me.me_button != mouse_button_t::BUTTON_LEFT) {
         return false;
     }
 
@@ -232,19 +344,19 @@ bool textview_curses::handle_mouse(mouse_event &me)
     this->get_dimensions(height, width);
 
     switch (me.me_state) {
-    case BUTTON_STATE_PRESSED:
+    case mouse_button_state_t::BUTTON_STATE_PRESSED:
         this->tc_selection_start = mouse_line;
-        this->tc_selection_last = vis_line_t(-1);
+        this->tc_selection_last = -1_vl;
         this->tc_selection_cleared = false;
         break;
-    case BUTTON_STATE_DRAGGED:
+    case mouse_button_state_t::BUTTON_STATE_DRAGGED:
         if (me.me_y <= 0) {
-            this->shift_top(vis_line_t(-1));
+            this->shift_top(-1_vl);
             me.me_y = 0;
             mouse_line = this->get_top();
         }
         if (me.me_y >= height && this->get_top() < this->get_top_for_last_row()) {
-            this->shift_top(vis_line_t(1));
+            this->shift_top(1_vl);
             me.me_y = height;
             mouse_line = this->get_bottom();
         }
@@ -258,11 +370,11 @@ bool textview_curses::handle_mouse(mouse_event &me)
                this->tc_selection_last);
         }
         if (this->tc_selection_start == mouse_line) {
-            this->tc_selection_last = vis_line_t(-1);
+            this->tc_selection_last = -1_vl;
         }
         else {
             if (!this->tc_selection_cleared) {
-                if (this->tc_sub_source != NULL) {
+                if (this->tc_sub_source != nullptr) {
                     this->tc_sub_source->text_clear_marks(&BM_USER);
                 }
                 this->tc_bookmarks[&BM_USER].clear();
@@ -276,9 +388,9 @@ bool textview_curses::handle_mouse(mouse_event &me)
         }
         this->reload_data();
         break;
-    case BUTTON_STATE_RELEASED:
-        this->tc_selection_start = vis_line_t(-1);
-        this->tc_selection_last = vis_line_t(-1);
+    case mouse_button_state_t::BUTTON_STATE_RELEASED:
+        this->tc_selection_start = -1_vl;
+        this->tc_selection_last = -1_vl;
         this->tc_selection_cleared = false;
         break;
     }
@@ -289,8 +401,6 @@ bool textview_curses::handle_mouse(mouse_event &me)
 void textview_curses::textview_value_for_row(vis_line_t row,
                                              attr_line_t &value_out)
 {
-    view_colors &vc = view_colors::singleton();
-    bookmark_vector<vis_line_t> &user_marks = this->tc_bookmarks[&BM_USER];
     string_attrs_t &sa = value_out.get_attrs();
     string &str = value_out.get_string();
     text_format_t source_format = this->tc_sub_source->get_text_format();
@@ -322,10 +432,11 @@ void textview_curses::textview_value_for_row(vis_line_t row,
 
     for (auto &tc_highlight : this->tc_highlights) {
         bool internal_hl =
-            tc_highlight.first.first == highlight_source_t::INTERNAL;
+            tc_highlight.first.first == highlight_source_t::INTERNAL ||
+            tc_highlight.first.first == highlight_source_t::THEME;
 
-        if (tc_highlight.second.h_text_format != text_format_t::TF_UNKNOWN &&
-            source_format != tc_highlight.second.h_text_format) {
+        if (!tc_highlight.second.h_text_formats.empty() &&
+            tc_highlight.second.h_text_formats.count(source_format) == 0) {
             continue;
         }
 
@@ -334,6 +445,10 @@ void textview_curses::textview_value_for_row(vis_line_t row,
             continue;
         }
 
+        if (this->tc_disabled_highlights.count(tc_highlight.first.first)) {
+            continue;
+        }
+
         // Internal highlights should only apply to the log message body so
         // that we don't start highlighting other fields.  User-provided
         // highlights should apply only to the line itself and not any of the
@@ -344,26 +459,7 @@ void textview_curses::textview_value_for_row(vis_line_t row,
     }
 
     if (this->tc_hide_fields) {
-        for (auto &sattr : sa) {
-            if (sattr.sa_type == &SA_HIDDEN &&
-                sattr.sa_range.length() > 3) {
-                struct line_range &lr = sattr.sa_range;
-
-                str.replace(lr.lr_start, lr.length(), "\xE2\x8B\xAE");
-                shift_string_attrs(sa, lr.lr_start + 1, -(lr.length() - 3));
-                sattr.sa_type = &VC_STYLE;
-                sattr.sa_value.sav_int =
-                    vc.attrs_for_role(view_colors::VCR_HIDDEN);
-                lr.lr_end = lr.lr_start + 3;
-
-                for_each(sa.begin(), sa.end(), [&] (string_attr &attr) {
-                    if (attr.sa_type == &VC_STYLE &&
-                        attr.sa_range.lr_start == lr.lr_start) {
-                        attr.sa_type = &SA_REMOVED;
-                    }
-                });
-            }
-        }
+        value_out.apply_hide();
     }
 
 #if 0
@@ -417,7 +513,10 @@ void textview_curses::textview_value_for_row(vis_line_t row,
     }
 #endif
 
-    if (binary_search(user_marks.begin(), user_marks.end(), row)) {
+    const auto &user_marks = this->tc_bookmarks[&BM_USER];
+    const auto &user_expr_marks = this->tc_bookmarks[&BM_USER_EXPR];
+    if (binary_search(user_marks.begin(), user_marks.end(), row) ||
+        binary_search(user_expr_marks.begin(), user_expr_marks.end(), row)) {
         sa.emplace_back(line_range{orig_line.lr_start, -1},
                         &view_curses::VC_STYLE,
                         A_REVERSE);
@@ -429,10 +528,12 @@ void textview_curses::execute_search(const std::string &regex_orig)
     std::string regex = regex_orig;
     pcre *code = nullptr;
 
-    if ((this->tc_search_child == nullptr) || (regex != this->tc_last_search)) {
+    if ((this->tc_search_child == nullptr) ||
+        (regex != this->tc_current_search)) {
         const char *errptr;
         int         eoff;
 
+        this->tc_previous_search = this->tc_current_search;
         this->match_reset();
 
         this->tc_search_child.reset();
@@ -449,7 +550,7 @@ void textview_curses::execute_search(const std::string &regex_orig)
                                       nullptr)) == nullptr) {
             string errmsg = string(errptr);
 
-            regex = pcrecpp::RE::QuoteMeta(regex);
+            regex = pcrepp::quote(regex);
 
             log_info("invalid search regex, using quoted: %s", regex.c_str());
             if ((code = pcre_compile(regex.c_str(),
@@ -466,7 +567,7 @@ void textview_curses::execute_search(const std::string &regex_orig)
 
             hl.with_role(view_colors::VCR_SEARCH);
 
-            textview_curses::highlight_map_t &hm = this->get_highlights();
+            highlight_map_t &hm = this->get_highlights();
             hm[{highlight_source_t::PREVIEW, "search"}] = hl;
 
             unique_ptr<grep_proc<vis_line_t>> gp = make_unique<grep_proc<vis_line_t>>(code, *this);
@@ -501,40 +602,166 @@ void textview_curses::execute_search(const std::string &regex_orig)
         }
     }
 
-    this->tc_last_search = regex;
+    this->tc_current_search = regex;
     if (this->tc_state_event_handler) {
         this->tc_state_event_handler(*this);
     }
 }
 
+void
+textview_curses::horiz_shift(vis_line_t start, vis_line_t end, int off_start,
+                             pair<int, int> &range_out)
+{
+    highlighter &hl = this->tc_highlights[{highlight_source_t::PREVIEW, "search"}];
+    int          prev_hit = -1, next_hit = INT_MAX;
+
+    for (; start < end; ++start) {
+        std::vector<attr_line_t> rows(1);
+        int off;
+
+        this->listview_value_for_rows(*this, start, rows);
+
+        const std::string &str = rows[0].get_string();
+        for (off = 0; off < (int)str.size(); ) {
+            int rc, matches[128];
+
+            rc = pcre_exec(hl.h_code,
+                           hl.h_code_extra,
+                           str.c_str(),
+                           str.size(),
+                           off,
+                           0,
+                           matches,
+                           128);
+            if (rc > 0) {
+                struct line_range lr;
+
+                if (rc == 2) {
+                    lr.lr_start = matches[2];
+                    lr.lr_end   = matches[3];
+                }
+                else {
+                    lr.lr_start = matches[0];
+                    lr.lr_end   = matches[1];
+                }
+
+                if (lr.lr_start < off_start) {
+                    prev_hit = std::max(prev_hit, lr.lr_start);
+                }
+                else if (lr.lr_start > off_start) {
+                    next_hit = std::min(next_hit, lr.lr_start);
+                }
+                if (lr.lr_end > lr.lr_start) {
+                    off = matches[1];
+                }
+                else {
+                    off += 1;
+                }
+            }
+            else {
+                off = str.size();
+            }
+        }
+    }
+
+    range_out = std::make_pair(prev_hit, next_hit);
+}
+
+void
+textview_curses::set_user_mark(bookmark_type_t *bm, vis_line_t vl, bool marked)
+{
+    bookmark_vector<vis_line_t> &bv = this->tc_bookmarks[bm];
+    bookmark_vector<vis_line_t>::iterator iter;
+
+    if (marked) {
+        bv.insert_once(vl);
+    }
+    else {
+        iter = std::lower_bound(bv.begin(), bv.end(), vl);
+        if (iter != bv.end() && *iter == vl) {
+            bv.erase(iter);
+        }
+    }
+    if (this->tc_sub_source) {
+        this->tc_sub_source->text_mark(bm, vl, marked);
+    }
+
+    if (marked) {
+        this->search_range(vl, vl + 1_vl);
+        this->search_new_data();
+    }
+    this->set_needs_update();
+}
+
+void
+textview_curses::toggle_user_mark(bookmark_type_t *bm, vis_line_t start_line,
+                                  vis_line_t end_line)
+{
+    if (end_line == -1) {
+        end_line = start_line;
+    }
+    if (start_line > end_line) {
+        std::swap(start_line, end_line);
+    }
+
+    if (start_line >= this->get_inner_height()) {
+        return;
+    }
+    if (end_line >= this->get_inner_height()) {
+        end_line = vis_line_t(this->get_inner_height() - 1);
+    }
+    for (vis_line_t curr_line = start_line; curr_line <= end_line;
+         ++curr_line) {
+        bookmark_vector<vis_line_t> &bv = this->tc_bookmarks[bm];
+        bookmark_vector<vis_line_t>::iterator iter;
+        bool added;
+
+        iter = bv.insert_once(curr_line);
+        if (iter == bv.end()) {
+            added = true;
+        }
+        else {
+            bv.erase(iter);
+            added = false;
+        }
+        if (this->tc_sub_source) {
+            this->tc_sub_source->text_mark(bm, curr_line, added);
+        }
+    }
+    this->search_range(start_line, end_line + 1_vl);
+    this->search_new_data();
+}
+
 void text_time_translator::scroll_invoked(textview_curses *tc)
 {
     if (tc->get_inner_height() > 0) {
-        this->ttt_top_time = this->time_for_row((int) tc->get_top());
+        this->time_for_row(tc->get_top()) | [this](auto new_top_time) {
+            this->ttt_top_time = new_top_time;
+        };
     }
 }
 
 void text_time_translator::data_reloaded(textview_curses *tc)
 {
     if (tc->get_inner_height() > 0) {
-        struct timeval top_time = this->time_for_row((int) tc->get_top());
-
-        if (top_time != this->ttt_top_time) {
-            if (this->ttt_top_time.tv_sec != 0) {
-                vis_line_t new_top(this->row_for_time(this->ttt_top_time));
-
-                if (new_top >= 0) {
-                    tc->set_top(new_top);
+        this->time_for_row(tc->get_top()) | [this, tc](auto top_time) {
+            if (top_time != this->ttt_top_time) {
+                if (this->ttt_top_time.tv_sec != 0) {
+                    this->row_for_time(this->ttt_top_time) | [tc](auto new_top) {
+                        tc->set_top(new_top);
+                    };
                 }
+                this->time_for_row(tc->get_top()) | [this](auto new_top_time) {
+                    this->ttt_top_time = new_top_time;
+                };
             }
-            this->ttt_top_time = this->time_for_row((int) tc->get_top());
-        }
+        };
     }
 }
 
 template class bookmark_vector<vis_line_t>;
 
-bool empty_filter::matches(const logfile &lf, const logline &ll,
+bool empty_filter::matches(const logfile &lf, logfile::const_iterator ll,
                            shared_buffer_ref &line)
 {
     return false;
