@@ -30,16 +30,11 @@
 #include "config.h"
 
 #include <assert.h>
-#include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
-#include <sys/types.h>
-#include <pwd.h>
-#include <grp.h>
-
-#include <utility>
 
 #include "lnav.hh"
+#include "base/injector.bind.hh"
 #include "base/lnav_log.hh"
 #include "sql_util.hh"
 #include "views_vtab.hh"
@@ -51,22 +46,13 @@ template<>
 struct from_sqlite<lnav_view_t> {
     inline lnav_view_t operator()(int argc, sqlite3_value **val, int argi) {
         const char *view_name = (const char *) sqlite3_value_text(val[argi]);
+        auto view_index_opt = view_from_string(view_name);
 
-        if (view_name == nullptr) {
+        if (!view_index_opt) {
             throw from_sqlite_conversion_error("lnav view name", argi);
         }
 
-        auto view_name_iter = find_if(
-            ::begin(lnav_view_strings), ::end(lnav_view_strings),
-            [&](const char *v) {
-                return v != nullptr && strcasecmp(v, view_name) == 0;
-            });
-
-        if (view_name_iter == ::end(lnav_view_strings)) {
-            throw from_sqlite_conversion_error("lnav view name", argi);
-        }
-
-        return lnav_view_t(view_name_iter - lnav_view_strings);
+        return view_index_opt.value();
     }
 };
 
@@ -86,11 +72,11 @@ struct from_sqlite<text_filter::type_t> {
 };
 
 template<>
-struct from_sqlite<pair<string, pcre *>> {
-    inline pair<string, pcre *>operator()(int argc, sqlite3_value **val, int argi) {
+struct from_sqlite<pair<string, auto_mem<pcre>>> {
+    inline pair<string, auto_mem<pcre>> operator()(int argc, sqlite3_value **val, int argi) {
         const char *pattern = (const char *) sqlite3_value_text(val[argi]);
         const char *errptr;
-        pcre *code;
+        auto_mem<pcre> code;
         int eoff;
 
         if (pattern == nullptr || pattern[0] == '\0') {
@@ -104,22 +90,17 @@ struct from_sqlite<pair<string, pcre *>> {
                             nullptr);
 
         if (code == nullptr) {
-            throw from_sqlite_conversion_error(errptr, argi);
+            throw sqlite_func_error(
+                "Invalid regular expression in column {}: {} at offset {}",
+                argi, errptr, eoff);
         }
 
-        return make_pair(string(pattern), code);
+        return make_pair(string(pattern), std::move(code));
     }
 };
 
 struct lnav_views : public tvt_iterator_cursor<lnav_views> {
-    struct vtab {
-        sqlite3_vtab base;
-
-        operator sqlite3_vtab *() {
-            return &this->base;
-        };
-    };
-
+    static constexpr const char *NAME = "lnav_views";
     static constexpr const char *CREATE_STMT = R"(
 -- Access lnav's views through this table.
 CREATE TABLE lnav_views (
@@ -129,8 +110,10 @@ CREATE TABLE lnav_views (
     height INTEGER,         -- The height of the viewport.
     inner_height INTEGER,   -- The number of lines in the view.
     top_time DATETIME,      -- The time of the top line in the view, if the content is time-based.
+    top_file TEXT,          -- The file the top line is from.
     paused INTEGER,         -- Indicates if the view is paused and will not load new data.
-    search TEXT             -- The text to search for in the view.
+    search TEXT,            -- The text to search for in the view.
+    filtering INTEGER       -- Indicates if the view is applying filters.
 );
 )";
 
@@ -173,23 +156,45 @@ CREATE TABLE lnav_views (
                 auto *time_source = dynamic_cast<text_time_translator *>(tc.get_sub_source());
 
                 if (time_source != nullptr && tc.get_inner_height() > 0) {
-                    char timestamp[64];
+                    auto top_time_opt = time_source->time_for_row(tc.get_top());
 
-                    sql_strftime(timestamp, sizeof(timestamp), time_source->time_for_row(tc.get_top()), 'T');
+                    if (top_time_opt) {
+                        char timestamp[64];
 
-                    sqlite3_result_text(ctx, timestamp, -1, SQLITE_TRANSIENT);
+                        sql_strftime(timestamp, sizeof(timestamp), top_time_opt.value(), 'T');
+                        sqlite3_result_text(ctx, timestamp, -1, SQLITE_TRANSIENT);
+                    } else {
+                        sqlite3_result_null(ctx);
+                    }
                 } else {
                     sqlite3_result_null(ctx);
                 }
                 break;
             }
-            case 6:
+            case 6: {
+                to_sqlite(ctx, tc.map_top_row([](const auto& al) {
+                    return get_string_attr(al.get_attrs(), &logline::L_FILE) | [](const auto* sa) {
+                        auto lf = (logfile *) sa->sa_value.sav_ptr;
+
+                        return nonstd::make_optional(lf->get_filename());
+                    };
+                }));
+                break;
+            }
+            case 7:
                 sqlite3_result_int(ctx, tc.is_paused());
                 break;
-            case 7: {
-                const string &str = tc.get_last_search();
+            case 8:
+                to_sqlite(ctx, tc.get_current_search());
+                break;
+            case 9: {
+                auto tss = tc.get_sub_source();
 
-                sqlite3_result_text(ctx, str.c_str(), str.length(), SQLITE_TRANSIENT);
+                if (tss != nullptr && tss->tss_supports_filtering) {
+                    sqlite3_result_int(ctx, tss->tss_apply_filters);
+                } else {
+                    sqlite3_result_int(ctx, 0);
+                }
                 break;
             }
         }
@@ -217,8 +222,10 @@ CREATE TABLE lnav_views (
                    int64_t height,
                    int64_t inner_height,
                    const char *top_time,
+                   const char *top_file,
                    bool is_paused,
-                   const char *search) {
+                   const char *search,
+                   bool do_filtering) {
         textview_curses &tc = lnav_data.ld_views[index];
         text_time_translator *time_source = dynamic_cast<text_time_translator *>(tc.get_sub_source());
 
@@ -229,12 +236,15 @@ CREATE TABLE lnav_views (
             struct timeval tv;
 
             if (dts.convert_to_timeval(top_time, -1, nullptr, tv)) {
-                struct timeval last_time = time_source->time_for_row(tc.get_top());
+                auto last_time_opt = time_source->time_for_row(tc.get_top());
 
-                if (tv != last_time) {
-                    int row = time_source->row_for_time(tv);
-
-                    tc.set_top(vis_line_t(row));
+                if (last_time_opt) {
+                    auto last_time = last_time_opt.value();
+                    if (tv != last_time) {
+                        time_source->row_for_time(tv) | [&tc](auto row) {
+                            tc.set_top(row);
+                        };
+                    }
                 }
             } else {
                 tab->zErrMsg = sqlite3_mprintf("Invalid time: %s", top_time);
@@ -244,6 +254,13 @@ CREATE TABLE lnav_views (
         tc.set_left(left);
         tc.set_paused(is_paused);
         tc.execute_search(search);
+        auto tss = tc.get_sub_source();
+        if (tss != nullptr &&
+            tss->tss_supports_filtering &&
+            tss->tss_apply_filters != do_filtering) {
+            tss->tss_apply_filters = do_filtering;
+            tss->text_filters_changed();
+        }
 
         return SQLITE_OK;
     };
@@ -252,6 +269,7 @@ CREATE TABLE lnav_views (
 struct lnav_view_stack : public tvt_iterator_cursor<lnav_view_stack> {
     using iterator = vector<textview_curses *>::iterator;
 
+    static constexpr const char *NAME = "lnav_view_stack";
     static constexpr const char *CREATE_STMT = R"(
 -- Access lnav's view stack through this table.
 CREATE TABLE lnav_view_stack (
@@ -259,25 +277,17 @@ CREATE TABLE lnav_view_stack (
 );
 )";
 
-    struct vtab {
-        sqlite3_vtab base;
-
-        operator sqlite3_vtab *() {
-            return &this->base;
-        };
-    };
-
     iterator begin() {
-        return lnav_data.ld_view_stack.vs_views.begin();
+        return lnav_data.ld_view_stack.begin();
     }
 
     iterator end() {
-        return lnav_data.ld_view_stack.vs_views.end();
+        return lnav_data.ld_view_stack.end();
     }
 
     int get_column(cursor &vc, sqlite3_context *ctx, int col) {
         textview_curses *tc = *vc.iter;
-        lnav_view_t view = lnav_view_t(tc - lnav_data.ld_views);
+        auto view = lnav_view_t(tc - lnav_data.ld_views);
 
         switch (col) {
             case 0:
@@ -291,21 +301,14 @@ CREATE TABLE lnav_view_stack (
     };
 
     int delete_row(sqlite3_vtab *tab, sqlite3_int64 rowid) {
-        if ((size_t)rowid != lnav_data.ld_view_stack.vs_views.size() - 1) {
+        if ((size_t)rowid != lnav_data.ld_view_stack.size() - 1) {
             tab->zErrMsg = sqlite3_mprintf(
                 "Only the top view in the stack can be deleted");
             return SQLITE_ERROR;
         }
 
         lnav_data.ld_last_view = *lnav_data.ld_view_stack.top();
-        lnav_data.ld_view_stack.vs_views.pop_back();
-        if (!lnav_data.ld_view_stack.vs_views.empty()) {
-            textview_curses *tc = *lnav_data.ld_view_stack.top();
-
-            tc->set_needs_update();
-            lnav_data.ld_view_stack_broadcaster.invoke(tc);
-        }
-
+        lnav_data.ld_view_stack.pop_back();
         return SQLITE_OK;
     };
 
@@ -315,7 +318,7 @@ CREATE TABLE lnav_view_stack (
         textview_curses *tc = &lnav_data.ld_views[view_index];
 
         ensure_view(tc);
-        rowid_out = lnav_data.ld_view_stack.vs_views.size() - 1;
+        rowid_out = lnav_data.ld_view_stack.size() - 1;
 
         return SQLITE_OK;
     };
@@ -328,15 +331,6 @@ CREATE TABLE lnav_view_stack (
 };
 
 struct lnav_view_filter_base {
-
-    struct vtab {
-        sqlite3_vtab base;
-
-        operator sqlite3_vtab *() {
-            return &this->base;
-        };
-    };
-
     struct iterator {
         using difference_type = int;
         using value_type = text_filter;
@@ -392,7 +386,7 @@ struct lnav_view_filter_base {
     }
 
     iterator end() {
-        return iterator(LNV__MAX, -1);
+        return {LNV__MAX, -1};
     }
 
     sqlite_int64 get_rowid(iterator iter) {
@@ -412,6 +406,7 @@ struct lnav_view_filter_base {
 
 struct lnav_view_filters : public tvt_iterator_cursor<lnav_view_filters>,
     public lnav_view_filter_base {
+    static constexpr const char *NAME = "lnav_view_filters";
     static constexpr const char *CREATE_STMT = R"(
 -- Access lnav's filters through this table.
 CREATE TABLE lnav_view_filters (
@@ -470,15 +465,19 @@ CREATE TABLE lnav_view_filters (
                    nonstd::optional<int64_t> _filter_id,
                    nonstd::optional<bool> enabled,
                    nonstd::optional<text_filter::type_t> type,
-                   pair<string, pcre *> pattern) {
+                   pair<string, auto_mem<pcre>> pattern) {
         textview_curses &tc = lnav_data.ld_views[view_index];
         text_sub_source *tss = tc.get_sub_source();
         filter_stack &fs = tss->get_filters();
+        auto filter_index = fs.next_index();
+        if (!filter_index) {
+            throw sqlite_func_error("Too many filters");
+        }
         auto pf = make_shared<pcre_filter>(
             type.value_or(text_filter::type_t::EXCLUDE),
             pattern.first,
-            fs.next_index(),
-            pattern.second);
+            *filter_index,
+            pattern.second.release());
         fs.add_filter(pf);
         if (!enabled.value_or(true)) {
             pf->disable();
@@ -514,13 +513,19 @@ CREATE TABLE lnav_view_filters (
                    int64_t new_filter_id,
                    bool enabled,
                    text_filter::type_t type,
-                   pair<string, pcre *> pattern) {
-        lnav_view_t view_index = lnav_view_t(rowid >> 32);
-        int filter_index = rowid & 0xffffffffLL;
+                   pair<string, auto_mem<pcre>> pattern) {
+        auto view_index = lnav_view_t(rowid >> 32);
+        auto filter_index = rowid & 0xffffffffLL;
         textview_curses &tc = lnav_data.ld_views[view_index];
         text_sub_source *tss = tc.get_sub_source();
         filter_stack &fs = tss->get_filters();
-        auto iter = fs.begin() + filter_index;
+        auto iter = fs.begin();
+        for (; iter != fs.end(); ++iter) {
+            if ((*iter)->get_index() == (size_t) filter_index) {
+                break;
+            }
+        }
+
         shared_ptr<text_filter> tf = *iter;
 
         if (new_view_index != view_index) {
@@ -535,7 +540,7 @@ CREATE TABLE lnav_view_filters (
         auto pf = make_shared<pcre_filter>(type,
                                            pattern.first,
                                            tf->get_index(),
-                                           pattern.second);
+                                           pattern.second.release());
 
         if (!enabled) {
             pf->disable();
@@ -551,6 +556,7 @@ CREATE TABLE lnav_view_filters (
 
 struct lnav_view_filter_stats : public tvt_iterator_cursor<lnav_view_filter_stats>,
     public lnav_view_filter_base {
+    static constexpr const char *NAME = "lnav_view_filter_stats";
     static constexpr const char *CREATE_STMT = R"(
 -- Access statistics for filters through this table.
 CREATE TABLE lnav_view_filter_stats (
@@ -584,37 +590,98 @@ CREATE TABLE lnav_view_filter_stats (
     }
 };
 
+struct lnav_view_files : public tvt_iterator_cursor<lnav_view_files> {
+    static constexpr const char *NAME = "lnav_view_files";
+    static constexpr const char *CREATE_STMT = R"(
+--
+CREATE TABLE lnav_view_files (
+    view_name TEXT,     -- The name of the view.
+    filepath  TEXT,     -- The path to the file.
+    visible   INTEGER   -- Indicates whether or not the file is shown.
+);
+)";
+
+    using iterator = logfile_sub_source::iterator;
+
+    iterator begin() {
+        return lnav_data.ld_log_source.begin();
+    }
+
+    iterator end() {
+        return lnav_data.ld_log_source.end();
+    }
+
+    int get_column(cursor &vc, sqlite3_context *ctx, int col) {
+        auto& ld = *vc.iter;
+
+        switch (col) {
+            case 0:
+                sqlite3_result_text(ctx,
+                                    lnav_view_strings[LNV_LOG], -1,
+                                    SQLITE_STATIC);
+                break;
+            case 1:
+                to_sqlite(ctx, ld->ld_filter_state.lfo_filter_state
+                    .tfs_logfile->get_filename());
+                break;
+            case 2:
+                to_sqlite(ctx, ld->ld_visible);
+                break;
+        }
+
+        return SQLITE_OK;
+    }
+
+    int delete_row(sqlite3_vtab *tab, sqlite3_int64 rowid) {
+        tab->zErrMsg = sqlite3_mprintf(
+            "Rows cannot be deleted from the lnav_view_files table");
+        return SQLITE_ERROR;
+    }
+
+    int insert_row(sqlite3_vtab *tab, sqlite3_int64 &rowid_out) {
+        tab->zErrMsg = sqlite3_mprintf(
+            "Rows cannot be inserted into the lnav_view_files table");
+        return SQLITE_ERROR;
+    };
+
+    int update_row(sqlite3_vtab *tab,
+                   sqlite3_int64 &rowid,
+                   const char *view_name,
+                   const char *file_path,
+                   bool visible) {
+        auto &lss = lnav_data.ld_log_source;
+        auto iter = this->begin();
+
+        std::advance(iter, rowid);
+
+        auto& ld = *iter;
+        if (ld->ld_visible != visible) {
+            ld->set_visibility(visible);
+            lss.text_filters_changed();
+        }
+
+        return SQLITE_OK;
+    }
+};
+
 static const char *CREATE_FILTER_VIEW = R"(
 CREATE VIEW lnav_view_filters_and_stats AS
   SELECT * FROM lnav_view_filters LEFT NATURAL JOIN lnav_view_filter_stats
 )";
 
+static auto a = injector::bind_multiple<vtab_module_base>()
+    .add<vtab_module<lnav_views>>()
+    .add<vtab_module<lnav_view_stack>>()
+    .add<vtab_module<lnav_view_filters>>()
+    .add<vtab_module<tvt_no_update<lnav_view_filter_stats>>>()
+    .add<vtab_module<lnav_view_files>>();
+
 int register_views_vtab(sqlite3 *db)
 {
-    static vtab_module<lnav_views> LNAV_VIEWS_MODULE;
-    static vtab_module<lnav_view_stack> LNAV_VIEW_STACK_MODULE;
-    static vtab_module<lnav_view_filters> LNAV_VIEW_FILTERS_MODULE;
-    static vtab_module<tvt_no_update<lnav_view_filter_stats>> LNAV_VIEW_FILTER_STATS_MODULE;
-
-    int rc;
-
-    rc = LNAV_VIEWS_MODULE.create(db, "lnav_views");
-    assert(rc == SQLITE_OK);
-
-    rc = LNAV_VIEW_STACK_MODULE.create(db, "lnav_view_stack");
-    assert(rc == SQLITE_OK);
-
-    rc = LNAV_VIEW_FILTERS_MODULE.create(db, "lnav_view_filters");
-    assert(rc == SQLITE_OK);
-
-    rc = LNAV_VIEW_FILTER_STATS_MODULE.create(db, "lnav_view_filter_stats");
-    assert(rc == SQLITE_OK);
-
     char *errmsg;
     if (sqlite3_exec(db, CREATE_FILTER_VIEW, nullptr, nullptr, &errmsg) != SQLITE_OK) {
         log_error("Unable to create filter view: %s", errmsg);
-
     }
 
-    return rc;
+    return 0;
 }
