@@ -31,7 +31,8 @@
 
 #include "config.h"
 
-#include <pcrecpp.h>
+#include <regex>
+#include <utility>
 
 #include "fmt/format.h"
 
@@ -47,7 +48,7 @@ json_path_handler_base::json_path_handler_base(const string &property)
     : jph_property(property.back() == '#' ?
                    property.substr(0, property.size() - 1) :
                    property),
-      jph_regex(pcrecpp::RE::QuoteMeta(property), PCRE_ANCHORED),
+      jph_regex(pcrepp::quote(property), PCRE_ANCHORED),
       jph_is_array(property.back() == '#')
 {
     memset(&this->jph_callbacks, 0, sizeof(this->jph_callbacks));
@@ -55,12 +56,9 @@ json_path_handler_base::json_path_handler_base(const string &property)
 
 static std::string scrub_pattern(const std::string &pattern)
 {
-    static pcrecpp::RE CAPTURE(R"(\(\?\<\w+\>)");
-    std::string retval = pattern;
+    static const std::regex CAPTURE(R"(\(\?\<\w+\>)");
 
-    CAPTURE.GlobalReplace("(", &retval);
-
-    return retval;
+    return std::regex_replace(pattern, CAPTURE, "(");
 }
 
 json_path_handler_base::json_path_handler_base(const pcrepp &property)
@@ -72,9 +70,9 @@ json_path_handler_base::json_path_handler_base(const pcrepp &property)
     memset(&this->jph_callbacks, 0, sizeof(this->jph_callbacks));
 }
 
-json_path_handler_base::json_path_handler_base(const string &property,
+json_path_handler_base::json_path_handler_base(string property,
                                                const pcrepp &property_re)
-    : jph_property(property),
+    : jph_property(std::move(property)),
       jph_regex(property_re),
       jph_is_array(property_re.p_pattern.find('#') != string::npos)
 {
@@ -126,6 +124,9 @@ yajl_gen_status json_path_handler_base::gen(yajlpp_gen_context &ygc, yajl_gen ha
                 size_t len;
                 yajl_gen_get_buf(handle, &buf, &len);
                 if (status != yajl_gen_status_ok) {
+                    log_error("yajl_gen failure for: %s -- %d",
+                              jph.jph_property.c_str(),
+                              status);
                     return status;
                 }
             }
@@ -738,13 +739,15 @@ yajlpp_parse_context::parse(const unsigned char *jsonText, size_t jsonTextLen)
     this->ypc_json_text = nullptr;
 
     if (retval != yajl_status_ok && this->ypc_error_reporter) {
+        auto msg = yajl_get_error(this->ypc_handle, 1, jsonText, jsonTextLen);
+
         this->ypc_error_reporter(
             *this, lnav_log_level_t::ERROR,
             fmt::format("error:{}:{}:invalid json -- {}",
                         this->ypc_source,
                         this->get_line_number(),
-                        yajl_get_error(this->ypc_handle, 1,
-                                       jsonText, jsonTextLen)).c_str());
+                        msg).c_str());
+        yajl_free_error(this->ypc_handle, msg);
     }
 
     return retval;
@@ -755,12 +758,14 @@ yajl_status yajlpp_parse_context::complete_parse()
     yajl_status retval = yajl_complete_parse(this->ypc_handle);
 
     if (retval != yajl_status_ok && this->ypc_error_reporter) {
+        auto msg = yajl_get_error(this->ypc_handle, 0, nullptr, 0);
+
         this->ypc_error_reporter(
             *this, lnav_log_level_t::ERROR,
             fmt::format("error:{}:invalid json -- {}",
                         this->ypc_source,
-                        yajl_get_error(this->ypc_handle, 0,
-                                       nullptr, 0)).c_str());
+                        msg).c_str());
+        yajl_free_error(this->ypc_handle, msg);
     }
 
     return retval;
@@ -830,7 +835,8 @@ yajlpp_parse_context &yajlpp_parse_context::set_path(const string &path)
     for (size_t lpc = 0; lpc < path.size(); lpc++) {
         switch (path[lpc]) {
             case '/':
-                this->ypc_path_index_stack.push_back(1 + lpc);
+                this->ypc_path_index_stack.push_back(
+                    this->ypc_path_index_stack.empty() ? 1 : 0 + lpc);
                 break;
         }
     }
@@ -844,7 +850,7 @@ const char *yajlpp_parse_context::get_path_fragment(int offset, char *frag_in,
     size_t start, end;
 
     if (offset < 0) {
-        offset = this->ypc_path_index_stack.size() + offset;
+        offset = ((int) this->ypc_path_index_stack.size()) + offset;
     }
     start = this->ypc_path_index_stack[offset] + ((offset == 0) ? 0 : 1);
     if ((offset + 1) < (int)this->ypc_path_index_stack.size()) {
@@ -865,6 +871,21 @@ const char *yajlpp_parse_context::get_path_fragment(int offset, char *frag_in,
     return retval;
 }
 
+int yajlpp_parse_context::get_line_number() const
+{
+    if (this->ypc_handle != NULL && this->ypc_json_text) {
+        size_t consumed = yajl_get_bytes_consumed(this->ypc_handle);
+        long current_count = std::count(&this->ypc_json_text[0],
+                                        &this->ypc_json_text[consumed],
+                                        '\n');
+
+        return this->ypc_line_number + current_count;
+    }
+    else {
+        return this->ypc_line_number;
+    }
+}
+
 void yajlpp_gen_context::gen()
 {
     yajlpp_map root(this->ygc_handle);
@@ -902,6 +923,10 @@ void yajlpp_gen_context::gen_schema(const json_path_container *handlers)
 
                 def.gen("title");
                 def.gen(container.first);
+                if (!container.second->jpc_description.empty()) {
+                    def.gen("description");
+                    def.gen(container.second->jpc_description);
+                }
                 def.gen("type");
                 def.gen("object");
                 def.gen("$$target");
