@@ -35,7 +35,7 @@
 #include <fstream>
 #include <iostream>
 
-#include "pcrepp/pcrepp.hh"
+#include "base/injector.hh"
 #include "textview_curses.hh"
 #include "data_scanner.hh"
 #include "data_parser.hh"
@@ -51,22 +51,23 @@ using namespace std;
 
 const char *TMP_NAME = "scanned.tmp";
 
-string execute_any(exec_context &ec, const string &cmdline_with_mode)
-{
-    return "";
-}
-
-void add_global_vars(exec_context &ec)
-{
-}
-
 int main(int argc, char *argv[])
 {
     int  c, retval = EXIT_SUCCESS;
     bool prompt = false, is_log = false, pretty_print = false;
 
     {
-        std::vector<filesystem::path> paths;
+        static auto builtin_formats =
+            injector::get<std::vector<std::shared_ptr<log_format>>>();
+        auto& root_formats = log_format::get_root_formats();
+
+        log_format::get_root_formats().insert(
+            root_formats.begin(), builtin_formats.begin(), builtin_formats.end());
+        builtin_formats.clear();
+    }
+
+    {
+        std::vector<ghc::filesystem::path> paths;
         vector<string> errors;
 
         load_formats(paths, errors);
@@ -102,30 +103,32 @@ int main(int argc, char *argv[])
     }
     else {
         for (int lpc = 0; lpc < argc; lpc++) {
+            std::unique_ptr<ifstream> in_ptr;
             istream *in;
-            FILE *   out;
+            FILE *out;
 
             if (strcmp(argv[lpc], "-") == 0) {
                 in = &cin;
             }
             else {
-                ifstream *ifs = new ifstream(argv[lpc]);
+                auto ifs = std::make_unique<ifstream>(argv[lpc]);
 
                 if (!ifs->is_open()) {
                     fprintf(stderr, "error: unable to open file\n");
                     retval = EXIT_FAILURE;
                 }
                 else {
-                    in = ifs;
+                    in_ptr = std::move(ifs);
+                    in = in_ptr.get();
                 }
             }
 
-            if ((out = fopen(TMP_NAME, "w")) == NULL) {
+            if ((out = fopen(TMP_NAME, "w")) == nullptr) {
                 fprintf(stderr, "error: unable to temporary file for writing\n");
                 retval = EXIT_FAILURE;
             }
             else {
-                unique_ptr<log_format> format;
+                shared_ptr<log_format> format;
                 char *log_line;
                 bool found = false;
                 char   cmd[2048];
@@ -146,14 +149,14 @@ int main(int argc, char *argv[])
 
                 sbr.share(share_manager, (char *)sub_line.c_str(), sub_line.size());
 
-                vector<log_format *> &root_formats = log_format::get_root_formats();
-                vector<log_format *>::iterator iter;
+                auto &root_formats = log_format::get_root_formats();
+                vector<std::shared_ptr<log_format>>::iterator iter;
                 vector<logline> index;
 
                 if (is_log) {
                     for (iter = root_formats.begin();
-                       iter != root_formats.end() && !found;
-                       ++iter) {
+                         iter != root_formats.end() && !found;
+                         ++iter) {
                         line_info li = { {13}};
                         logfile *lf = nullptr; // XXX
 
@@ -163,14 +166,19 @@ int main(int argc, char *argv[])
                             found = true;
                         }
                     }
+
+                    if (!found) {
+                        fprintf(stderr, "error: log sample does not match\n");
+                        return EXIT_FAILURE;
+                    }
                 }
 
                 vector<logline_value> ll_values;
                 string_attrs_t sa;
 
-                if (format.get() != NULL) {
+                if (format.get() != nullptr) {
                     format->annotate(0, sbr, sa, ll_values);
-                    body = find_string_attr_range(sa, &textview_curses::SA_BODY);
+                    body = find_string_attr_range(sa, &SA_BODY);
                 }
 
                 data_parser::TRACE_FILE = fopen("scanned.dpt", "w");
@@ -196,7 +204,7 @@ int main(int argc, char *argv[])
 
                 auto_mem<yajl_gen_t> gen(yajl_gen_free);
 
-                gen = yajl_gen_alloc(NULL);
+                gen = yajl_gen_alloc(nullptr);
                 yajl_gen_config(gen.in(), yajl_gen_beautify, true);
 
                 elements_to_json(gen, dp, &dp.dp_pairs);
@@ -233,7 +241,7 @@ int main(int argc, char *argv[])
                 }
 
                 fclose(data_parser::TRACE_FILE);
-                data_parser::TRACE_FILE = NULL;
+                data_parser::TRACE_FILE = nullptr;
             }
         }
     }
