@@ -29,26 +29,18 @@
 
 #include "config.h"
 
-#include <fstream>
-
 #define DOCTEST_CONFIG_IMPLEMENT_WITH_MAIN
-#include "doctest.hh"
+#include "doctest/doctest.h"
 
+#include "byte_array.hh"
 #include "lnav_config.hh"
-#include "view_curses.hh"
 #include "relative_time.hh"
 #include "unique_path.hh"
 #include "logfile.hh"
+#include "log_format.hh"
 
 using namespace std;
 
-std::vector<log_format *>& log_format::get_root_formats()
-{
-    static std::vector<log_format *> retval;
-
-    return retval;
-}
-
 #if 0
 TEST_CASE("overwritten-logfile") {
     string fname = "reload_test.0";
@@ -69,29 +61,27 @@ TEST_CASE("overwritten-logfile") {
 }
 #endif
 
-TEST_CASE("duration2str") {
-    string val;
+TEST_CASE("byte_array") {
+    using my_array_t = byte_array<8>;
 
-    duration2str(25 * 60 * 60 * 1000 + 123, val);
-    CHECK(val == "1d1h0m0s");
-    val.clear();
-    duration2str(10 * 1000 + 123, val);
-    CHECK(val == "10s123");
-    val.clear();
-    duration2str(10 * 1000, val);
-    CHECK(val == "10s000");
-    val.clear();
-    duration2str(100, val);
-    CHECK(val == "100");
-    val.clear();
-    duration2str(0, val);
-    CHECK(val == "");
-    val.clear();
-    duration2str(-10, val);
-    CHECK(val == "-010");
-    val.clear();
-    duration2str(-10 * 1000, val);
-    CHECK(val == "-10s000");
+    my_array_t ba1;
+
+    memcpy(ba1.out(), "abcd1234", my_array_t::BYTE_COUNT);
+    CHECK(ba1.to_string() == "6162636431323334");
+    auto ba2 = ba1;
+    CHECK(ba1 == ba2);
+    CHECK_FALSE(ba1 != ba2);
+    CHECK_FALSE(ba1 < ba2);
+
+    my_array_t ba3;
+
+    memcpy(ba3.out(), "abcd1235", my_array_t::BYTE_COUNT);
+    CHECK(ba1 < ba3);
+    CHECK_FALSE(ba3 < ba1);
+
+    ba1.clear();
+    CHECK(ba1.to_string() == "0000000000000000");
+    CHECK(ba2.to_string() == "6162636431323334");
 }
 
 TEST_CASE("ptime_fmt") {
@@ -106,12 +96,7 @@ TEST_CASE("ptime_fmt") {
 
 TEST_CASE("rgb_color from string") {
     string name = "SkyBlue1";
-    rgb_color color;
-    string errmsg;
-    bool rc;
-
-    rc = rgb_color::from_str(name, color, errmsg);
-    CHECK(rc);
+    auto color = rgb_color::from_str(name).unwrap();
     CHECK(color.rc_r == 135);
     CHECK(color.rc_g == 215);
     CHECK(color.rc_b == 255);
@@ -155,15 +140,15 @@ TEST_CASE("ptime_roundtrip") {
 
 class my_path_source : public unique_path_source {
 public:
-    explicit my_path_source(const filesystem::path &p) : mps_path(p) {
+    explicit my_path_source(ghc::filesystem::path p) : mps_path(std::move(p)) {
 
     }
 
-    filesystem::path get_path() const override {
+    ghc::filesystem::path get_path() const override {
         return this->mps_path;
     }
 
-    filesystem::path mps_path;
+    ghc::filesystem::path mps_path;
 };
 
 TEST_CASE("unique_path") {
