@@ -1,5 +1,177 @@
 #! /bin/bash
 
+run_test ${lnav_test} -n \
+    -c ":goto 1" \
+    -c ":mark" \
+    -c ":hide-unmarked-lines" \
+    -c ":goto 0" \
+    ${test_dir}/logfile_access_log.0
+
+check_output "hide-unmarked-lines does not work?" <<EOF
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkboot.gz HTTP/1.0" 404 46210 "-" "gPXE/0.9.7"
+EOF
+
+run_test ${lnav_test} -n \
+    -c ":unix-time" \
+    "${test_dir}/logfile_access_log.*"
+
+check_error_output ":unix-time works without arg?" <<EOF
+command-option:1: error: expecting a unix time value
+EOF
+
+
+run_test ${lnav_test} -n \
+    -c ":unix-time abc" \
+    "${test_dir}/logfile_access_log.*"
+
+check_error_output ":unix-time works without arg?" <<EOF
+command-option:1: error: invalid unix time -- abc
+EOF
+
+
+run_test env TZ=UTC ${lnav_test} -n \
+    -c ":unix-time 1612072409" \
+    "${test_dir}/logfile_access_log.*"
+
+check_output ":unix-time does not convert time correctly?" <<EOF
+Sun Jan 31 05:53:29 2021  +0000 UTC -- 1612072409
+EOF
+
+
+run_test env TZ=UTC ${lnav_test} -n \
+    -c ":current-time" \
+    "${test_dir}/logfile_access_log.*"
+
+check_output ":current-time does not work?" <<EOF
+Thu Jun 06 19:13:20 2013  +0000 UTC -- 1370546000
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":write-to" \
+    "${test_dir}/logfile_access_log.*"
+
+check_error_output "able to write without a file name" <<EOF
+command-option:1: error: expecting file name or '-' to write to the terminal
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ";SELECT 1 AS c1, 'Hello ' || char(10) || 'World!' AS c2" \
+    -c ":write-csv-to -" \
+    "${test_dir}/logfile_access_log.*"
+
+check_output "writing CSV does not work" <<EOF
+c1,c2
+1,"Hello
+World!"
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ";SELECT 1 AS c1, 'Hello, World!' AS c2" \
+    -c ":write-table-to -" \
+    "${test_dir}/logfile_access_log.*"
+
+check_output "writing columns does not work?" <<EOF
+┏━━┳━━━━━━━━━━━━━┓
+┃c1┃     c2      ┃
+┡━━╇━━━━━━━━━━━━━┩
+│ 1│Hello, World!│
+└━━┴━━━━━━━━━━━━━┘
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ";SELECT 1 AS c1, 'Hello, World!' AS c2" \
+    -c ":write-raw-to -" \
+    "${test_dir}/logfile_access_log.*"
+
+check_output "writing raw DB does not work?" <<EOF
+ 1 Hello, World!
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":write-view-to -" \
+    "${test_dir}/logfile_access_log.0"
+
+check_output "writing raw log does not work?" <<EOF
+192.168.202.254 - - [20/Jul/2009:22:59:26 +0000] "GET /vmw/cgi/tramp HTTP/1.0" 200 134 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkboot.gz HTTP/1.0" 404 46210 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":filter-expr timeslice(:log_time_msecs, 'bad') is not null" \
+    "${test_dir}/logfile_multiline.0"
+
+check_error_output "filter-expr with bad SQL works?" <<EOF
+command-option:1: error: filter expression failed with: unable to parse time slice value: bad -- Unrecognized input
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":filter-expr :log_text LIKE '%How are%'" \
+    "${test_dir}/logfile_multiline.0"
+
+check_output "filter-expr for multiline not working" <<EOF
+2009-07-20 22:59:27,672:DEBUG:Hello, World!
+  How are you today?
+EOF
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":filter-expr not json_contains(:log_tags, '#bad')" \
+    -c ":goto 0" \
+    -c ":tag #bad" \
+    "${test_dir}/logfile_access_log.0"
+
+check_output "filter-expr for multiline not working" <<EOF
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkboot.gz HTTP/1.0" 404 46210 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
+EOF
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":filter-expr :sc_bytes > 2000" \
+    "${test_dir}/logfile_access_log.*"
+
+check_output "filter-expr not working" <<EOF
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkboot.gz HTTP/1.0" 404 46210 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":filter-expr :sc_bytes # ff" \
+    "${test_dir}/logfile_access_log.*"
+
+check_error_output "filter-expr error not working" <<EOF
+command-option:1: error: unrecognized token: "#"
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":goto 0" \
+    -c ":close" \
+    -c ":goto 0" \
+    "${test_dir}/logfile_access_log.*"
+
+check_output "close not sticking" <<EOF
+10.112.81.15 - - [15/Feb/2013:06:00:31 +0000] "-" 400 0 "-" "-"
+EOF
+
+
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":goto 0" \
+    -c ":hide-file" \
+    ${test_dir}/logfile_access_log.*
+
+check_output "hide-file with log file does not work" <<EOF
+10.112.81.15 - - [15/Feb/2013:06:00:31 +0000] "-" 400 0 "-" "-"
+EOF
+
+
 run_test ${lnav_test} -n -d /tmp/lnav.err \
     -c ":goto 0" \
     -c ":next-mark error" \
@@ -24,6 +196,16 @@ check_output "next-location is not working" <<EOF
 192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
 EOF
 
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":filter-in vmk" \
+    -c ":disable-filter vmk" \
+    ${test_dir}/logfile_access_log.0
+
+check_output "disable after :filter-in is not working" <<EOF
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkboot.gz HTTP/1.0" 404 46210 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
+EOF
+
 run_test ${lnav_test} -n -d /tmp/lnav.err \
     -c ":filter-in vmk" \
     -c ":rebuild" \
@@ -36,6 +218,18 @@ check_output "filter-in vmk is not working" <<EOF
 192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
 EOF
 
+run_test ${lnav_test} -n -d /tmp/lnav.err \
+    -c ":goto 0" \
+    -c ":filter-out vmk" \
+    -c ":toggle-filtering" \
+    ${test_dir}/logfile_access_log.0
+
+check_output "toggle-filtering is not working" <<EOF
+192.168.202.254 - - [20/Jul/2009:22:59:26 +0000] "GET /vmw/cgi/tramp HTTP/1.0" 200 134 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkboot.gz HTTP/1.0" 404 46210 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
+EOF
+
 run_test ${lnav_test} -n \
     -c ":switch-to-view help" \
     ${test_dir}/logfile_access_log.0
@@ -93,11 +287,11 @@ run_test ${lnav_test} -nvq \
     -c ":config /ui/clock-format" \
     ${test_dir}/logfile_access_log.0
 
-check_error_output "config clock-format" <<EOF
+check_error_output "config clock-format 1" <<EOF
 EOF
 
-check_output "config clock-format" <<EOF
-info: /ui/clock-format = "%a %b %d %H:%M:%S %Z"
+check_output "config clock-format 1" <<EOF
+/ui/clock-format = "%a %b %d %H:%M:%S %Z"
 EOF
 
 run_test ${lnav_test} -nvq \
@@ -106,13 +300,13 @@ run_test ${lnav_test} -nvq \
     -c ":config /ui/clock-format" \
     ${test_dir}/logfile_access_log.0
 
-check_error_output "config clock-format" <<EOF
+check_error_output "config clock-format 2" <<EOF
 EOF
 
-check_output "config clock-format" <<EOF
-info: /ui/clock-format = "%a %b %d %H:%M:%S %Z"
+check_output "config clock-format 2" <<EOF
+/ui/clock-format = "%a %b %d %H:%M:%S %Z"
 info: changed config option -- /ui/clock-format
-info: /ui/clock-format = "abc"
+/ui/clock-format = "abc"
 EOF
 
 run_test ${lnav_test} -nvq \
@@ -121,13 +315,13 @@ run_test ${lnav_test} -nvq \
     -c ":config /ui/clock-format" \
     ${test_dir}/logfile_access_log.0
 
-check_error_output "config clock-format" <<EOF
+check_error_output "config clock-format 3" <<EOF
 EOF
 
-check_output "config clock-format" <<EOF
+check_output "config clock-format 3" <<EOF
 info: changed config option -- /ui/clock-format
 info: reset option -- /ui/clock-format
-info: /ui/clock-format = "%a %b %d %H:%M:%S %Z"
+/ui/clock-format = "%a %b %d %H:%M:%S %Z"
 EOF
 
 
@@ -419,14 +613,14 @@ EOF
 
 
 TOO_MANY_FILTERS=""
-for i in `seq 1 33`; do
+for i in `seq 1 32`; do
     TOO_MANY_FILTERS="$TOO_MANY_FILTERS -c ':filter-out $i'"
 done
 run_test eval ${lnav_test} -d /tmp/lnav.err -n \
     $TOO_MANY_FILTERS \
     ${test_dir}/logfile_filter.0
 check_error_output "able to create too many filters?" <<EOF
-command-option:33: error: filter limit reached, try combining filters with a pipe symbol (e.g. foo|bar)
+command-option:32: error: filter limit reached, try combining filters with a pipe symbol (e.g. foo|bar)
 EOF
 
 
@@ -450,6 +644,13 @@ EOF
 check_output "double close is working" <<EOF
 EOF
 
+run_test ${lnav_test} -n \
+    -c ":open" \
+    ${test_dir}/logfile_access_log.0
+
+check_error_output "open does not require arg?" <<EOF
+command-option:1: error: expecting file name to open
+EOF
 
 run_test ${lnav_test} -n \
     -c ":close" \
@@ -476,16 +677,6 @@ check_output "open non-existent is not working" <<EOF
 EOF
 
 
-run_test ${lnav_test} -n \
-    -c ":goto 0" \
-    -c ":close" \
-    -c ":goto 0" \
-    "${test_dir}/logfile_access_log.*"
-
-check_output "close not sticking" <<EOF
-10.112.81.15 - - [15/Feb/2013:06:00:31 +0000] "-" 400 0 "-" "-"
-EOF
-
 
 run_test ${lnav_test} -n \
     -c ":goto 1" \
@@ -526,7 +717,8 @@ check_output "write-json-to is not working" <<EOF
         "cs_username": "-",
         "cs_version": "HTTP/1.0",
         "sc_bytes": 134,
-        "sc_status": 200
+        "sc_status": 200,
+        "cs_host": null
     },
     {
         "log_line": 1,
@@ -547,7 +739,8 @@ check_output "write-json-to is not working" <<EOF
         "cs_username": "-",
         "cs_version": "HTTP/1.0",
         "sc_bytes": 46210,
-        "sc_status": 404
+        "sc_status": 404,
+        "cs_host": null
     },
     {
         "log_line": 2,
@@ -568,7 +761,8 @@ check_output "write-json-to is not working" <<EOF
         "cs_username": "-",
         "cs_version": "HTTP/1.0",
         "sc_bytes": 78929,
-        "sc_status": 200
+        "sc_status": 200,
+        "cs_host": null
     }
 ]
 EOF
@@ -580,9 +774,9 @@ run_test ${lnav_test} -n \
     ${test_dir}/logfile_access_log.0
 
 check_output "write-jsonlines-to is not working" <<EOF
-{"log_line":0,"log_part":null,"log_time":"2009-07-20 22:59:26.000","log_idle_msecs":0,"log_level":"info","log_mark":0,"log_comment":null,"log_tags":null,"log_filters":null,"c_ip":"192.168.202.254","cs_method":"GET","cs_referer":"-","cs_uri_query":null,"cs_uri_stem":"/vmw/cgi/tramp","cs_user_agent":"gPXE/0.9.7","cs_username":"-","cs_version":"HTTP/1.0","sc_bytes":134,"sc_status":200}
-{"log_line":1,"log_part":null,"log_time":"2009-07-20 22:59:29.000","log_idle_msecs":3000,"log_level":"error","log_mark":0,"log_comment":null,"log_tags":null,"log_filters":null,"c_ip":"192.168.202.254","cs_method":"GET","cs_referer":"-","cs_uri_query":null,"cs_uri_stem":"/vmw/vSphere/default/vmkboot.gz","cs_user_agent":"gPXE/0.9.7","cs_username":"-","cs_version":"HTTP/1.0","sc_bytes":46210,"sc_status":404}
-{"log_line":2,"log_part":null,"log_time":"2009-07-20 22:59:29.000","log_idle_msecs":0,"log_level":"info","log_mark":0,"log_comment":null,"log_tags":null,"log_filters":null,"c_ip":"192.168.202.254","cs_method":"GET","cs_referer":"-","cs_uri_query":null,"cs_uri_stem":"/vmw/vSphere/default/vmkernel.gz","cs_user_agent":"gPXE/0.9.7","cs_username":"-","cs_version":"HTTP/1.0","sc_bytes":78929,"sc_status":200}
+{"log_line":0,"log_part":null,"log_time":"2009-07-20 22:59:26.000","log_idle_msecs":0,"log_level":"info","log_mark":0,"log_comment":null,"log_tags":null,"log_filters":null,"c_ip":"192.168.202.254","cs_method":"GET","cs_referer":"-","cs_uri_query":null,"cs_uri_stem":"/vmw/cgi/tramp","cs_user_agent":"gPXE/0.9.7","cs_username":"-","cs_version":"HTTP/1.0","sc_bytes":134,"sc_status":200,"cs_host":null}
+{"log_line":1,"log_part":null,"log_time":"2009-07-20 22:59:29.000","log_idle_msecs":3000,"log_level":"error","log_mark":0,"log_comment":null,"log_tags":null,"log_filters":null,"c_ip":"192.168.202.254","cs_method":"GET","cs_referer":"-","cs_uri_query":null,"cs_uri_stem":"/vmw/vSphere/default/vmkboot.gz","cs_user_agent":"gPXE/0.9.7","cs_username":"-","cs_version":"HTTP/1.0","sc_bytes":46210,"sc_status":404,"cs_host":null}
+{"log_line":2,"log_part":null,"log_time":"2009-07-20 22:59:29.000","log_idle_msecs":0,"log_level":"info","log_mark":0,"log_comment":null,"log_tags":null,"log_filters":null,"c_ip":"192.168.202.254","cs_method":"GET","cs_referer":"-","cs_uri_query":null,"cs_uri_stem":"/vmw/vSphere/default/vmkernel.gz","cs_user_agent":"gPXE/0.9.7","cs_username":"-","cs_version":"HTTP/1.0","sc_bytes":78929,"sc_status":200,"cs_host":null}
 EOF
 
 # By setting the LNAVSECURE mode before executing the command, we will disable
@@ -873,6 +1067,36 @@ check_output "histogram is not working?" <<EOF
  Sat Nov 03 08:00:00          1 normal         0 errors         0 warnings         0 marks
 EOF
 
+
+run_test ${lnav_test} -n \
+    -c ":mark-expr" \
+    ${test_dir}/logfile_syslog.0
+
+check_error_output ":mark-expr works without an expression?" <<EOF
+command-option:1: error: expecting an SQL expression
+EOF
+
+
+run_test ${lnav_test} -n \
+    -c ":mark-expr :log_procname lik" \
+    ${test_dir}/logfile_syslog.0
+
+check_error_output ":mark-expr works with a bad expression?" <<EOF
+command-option:1: error: near "lik": syntax error
+EOF
+
+
+run_test ${lnav_test} -n \
+    -c ":mark-expr :cs_uri_stem LIKE '%vmk%'" \
+    -c ":write-to -" \
+    ${test_dir}/logfile_access_log.0
+
+check_output "mark-expr is not working?" <<EOF
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkboot.gz HTTP/1.0" 404 46210 "-" "gPXE/0.9.7"
+192.168.202.254 - - [20/Jul/2009:22:59:29 +0000] "GET /vmw/vSphere/default/vmkernel.gz HTTP/1.0" 200 78929 "-" "gPXE/0.9.7"
+EOF
+
+
 run_test ${lnav_test} -n \
     -c ":goto 0" \
     -c ":mark" \
@@ -901,15 +1125,44 @@ check_output "multiline commands do not work?" <<EOF
 Hello: Jules
 EOF
 
-
-printf 'Hello, World!' | run_test ${lnav_test} -n \
+printf "Hello, World!" | run_test ${lnav_test} -n \
   -c ":switch-to-view text"
+test_num=`expr ${test_num} \+ 1`
 
 check_output "stdin with no line feed failed" <<EOF
 Hello, World!
 EOF
 
 
+run_test ${lnav_test} -Nnv \
+    -c ":hide-lines-before 2009-07-20T22:59:29" \
+    -c ":hide-lines-before"
+
+check_output "hide-lines-before with no args does not work?" <<EOF
+info: hiding lines before 2009-07-20 22:59:29.000
+info: hiding lines before 2009-07-20 22:59:29.000
+EOF
+
+run_test ${lnav_test} -Nnv \
+    -c ":hide-lines-after 2009-07-20T22:59:29" \
+    -c ":hide-lines-after"
+
+check_output "hide-lines-after with no args does not work?" <<EOF
+info: hiding lines after 2009-07-20 22:59:29.000
+info: hiding lines after 2009-07-20 22:59:29.000
+EOF
+
+run_test ${lnav_test} -Nnv \
+    -c ":hide-lines-before 2009-07-20T22:00:29" \
+    -c ":hide-lines-after 2009-07-20T22:59:29" \
+    -c ":hide-lines-before"
+
+check_output "hide-lines-after with no args does not work?" <<EOF
+info: hiding lines before 2009-07-20 22:00:29.000
+info: hiding lines after 2009-07-20 22:59:29.000
+info: hiding lines before 2009-07-20 22:00:29.000 and after 2009-07-20 22:59:29.000
+EOF
+
 run_test ${lnav_test} -n \
     -c ":hide-lines-before 2009-07-20T22:59:29" \
     ${test_dir}/logfile_access_log.0
@@ -941,7 +1194,7 @@ EOF
 export XYZ="World"
 
 run_test ${lnav_test} -n \
-    -c ':echo Hello, $XYZ!' \
+    -c ':echo Hello, \$XYZ!' \
     ${test_dir}/logfile_access_log.0
 
 check_output "echo hello" <<EOF
@@ -961,7 +1214,7 @@ EOF
 
 
 run_test ${lnav_test} -n \
-    -c ':eval :echo Hello, $XYZ!' \
+    -c ':echo Hello, $XYZ!' \
     ${test_dir}/logfile_access_log.0
 
 check_output "eval echo hello" <<EOF
