@@ -1,4 +1,5 @@
 
+#include <stdlib.h>
 #include "config.h"
 #define _XOPEN_SOURCE_EXTENDED 1
 #include <locale.h>
@@ -19,6 +20,7 @@
 
 int main(int argc, char *argv[])
 {
+    setenv("LANG", "en_US.utf-8", 1);
     setlocale(LC_ALL, "");
 
     WINDOW *stdscr = initscr();
