@@ -34,6 +34,7 @@
 
 #include "lnav_config.hh"
 #include "top_status_source.hh"
+#include "command_executor.hh"
 
 using namespace std;
 
@@ -47,15 +48,6 @@ int gettimeofday(struct timeval * tp, void * tzp)
     return 0;
 }
 
-string execute_any(exec_context &ec, const string &cmdline_with_mode)
-{
-    return "";
-}
-
-void add_global_vars(exec_context &ec)
-{
-}
-
 int main(int argc, char *argv[])
 {
     int retval = EXIT_SUCCESS;
@@ -65,7 +57,7 @@ int main(int argc, char *argv[])
     setenv("HOME", "/", 1);
 
     vector<string> errors;
-    vector<filesystem::path> paths;
+    vector<ghc::filesystem::path> paths;
 
     load_config(paths, errors);
 
